/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    AprioriSets.java 
 *    by Zack Stoecker-Sylvia, Worcester Polytechnic Institute, 2002
 *      Based on 
 *      Apriori.java, Copyright (C) 1999 Eibe Frank,Mark Hall
 *      and 
 *      ARMiner, Copyright (C) 2000  UMass/Boston - Computer Science Department
 *
 */

package wpi.associations;

import java.io.*;
import java.util.*;

import weka.core.*;
import weka.filters.Filter;
import weka.filters.AttributeFilter;
import weka.associations.*;

import wpi.associations.arminer.*;

/*
  Maintainance log started on January a3th, 2003

  January 13th, 2003 - changed buildAssociations to mineAssociations
                       so that we could override the buildAssociations
		       method in our own abstract Associator class.
		       This facilitated the measuring of mining time
		       across all our association rule mining classes.

		     - added main method for testing

		     - cleaned up some formatting -kap
*/

/**
 * Class implementing an Apriori-type algorithm that may understand sets.  
 * Sets must be loaded as strings into WEKA and are then parsed using ^ or &
 * to delimit items.  Items are split into individual Attribute-Value pairs 
 * for the item based ARMiner algorithm.  WEKA iteratively reduces the minimum
 * support until it finds the required number of rules with the given minimum 
 * confidence. <p>
 *
 * This class implements the Apriori algorithm 
 * for finding large itemsets.
 *
 * (see "Fast Algorithms for Mining Association Rules"
 * by Rakesh Agrawal and Ramakrishnan Srikant
 * from IBM Almaden Research Center 1994)
 *
 * Valid options are:<p>
 *
 * -N required number of rules <br>
 * The required number of rules (default: 10). <p>
 * 
 * -C minimum confidence score of a rule <br>
 * The minimum confidence of a rule (default: 0.9). <p>
 * 
 * -D delta for minimum support <br>
 * The delta by which the minimum support is decreased in
 * each iteration (default: 0.05).
 * 
 * -U upper bound for minimum support <br>
 * The upper bound for minimum support. Don't explicitly look for 
 * rules with more than this level of support. <p>
 * 
 * -M lower bound for minimum support <br>
 * The lower bound for the minimum support (default = 0.1). <p>
 * 
 * -V <br>
 * If set then progress is reported iteratively during execution. <p>
 * 
 * -A <br>
 * The range of required attributes for the antecedents. <p>
 * 
 * -B <br>
 * The maximum number of antecedent items. <p>
 * 
 * -Y <br>
 * The range of required attributes for the consequents. <p>
 * 
 * -Z <br>
 * The minimum number of consequent items. <p>
 * 
 * @author Eibe Frank (eibe@cs.waikato.ac.nz)
 * @author Mark Hall (mhall@cs.waikato.ac.nz)
 * @author Stoecker-Sylvia (zss@wpi.edu) Worcester Polytechnic Institute, 
 * @version 1.11
 */

public class AprioriSets extends Associator implements OptionHandler 
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The minimum support. */
  protected double m_minSupport;
  
  /** The upper bound on the support */
  protected double m_upperBoundMinSupport;
  
  /** The lower bound for the minimum support. */
  protected double m_lowerBoundMinSupport;
  
  /** The minimum confidence score. */
  protected double m_minConfidence;
  
  /** The maximum number of rules that are output. */
  protected int m_maxRulesOutput;
  
  /** Verbose mode? */
  protected boolean m_verbose;
  
  /** Delta by which m_minSupport is decreased in each iteration. */
  protected double m_delta;
  
  /** Number of cycles used before required number of rules was one. */
  protected int m_cycles;
  
  /** The set of all sets of itemsets L. */
  protected FastVector m_Ls;
  
  /** The same information stored in hash tables. */
  protected FastVector m_hashtables;
  
  /** The list of all generated rules. */
  protected FastVector[] m_allTheRules;
  
  /** The instances (transactions) to be used for generating 
   *  the association rules. */
  protected Instances m_instances;
  
  /** Output itemsets found? */
  protected boolean m_outputItemSets;
  
  /********* zss: New options ************/
  /** The maximum number of antecedents in a rule */
  protected int m_maxAntecedents;
  
  /** The minimum number of consequents in a rule */
  protected int m_minConsequents;
  
  /** The vector containing all found AssociationRules */
  protected Vector m_rulesFound;
  
  /** The attributes required to be in the antecedents of the rules */
  protected Range m_requiredAntecedentAttributes = new Range();
  
  /** The attributes required to be in the consequents of the rules */
  protected Range m_requiredConsequentAttributes = new Range();
	
  /** A hash listing attribute value pair Strings by Integer key */
  protected Hashtable m_attributeHash;

//********************************************************//
// ----- Methods-----
//********************************************************//
  
  /**
   * Returns a string describing this associator
   * @return a description of the evaluator suitable for
   * displaying in the explorer/experimenter gui
   */

  public String globalInfo()
  {
    return ( "Finds association rules." );
  }

//********************************************************//
  
  /**
   * Constructor that allows to sets default values for the 
   * minimum confidence and the maximum number of rules
   * the minimum confidence.
   */

  public AprioriSets() 
  {
    resetOptions();
  }

//********************************************************//
  
  /**
   * Resets the options to the default values.
   */

  public void resetOptions() 
  {
    m_verbose = false;
    m_delta = 0.05;
    m_minConfidence = 0.90;
    m_maxRulesOutput = 10;
    m_lowerBoundMinSupport = 0.1;
    m_upperBoundMinSupport = 1.0;
    m_outputItemSets = false;
    
    m_requiredAntecedentAttributes.setRanges ( "" );
    m_requiredConsequentAttributes.setRanges ( "" );
    m_maxAntecedents = 0;
    m_minConsequents = 0;
  }

//********************************************************//

  /*
     Changing the method below to be named mineAssociations
     instead of buildAssociations. This was done so that we
     could measure the time it takes to mine rules in the
     abstract class that this class extends.
     -kap
  */
  
  /**
   * Method that generates all large itemsets with a minimum support, and from
   * these all association rules with a minimum confidence.
   * 
   * @param instances
   * the instances to be used for generating the associations
   *
   * @exception	Exception	
   * if rules can't be built successfully
   */
  
  public void mineAssociations ( Instances instances ) 
    throws Exception 
  {
    m_instances = instances;
    m_requiredAntecedentAttributes.setUpper ( instances.numAttributes() - 1 );
    m_requiredConsequentAttributes.setUpper ( instances.numAttributes() - 1 );
    
    DBReader instanceReader = new DBReader ( instances );
    m_attributeHash = instanceReader.getNumberHash();

    int necSupport = 0;
    Vector requiredAntecedents = new Vector();
    Vector requiredConsequents = new Vector();
    ARMinerApriori itemsetMiner = new ARMinerApriori();
    AprioriRules rulesFinder = new AprioriRules();
    
    for ( int i = 0; i < instances.numAttributes(); i++ )
    {
      if ( m_requiredAntecedentAttributes.isInRange ( i ) )
      {
	requiredAntecedents.addElement ( instanceReader.
					 getAttributeItemset ( i ) );
      }
      
      if ( m_requiredConsequentAttributes.isInRange ( i ) )
      {
	requiredConsequents.addElement ( instanceReader.
					 getAttributeItemset ( i ) );
      }
    }
    
    if ( requiredAntecedents.isEmpty() )
    {
      requiredAntecedents = null;
    }
    
    if ( requiredConsequents.isEmpty() )
    {
      requiredConsequents = null;
    }
    
    // Decrease minimum support until desired number of rules found.
    m_cycles = 0;
    m_minSupport = m_upperBoundMinSupport - m_delta;
    m_minSupport = ( m_minSupport < m_lowerBoundMinSupport )
      ? m_lowerBoundMinSupport 
      : m_minSupport;
    do 
    {
      itemsetMiner.findLargeItemsets ( instanceReader,
				       new DBCacheWriter ( "temp.cache" ),
				       (float)m_minSupport );

      m_rulesFound = rulesFinder.findAssociations 
	( new DBCacheReader ( "temp.cache"),
	  (float)m_minSupport,
	  (float)m_minConfidence,
	  requiredAntecedents,
	  requiredConsequents,
	  m_maxAntecedents,
	  m_minConsequents );

      m_minSupport -= m_delta;
      m_minSupport = ( m_minSupport < m_lowerBoundMinSupport ) 
	? 0 
	: m_minSupport;
      necSupport = (int)( m_minSupport * (double)instances.numInstances() + 
			  0.5 );
      m_cycles++;
      
      if ( m_verbose ) 
      {
	if ( m_Ls.size() > 1 ) 
	{
	  System.out.println ( toString() );
	}
      }
    }
    while ( ( m_rulesFound.size() < m_maxRulesOutput ) && 
	   ( m_minSupport >= m_lowerBoundMinSupport ) && 
	   ( m_minSupport * (double)instances.numInstances() + 0.5 >= 1 ) );
    {
      m_minSupport += m_delta;
    }
  } // END public void mineAssociations ( Instances instances ) 
  //               throws Exception 

//********************************************************//
  
  /**
   * Returns an enumeration describing the available options
   * 
   * @return an enumeration of all the available options
   */

  public Enumeration listOptions() 
  {
    String 
      maxRulesOutputHint = "\tThe required number of rules. (default = " + 
      m_maxRulesOutput + ")",
      minConfidenceHint = "\tThe minimum confidence of a rule. (default = " + 
      m_minConfidence + ")",
      deltaHint = 
      "\tThe delta by which the minimum support is decreased in\n" +
      "\teach iteration. (default = " + m_delta + ")",
      lowerBoundMinSupportHint = 
      "\tThe lower bound for the minimum support. (default = " + 
      m_lowerBoundMinSupport + ")",
      upperBoundMinSupportHint = 
      "\tThe upper bound for the minimum support. (default = 1.0)",
      verboseHint = 
      "\tIf set the itemsets found are also output. (default = no)",
      requiredAntecedentAttributesHint = 
      "\tThe set of attributes required to be in the antecedent.",
      requiredConsequentAttributesHint = 
      "\tThe set of attributes required to be in the consequent.",
      maxAntecedentsHint = 
      "\tThe maximum number of attributes in the antecedent. (0 = no maximum)",
      minConsequentsHint = 
      "\tThe minimum number of attributes in the consequent. (0 = no minimum)";
    
    FastVector newVector = new FastVector ( 10 );
    
    newVector.addElement 
      ( new Option ( maxRulesOutputHint, "N", 1, 
		     "-N <required number of rules output>" ) );
    newVector.addElement 
      ( new Option ( minConfidenceHint, "C", 1, 
		     "-C <minimum confidence score of a rule>" ) );
    newVector.addElement ( new Option ( deltaHint, "D", 1,
					"-D <delta for minimum support>" ) );
    newVector.addElement 
      ( new Option ( upperBoundMinSupportHint, "U", 1,
		     "-U <upper bound for minimum support>" ) );
    newVector.addElement 
      ( new Option ( lowerBoundMinSupportHint, "M", 1,
		     "-M <lower bound for minimum support>" ) );
    newVector.addElement ( new Option ( verboseHint, "V", 0,
					"-V" ) );
    newVector.addElement 
      ( new Option ( requiredAntecedentAttributesHint, "A", 1, 
		     "-A <range of required antecedent attributes>" ) );
    newVector.addElement 
      ( new Option ( requiredConsequentAttributesHint, 
		     "Y", 1, 
		     "-Y <range of required consequent attributes>" ) );
    newVector.addElement ( new Option ( maxAntecedentsHint, "B", 1, 
					"-B <maximum antecedent items>" ) );
    newVector.addElement ( new Option ( minConsequentsHint, "Z", 1, 
					"-Z <minimum consequent items>" ) );

    return ( newVector.elements() );
  }

//********************************************************//
  
  /**
   * Parses a given list of options. Valid options are:<p>
   * 
   * -N required number of rules <br>
   * The required number of rules (default: 10). <p>
   * 
   * -C minimum confidence score of a rule <br>
   * The minimum confidence of a rule (default: 0.9). <p>
   * 
   * -D delta for minimum support <br>
   * The delta by which the minimum support is decreased in
   * each iteration (default: 0.05).
   * 
   * -U upper bound for minimum support <br>
   * The upper bound for minimum support. Don't explicitly look for 
   * rules with more than this level of support. <p>
   * 
   * -M lower bound for minimum support <br>
   * The lower bound for the minimum support (default = 0.1). <p>
   * 
   * -V <br>
   * If set then progress is reported iteratively during execution. <p>
   * 
   * -A <br>
   * The range of required attributes for the antecedents. <p>
   * 
   * -B <br>
   * The maximum number of antecedent items. <p>
   * 
   * -Y <br>
   * The range of required attributes for the consequents. <p>
   * 
   * -Z <br>
   * The minimum number of consequent items. <p>
   * 
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported 
   */

  public void setOptions ( String[] options ) throws Exception
  {
    resetOptions();
    
    String numRulesString = Utils.getOption ( 'N', options ),
      minConfidenceString = Utils.getOption ( 'C', options ),
      deltaString = Utils.getOption ( 'D', options ),
      maxSupportString = Utils.getOption ( 'U', options ),
      minSupportString = Utils.getOption ( 'M', options ),
      requiredAntecedentsString = Utils.getOption ( 'A', options ),
      requiredConsequentsString = Utils.getOption ( 'Y', options ),
      maxAntecedentsString = Utils.getOption ( 'B', options ),
      minConsequentsString = Utils.getOption ( 'Z', options );
    
    m_verbose = Utils.getFlag ( 'V', options );
    
    if ( numRulesString.length() != 0 )
    {
      m_maxRulesOutput = Integer.parseInt ( numRulesString );
    }
    if ( minConfidenceString.length() != 0 ) 
    {
      m_minConfidence = ( new Double ( minConfidenceString ) ).doubleValue();
    }
    if ( deltaString.length() != 0 ) 
    {
      m_delta = ( new Double ( deltaString ) ).doubleValue();
    }
    if ( maxSupportString.length() != 0 ) 
    {
      setUpperBoundMinSupport ( ( new Double ( maxSupportString ) ).
				doubleValue() );
    }
    if ( minSupportString.length() != 0 ) 
    {
      m_lowerBoundMinSupport = ( new Double ( minSupportString ) ).
	doubleValue();
    }
    if ( requiredAntecedentsString.length() != 0 )
    {
      setRequiredAntecedents ( requiredAntecedentsString );
    }
    if ( requiredConsequentsString.length() != 0 )
    {
      setRequiredConsequents ( requiredConsequentsString );
    }
    if ( maxAntecedentsString.length() != 0 )
    {
      m_maxAntecedents = ( new Integer ( maxAntecedentsString ).intValue() );
    }
    if ( minConsequentsString.length() != 0 )
    {
      m_minConsequents = ( new Integer ( minConsequentsString ).intValue() );
    }
  }

//********************************************************//
  
  /**
   * Gets the current settings of the Apriori object.
   * 
   * @return an array of strings suitable for passing to setOptions
   */

  public String [] getOptions() 
  {
    String [] options = new String [ 20 ];
    int current = 0;
    
    options [ current++ ] = "-N"; options [ current++ ] = "" + 
				    m_maxRulesOutput;
    options [ current++ ] = "-C"; options [ current++ ] = "" + 
				    m_minConfidence;
    options [ current++ ] = "-D"; options [ current++ ] = "" + 
				    m_delta;
    options [ current++ ] = "-U"; options [ current++ ] = "" + 
				    m_upperBoundMinSupport;
    options [ current++ ] = "-M"; options [ current++ ] = "" +
				    m_lowerBoundMinSupport;
    
    if ( getRequiredAntecedents() != "" )
    {
      options [ current++ ] = "-A"; options [ current++ ] = 
				      getRequiredAntecedents();
    }
    if ( getRequiredConsequents() != "" )
    {
      options [ current++ ] = "-Y"; options [ current++ ] = 
				      getRequiredConsequents();
    }
    
    options [ current++ ] = "-B"; options [ current++ ] = "" + 
				    m_maxAntecedents;
    options [ current++ ] = "-Z"; options [ current++ ] = "" + 
				    m_minConsequents;
    
    while ( current < options.length ) 
    {
      options [ current++ ] = "";
    }
    
    return ( options );
  }

//********************************************************//
  
  /**
   * Outputs the association rules and their confidences and supports.
   */

  public String toString() 
  {
    StringBuffer text = new StringBuffer();
    Enumeration e;
    AssociationRule as;
    int i;
    boolean comma = false;
    
    text.append ( "Required Attributes in Antecedents:\n" );
    
    for ( i = 0; i < m_instances.numAttributes(); i++ )
    {
      if ( m_requiredAntecedentAttributes.isInRange ( i ) )
      {
	if ( comma )
	{
	  text.append ( ", " );
	}
	else
	{
	  comma = true;
	}
	
	text.append ( m_instances.attribute ( i ).name() );
      }
    }
    
    if ( ! comma )
    {
      text.append ( "none" );
    }
    
    comma = false;
    
    text.append ( "\n\nRequired Attributes in Consequents:\n" );
    
    for ( i = 0; i < m_instances.numAttributes(); i++ )
    {
      if ( m_requiredConsequentAttributes.isInRange ( i ) )
      {
	if ( comma )
	{
	  text.append ( ", " );
	}
	else
	{
	  comma = true;
	}
	
	text.append ( m_instances.attribute ( i ).name() );
      }
    }
    
    if ( ! comma )
    {
      text.append ( "none" );
    }
    
    text.append ( "\n\n" );
    
    if ( m_rulesFound.isEmpty() )
    {
      text.append ( "No rules found!\n" );
    }
    else
    {
      Vector v1 = (Vector)m_rulesFound.clone();
      Vector v2 = new Vector ( m_rulesFound.size() );
      float highConf;
      float highSup;
      int bestRuleIndex;
      int rulesOutput = 0;
		
      while ( ! v1.isEmpty() && ( rulesOutput < m_maxRulesOutput ) )
      {
	rulesOutput++;
	highConf = -1;
	highSup = -1;
	bestRuleIndex = -1;
	
	for ( i = 0; i < v1.size(); i++ )
	{
	  as = (AssociationRule)v1.elementAt ( i );
	  
	  if ( ( as.getConfidence() > highConf ) || 
	       ( ( as.getConfidence() == highConf ) && 
		 ( as.getSupport() > highSup ) ) )
	  {
	    highConf = as.getConfidence();
	    highSup = as.getSupport();
	    bestRuleIndex = i;
	  }
	}
	
	v2.addElement ( v1.elementAt ( bestRuleIndex ) );
	v1.removeElementAt ( bestRuleIndex );
      }
      
      e = v2.elements();
      
      rulesOutput = 0;
      
      while ( e.hasMoreElements() && ( rulesOutput < m_maxRulesOutput ) )
      {
	rulesOutput++;
	if ( rulesOutput < 10 )
	{
	  text.append ( " " + rulesOutput + ". " );
	}
	else
	{
	  text.append ( "" + rulesOutput + ". " );
	}
	
	as = (AssociationRule)e.nextElement();
	
	for ( i = 0; i < as.antecedentSize(); i++ )
	{
	  if ( i > 0 )
	  {
	    text.append ( " && " );
	  }
	  
	  text.append ( (String)m_attributeHash.
			get ( new Integer ( as.getAntecedentItem ( i ) ) ) );
	}
	
	text.append ( " ==> " );
	
	for ( i = 0; i < as.consequentSize(); i++ )
	{
	  if ( i > 0 )
	  {
	    text.append ( " && " );
	  }
	  
	  text.append ( (String)m_attributeHash.
			get ( new Integer ( as.getConsequentItem ( i ) ) ) );
	}
	
	text.append ( "  [Conf: " + as.getConfidence() + 
		      ", Sup: " + as.getSupport() + "]\n" );
      }
    }
    
    return ( text.toString() );
  }

//********************************************************//
  
  public String requiredAntecedentsTipText()
  {
    return "Attributes required to be in the antecedent of the rules.";
  }

//********************************************************//

  public void setRequiredAntecedents ( String s )
  {
    m_requiredAntecedentAttributes.setRanges ( s );
  }

//********************************************************//

  public String getRequiredAntecedents()
  {
    return ( m_requiredAntecedentAttributes.getRanges() );
  }

//********************************************************//

  public String requiredConsequentsTipText()
  {
    return ( "Attributes required to be in the consequent of the rules." );
  }

//********************************************************//

  public void setRequiredConsequents ( String s )
  {
    m_requiredConsequentAttributes.setRanges ( s );
  }

//********************************************************//

  public String getRequiredConsequents()
  {
    return ( m_requiredConsequentAttributes.getRanges() );
  }

//********************************************************//
  
  public String maxAntecedentsTipText()
  {
    return ( "Maximum number of items allowed in the antecedent of the " +
	     "rules." );
  }

//********************************************************//

  public void setMaxAntecedents ( int i )
  {
    m_maxAntecedents = i;
  }

//********************************************************//

  public int getMaxAntecedents()
  {
    return ( m_maxAntecedents );
  }

//********************************************************//

  public String minConsequentsTipText()
  {
    return ( "Minimum number of items allowed in the consequent of the " +
	     "rules." );
  }

//********************************************************//

  public void setMinConsequents ( int i )
  {
    m_minConsequents = i;
  }

//********************************************************//

  public int getMinConsequents()
  {
    return ( m_minConsequents );
  }

//********************************************************//

  public String upperBoundMinSupportTipText() 
  {
    return ( "Upper bound for minimum support. Start iteratively decreasing "
	     + "minimum support from this value." );
  }

//********************************************************//

  public double getUpperBoundMinSupport()
  {
    return ( m_upperBoundMinSupport );
  }

//********************************************************//

  public void setUpperBoundMinSupport ( double v ) 
  {
    m_upperBoundMinSupport = v;
  }

//********************************************************//

  public String lowerBoundMinSupportTipText() 
  {
    return ( "Lower bound for minimum support." );
  }

//********************************************************//

  public double getLowerBoundMinSupport() 
  {
    return ( m_lowerBoundMinSupport );
  }

//********************************************************//

  public void setLowerBoundMinSupport ( double v ) 
  {
    m_lowerBoundMinSupport = v;
  }

//********************************************************//

  public String minMetricTipText() 
  {
    return ( "Minimum confidence score. Consider only rules with scores " + 
	     "higher than this value." );
  }

//********************************************************//

  public double getMinMetric() 
  {
    return ( m_minConfidence );
  }

//********************************************************//

  public void setMinMetric ( double v ) 
  {
    m_minConfidence = v;
  }

//********************************************************//

  public String numRulesTipText() 
  {
    return ( "Number of rules to find." );
  }

//********************************************************//

  public int getNumRules() 
  {
    return m_maxRulesOutput;
  }

//********************************************************//

  public void setNumRules(int v) 
  {
    m_maxRulesOutput = v;
  }

//********************************************************//

  public String deltaTipText() 
  {
    return ( "Iteratively decrease support by this factor. Reduces support "+ 
	     "until min support is reached or required number of rules has " +
	     "been generated." );
  }

//********************************************************//

  public double getDelta() 
  {
    return ( m_delta );
  }

//********************************************************//

  public void setDelta ( double v ) 
  {
    m_delta = v;
  }

//********************************************************//
  
  /**
   * Main method for testing this class.
   */

  public static void main ( String[] options ) 
  {
    String trainFileString;
    StringBuffer text = new StringBuffer();
    AprioriSets as = new AprioriSets();
    Reader reader;

    try 
    {
      text.append ( "\n\nApriori Sets options:\n\n" );
      text.append ( "-t <training file>\n" );
      text.append ( "\tThe name of the training file.\n" );

      Enumeration enum = as.listOptions();

      while ( enum.hasMoreElements() )
      {
	Option option = (Option)enum.nextElement();
	text.append ( option.synopsis() + '\n' );
	text.append ( option.description() + '\n' );

      } // end while more options

      trainFileString = Utils.getOption ( 't', options );

      if ( trainFileString.length() == 0 ) 
      {
	throw new Exception ( "No training file given!" );
      }
      
      as.setOptions ( options );

      reader = new BufferedReader ( new FileReader ( trainFileString ) );

      as.buildAssociations ( new Instances ( reader ) );

      System.out.println ( as );

    } // end try
    catch ( Exception e ) 
    {
      e.printStackTrace();
      System.out.println ( "\n" + e.getMessage() + text );

    } //end catch

  } // END public static void main ( String[] options ) 

//********************************************************//

} // END 



