/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Associator.java
 *    Copyright (C) 2003 Keith A. Pray
 *
 */

package wpi.associations;

import wpi.associations.arminerSequence.*;
import wpi.timesequence.TSUtility;

import weka.core.Instance;
import weka.core.Instances;
import weka.core.Attribute;
import weka.core.SerializedObject;
import weka.core.Utils;

import java.util.*;

/*
  Maintainance log started on April 30th, 2002

  January 8, 2004     - added data members logger and logStatsFileName -kap
                      - added initialization of Logger mechanism to
		        buildAssociations method -kap

  January 12, 2004    - added logging of NUM_INSTANCES, NUM_EVENT_ATTRIBUTES,
                        NUM_EVENT_OCCURRENCES, AVERAGE_LENGTH_OF_TIMELINE,
		        and TOTAL_SECONDS_TO_MINE to buildAssociations
			method -kap

  Febraury 22, 2004   - added calculation and logging of min-event-length,
                        max-event-length, min-per-event-length, and 
			max-per-event-length -kap
*/

/** 
 * Abstract class that measures the time it takes to mine
 * frequent itemsets and reports this to standard out.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version 1.0
 */

public abstract class Associator extends weka.associations.Associator
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The object responsible for logging. */
  protected Logger logger = null;

  /** The name of the log file to write to. */
  protected String logStatsFileName = "";

  /** Specifies debug info level 
   0: no debug info
   1: input to methods
   2: and output from methods
   3: and all sorts of stuff */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Generates an associator. Must initialize all fields
   * of the associator that are not being set via options
   * (ie. multiple calls of buildAssociator must always
   * lead to the same result). Must not change the dataset
   * in any way. This takes the place of the WEKA
   * buildAssociations method.
   *
   * @param data set of instances serving as training data 
   * @exception Exception if the associator has not been 
   * generated successfully
   */

  public abstract void mineAssociations ( Instances data )
    throws Exception;

//********************************************************//
 
  /**
   * Generates an associator. Must initialize all fields of the associator
   * that are not being set via options (ie. multiple calls of buildAssociator
   * must always lead to the same result). Must not change the dataset
   * in any way.
   *
   * @param data set of instances serving as training data 
   * @exception Exception if the associator has not been 
   * generated successfully
   */

  public void buildAssociations ( Instances data ) 
    throws Exception
  {
    // first, set up logging if required
    // are we specifying a file name?
    if ( logStatsFileName.length() != 0 )
    {
      System.out.println ( "\nAssociator.buildAssociations: " +
			   "Setting up log information..." );

      logger = new Logger ( logStatsFileName );

      System.out.println ( "Calculating data set statistics. " +
			   "Time will vary with data set size." );

      // log the time now
      logger.logValue ( Logger.TIMESTAMP,
			new Date().toString() );

      // log the relation name of the data set
      logger.logValue ( Logger.RELATION_NAME,
			data.relationName() );


      // let's get some initial stats about the data set
      logger.logValue ( Logger.NUM_INSTANCES, 
			Integer.toString ( data.numInstances() ) );
      
      System.out.println ("Logging data set size: " + data.numInstances() +
			  " ..." );

      // count number of event attributes

      // get an enumeration of the attributes
      Enumeration attributeEnumeration = data.enumerateAttributes();
      
      // need a temp attribute
      Attribute a;

      // and a temp string
      String s;

      // will need an enumeration of the value from each attribute
      Enumeration valueEnumeration;

      // might as well keep track of the event attributes
      Vector eventAttributeVector = new Vector();

      // and the time sequence attributes
      Vector timeSequenceAttributeVector = new Vector();

      // can we say it is an event set so far
      boolean isEventSet;

      // can we say it is a time sequence so far
      boolean isTimeSequence;

      while ( attributeEnumeration.hasMoreElements() )
      {
	a = (Attribute)attributeEnumeration.nextElement();

	// it has to be a string attribute
	if ( ! a.isString() )
	{
	  // skip the rest of this and move onto the next attribute
	  continue;

	} // end if attribute not string type

	// let's start optimistic
	isEventSet = true;
	isTimeSequence = true;

	// check all the values of this attribute and make
	// sure they are all events
	valueEnumeration = a.enumerateValues();

	while ( valueEnumeration.hasMoreElements() )
	{
	  s = (String)valueEnumeration.nextElement();

	  if ( ! TSUtility.isTimeSequenceEventSet ( s ) )
	  {
	    switch ( debug ) {
	    case 3: //case 0:
	      System.out.println ( "Associator.buildAssociations: " +
				   a + 
				   " is not time sequence event attribute " +
				   "because value: " + s +
				   " is not an event set." );
	    } // end switch debug

	    isEventSet = false;

	  } // end if not an event set

	  // is the value a time sequence
	  if ( ! TSUtility.isTimeSequence ( s ) )
	  {
	    isTimeSequence = false;

	  } // end if value is not time sequence

	  // if value was neither event set nor time sequence we can
	  // stop and continue witht he next attribute
	  if ( ! isEventSet && ! isTimeSequence )
	  {
	    // skip the rest of this and move onto the next attribute
	    break;

	  } // end if neither event set nor time sequence

	} // end while more values

	if ( isEventSet )
	{
	  // ok, if it made it this far it should be an event attribute
	  // add it to our list
	  eventAttributeVector.add ( a );

	} // end is event set

	if ( isTimeSequence )
	{
	  // add it to our list
	  timeSequenceAttributeVector.add ( a );

	} // end is time sequence

      } // end while more attributes

      // the number of other attributes is the total number minus
      // the number of event attributes
      logger.logValue ( Logger.NUM_ATTRIBUTES,
			Integer.toString ( data.numAttributes() - 
					   eventAttributeVector.size() ) );

      System.out.println ( "Logging number attributes: " + 
			   logger.getValue ( Logger.NUM_ATTRIBUTES ) + 
			   " ..." );

      // log the number of event attributes found
      logger.logValue ( Logger.NUM_EVENT_ATTRIBUTES, 
			Integer.toString ( eventAttributeVector.size() ) );

      System.out.println ( "Logging number event attributes: " + 
			   eventAttributeVector.size() + " ..." );

      // count number of event occurrences and set up calculation
      // of average time line length

      // while we are at it let's track the min and max length of
      // the events.
      double minEventLength = Double.MAX_VALUE;
      double maxEventLength = 0.0;

      double tempEventLength;

      // let's also capture these statistics per event attribute
      double [] minEventLengthArray = 
	new double [ eventAttributeVector.size() ];

      Arrays.fill ( minEventLengthArray, Double.MAX_VALUE );

      double [] maxEventLengthArray = 
	new double [ eventAttributeVector.size() ];

      Arrays.fill ( maxEventLengthArray, 0.0 );

      // is there a timeline attribute?
      a = data.attribute ( TSUtility.TIME_LINE_NAME );

      // to use time line for average time line length
      double numTimeUnitsTimeLine;

      // to use sequence for average time line length
      double numTimeUnitsSequence;

      // to use events for average time line length
      double numTimeUnitsEvent;

      // the total time units found regardless of source
      double numTimeUnits = 0.0;

      // number of instance found with time line info
      double numInstancesWithTimeLineInfo = 0.0;

      // did we find time line length info
      boolean foundTimeLineInfo;

      // to do this we must look at each instance in the data set
      Enumeration instanceEnumeration = data.enumerateInstances();

      // need a temp instance
      Instance instance;

      // iterating index
      int i;

      // a temp array for events
      String [] eventArray;

      // a temp event
      double [] tempEventTime;

      // track number of event occurrences
      int numEventOccurrences = 0;

      while ( instanceEnumeration.hasMoreElements() )
      {
	instance = (Instance)instanceEnumeration.nextElement();

	// did we find time line length info
	// assume no so far
	foundTimeLineInfo = false;

	numTimeUnitsTimeLine = 0.0;
	numTimeUnitsSequence = 0.0;
	numTimeUnitsEvent = 0.0;
	
	// if there's a time line attribute add the number of
	// time units to a running total
	if ( a != null )
	{
	  // is the time line present for this instance
	  if ( ! instance.isMissing ( a ) )
	  {
	    // how many time units?
	    numTimeUnitsTimeLine += TSUtility.getValues 
	      ( instance.stringValue ( a ) ).length;

	    // remember to say we found time line info
	    foundTimeLineInfo = true;

	  } // end if time line present for this instance

	} // end if there is a time line attribute

	// go through the sequence attributes and record the max length
	for ( i = 0; i < timeSequenceAttributeVector.size(); i++ )
	{
	  // make sure there's a value for the attribute
	  if ( ! instance.isMissing 
	       ( (Attribute)timeSequenceAttributeVector.elementAt ( i ) ) )
	  { 
	    numTimeUnitsSequence = 
	      Math.max 
	      ( numTimeUnitsSequence,
		TSUtility.getValues 
		( instance.stringValue 
		  ( (Attribute)timeSequenceAttributeVector.elementAt 
		    ( i ) ) ).length );

	    // remember to say we found time line info
	    foundTimeLineInfo = true;

	  } // end if sequence attribute value present

	} // end for sequence attributes

	// go through the event attributes
	// record the maximum end time

	// also count the occurrences of events

	for ( i = 0; i < eventAttributeVector.size(); i++ )
	{
	  // make sure there's a value for this attribute
	  if ( ! instance.isMissing
	       ( (Attribute)eventAttributeVector.elementAt ( i ) ) )
	  {
	    // get the value parsed into events
	    eventArray = 
	      TSUtility.getEvents 
	      ( instance.stringValue
		( (Attribute)eventAttributeVector.elementAt ( i ) ) );

	    // now since we tested the attribute as an event attribute
	    // and for a missing value we will not check to see if the
	    // event array is null. 

	    // add the number of event occurrences to our total count
	    numEventOccurrences += eventArray.length;

	    // record the max end time of all the events
	    for ( int k = 0; k < eventArray.length; k++ )
	    {
	      // get the begin and end time for the event
	      tempEventTime = TSUtility.getValues ( eventArray [ k ] );

	      numTimeUnitsEvent = 
		Math.max ( numTimeUnitsEvent, tempEventTime [ 1 ] );

	      // calculate length of event
	      tempEventLength = tempEventTime [ 1 ] - tempEventTime [ 0 ];

	      // compare with max length
	      maxEventLength = Math.max ( maxEventLength, tempEventLength );

	      // do the same for the per events
	      maxEventLengthArray [ i ] = 
		Math.max ( maxEventLengthArray [ i ], 
			   tempEventLength );

	      // compare with min length
	      // for min we have to catch when the temp event length
	      // is 0.
	      if ( tempEventLength != 0.0 )
	      {
		minEventLength = Math.min ( minEventLength, tempEventLength );

		minEventLengthArray [ i ] = 
		  Math.min ( minEventLengthArray [ i ], 
			     tempEventLength );

	      } // end if temp event length is 0

	    } // end for each event

	    // have to check if we got any events before saying
	    // we found time line info
	    if ( eventArray.length > 0 )
	    {
	      // remember to say we found time line info
	      foundTimeLineInfo = true;

	    } // end if events were found

	  } // end if event attribute value present 

	} // end for event attributes

	// take the maximum value for time line length found
	if ( foundTimeLineInfo )
	{
	  numTimeUnits += Math.max ( Math.max ( numTimeUnitsTimeLine,
						numTimeUnitsSequence ),
				     numTimeUnitsEvent );

	  numInstancesWithTimeLineInfo += 1;

	} // end if found time line info

      } // while more instances

      // log the number of event occurrences
      logger.logValue ( Logger.NUM_EVENT_OCCURRENCES, 
			Integer.toString ( numEventOccurrences ) );

      System.out.println ( "Logging number event occurrences: " + 
			 numEventOccurrences + " ..." );

      // log the average time line length
      // make sure we avoid a devide by zero situation
      if ( numInstancesWithTimeLineInfo == 0 )
      {
	logger.logValue ( Logger.AVERAGE_LENGTH_OF_TIMELINE, "0" );

      } // end if no time lines
      else
      {
	logger.logValue ( Logger.AVERAGE_LENGTH_OF_TIMELINE,
			  Double.toString 
			  ( numTimeUnits / numInstancesWithTimeLineInfo ) );
      } // end else there were time lines

      System.out.println ( "Logging average length of time line: " +
			   logger.getValue 
			   ( Logger.AVERAGE_LENGTH_OF_TIMELINE ) + " ..." );

      // log the min and max event length
      logger.logValue ( Logger.MAX_EVENT_LENGTH, 
			Double.toString ( maxEventLength ) );

      logger.logValue ( Logger.MIN_EVENT_LENGTH, 
			Double.toString ( minEventLength ) );

      System.out.println ( "Logging min and max event lengths. min: "
			   + minEventLength + " max: " + maxEventLength );

      // for the per event attribute min and max we have to do it 
      // using a loop
      String attName;

      System.out.println ( "Logging min and max length for : " );

      for ( i = 0; i < minEventLengthArray.length; i++ )
      {
	attName = ((Attribute)eventAttributeVector.elementAt ( i )).name();

	logger.logValue ( Logger.MAX_PER_EVENT_LENGTH, 
			  attName + "=" + maxEventLengthArray [ i ] );

	// if there were no events of a certain type
	// the min length will be max double which isn't good.
	// we should reset it to 0
	if ( minEventLengthArray [ i ] == Double.MAX_VALUE )
	{
	  minEventLengthArray [ i ] = 0.0;
	} // end if min is max

	logger.logValue ( Logger.MIN_PER_EVENT_LENGTH, 
			  attName + "=" + minEventLengthArray [ i ] );

	System.out.println ( "\t" + attName +
			     "\t\tmin= " + minEventLengthArray [ i ] +
			     "\tmax= " + maxEventLengthArray [ i ] );

      } // end for each min and max event length
      
      System.out.println ( "Done logging data set statistics.\n" );

    } // end if specifying log file name

    long startTime = System.currentTimeMillis();

    // mine associations was named build associations.
    // it was changed so that we may measure the time it 
    // takes to mine association rules from the WEKA GUI.
    // The WEKA system defines an abstract class Associator that
    // defines build associations. All WPI's association mining classes
    // should extend the WPI extention of that class defined in the
    // file you are currently reading.
    mineAssociations ( data );

    long endTime = System.currentTimeMillis();

    long seconds = ( endTime / 1000 ) - ( startTime / 1000 );

    System.out.println ( "\nMining Started : " + startTime );
    System.out.println ( "Mining Complete: " + endTime );
    System.out.println ( "time taken: " + seconds + " seconds\n" );

    if ( logger != null )
    {
      logger.logValue ( Logger.TOTAL_SECONDS_TO_MINE, 
			Long.toString ( seconds ) );

      // let's computer some summary stats
      // like the average time per frequent itemset found

      // do we have the numbe rof frequrnt itemsets found?
      String tempString = logger.getValue 
	( Logger.NUM_FREQUENT_ITEMSETS_FOUND );

      // did we get anything?
      if ( tempString != null )
      {
	// this could be a sequence, use the time sequence utils
	double [] numFrequentItemsetArray = TSUtility.getValues ( tempString );

	// do we have the total time?
	tempString = logger.getValue ( Logger.TOTAL_SECONDS_TO_MINE );

	if ( tempString != null )
	{
	  // this is probably not a sequence but we can use the
	  // same method to get the values
	  double [] totalTimeArray = TSUtility.getValues ( tempString );

	  // tally it up
	  double numItemsetsFound = 0;
	  
	  for ( int i = 0; i < numFrequentItemsetArray.length; i++ )
	  {
	    numItemsetsFound += numFrequentItemsetArray [ i ];

	  } // end for each number of frequent itemset found

	  // might as well log this
	  logger.logValue ( Logger.TOTAL_NUM_FREQUENT_ITEMSETS,
			    Double.toString ( numItemsetsFound ) );

	  // have to check if there's a zero there
	  if ( totalTimeArray [ 0 ] == 0.0 )
	  {
	    // reset it to something really small
	    totalTimeArray [ 0 ] = 0.0001;

	  } // end if total time zero

	  // log it baby
	  logger.logValue ( Logger.SECONDS_PER_FREQUENT_ITEMSET,
			    Double.toString ( totalTimeArray [ 0 ] /
					      numItemsetsFound ) );

	} // end if num frequent itemsets logged 

	// let's see if we can figure out the number of event
	// occurrences per event attribute
	// do we have the number of occurrences?
	tempString = logger.getValue ( Logger.NUM_EVENT_OCCURRENCES );

	if ( tempString != null )
	{
	  double [] numOccurrences = TSUtility.getValues ( tempString );

	  // do we have the number of event attributes?
	  tempString = logger.getValue ( Logger.NUM_EVENT_ATTRIBUTES );

	  if ( tempString != null )
	  {
	    double [] numEventAttributes = TSUtility.getValues ( tempString );

	    logger.logValue ( Logger.NUM_OCCURRENCES_PER_ATTRIBUTE,
			      Double.toString ( numOccurrences [ 0 ] /
						numEventAttributes [ 0 ] ) );
	  } // end if num event attributes logged

	} // end if num occurrences logged
	
      } // end if we have counts of frequent itemsets

      // all stats that should be logged have to be logged 
      // before this call
      logger.saveLog();
      logger.clear();

    } // end if we are logging stuff

  } // END public void buildAssociations ( Instances data ) throws Exception

//********************************************************//

} // END public abstract class Associator 
  //       extends weka.associations.Associator
