/*
ARMiner - Association Rules Miner
Copyright (C) 2000  UMass/Boston - Computer Science Department

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The ARMiner Server was written by Dana Cristofor and Laurentiu
Cristofor.

The ARMiner Client was written by Abdelmajid Karatihy, Xiaoyong Kuang,
and Lung-Tsung Li.

The ARMiner package is currently maintained by Laurentiu Cristofor
(laur@cs.umb.edu).
*/

package wpi.associations.arminerSequence;

import java.util.*;

// stuff for dealing with time sequence stuff
import wpi.timesequence.*;

/*
  Maintenance log started on November 30th, 2000

  Nov. 30th, 2000   - added pruneDuplicates method
                    - added doesIntersect method

  Feb. 14th, 2002   - added hasAttributes -zss

  Apr. 16th, 2002   - added support for time sequence 
                      stuff and documentation and cleaned
		      up formatting -kap

  Apr. 23rd, 2002   - added debug info levels -kap

  June 28, 2002     - modified doesIntersect to support event items -kap

  January 17, 2003  - modified checkRelativeTime to return a list
                      of maps of all possible matches instead of
		      just one. this is so we can count the number
		      of times an itemset, or rather, a list of time 
		      sequence event items, is found. -kap

  Febraury 22, 2003 - renamed data member index to iterateIndex to clear
                      ambiguity between the common loop counter name.
		    - added method startItemIteration() cause there was 
		      no way to ensure iteration over the items without
		      returning the first item and moving to the next 
		      item -kap

  February 24, 2003 - made consistent the convention of using a lower case
		      letter to start a method name. -kap

  August 13, 2003   - corrected a check condition in canCombineWith.
                      when comparing the last items in the this and the
		      other itemset it is only necessary to make sure that
		      they are not the exact same item. they can be the 
		      same type -kap

  November 13, 2003 - corrected checking for regular items in isIncludedIn.
                      see method for details. -kap

  December 2, 2003  - added buildItemset method. -kap
                    - added countMapExtentions, findAllMaps (6), and
		      isMapValid methods for new weight counting approach -kap

  January 1, 2004   - renamed isIncludedIn to numIncludedIn. this was done so
                      that the old code for counting the number of inclusions
		      of an itemset in an instance based on events could be 
		      kept for reference. isIncludedIn will be modified to
		      only identify if an itemset is included in another and
		      not how many times. -kap

  January 3, 2004   - changed the name of checkRelativeTime to 
                      mapRelativeTime to more accurately represent its 
		      function -kap
		    - added a new checkRelativeTime method that just checks
		      if an itemset contains a set of event items with
		      the same relative temporal relationships -kap
*/

/*  
   This file is a part of the ARMiner project.
   
   (P)1999-2000 by ARMiner Server Team:

   Dana Cristofor
   Laurentiu Cristofor

   modified by Zack Stoecker-Sylvia 2002
   modified by Keith A. Pray 2002
*/

/**
   ARMinerItemset.java<P>

   An item set is an ordered list of integers that identify items
   coupled with a float value representing the support of the item set
   as a percentage.<P>

@author Keith A. Pray (kap@wpi.edu)
*/

public class ARMinerItemset implements java.io.Serializable
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  // let's set this to 2 instead of seven
  //  private static final int SIZE_INCR = 7;
  private static final int SIZE_INCR = 2;

  /** The capacity of the item set. @serial */
  private int capacity;

  /** The number of items in the item set. @serial */
  private int size;

  /** The item set. @serial */
  private int[] set;

  /** The support of the item set. @serial */
  private float support;

  /** The weight of the item set. @serial */
  private long weight;

  /** The mark of the item set.
   * This can be used to mark the item set for various purposes. @serial */
  private boolean mark; 

  /** Internal index used for cycling through the item set's items. @serial */
  private int iterateIndex;

  private int sum = 0;

  // ----- Stuff needed for time sequence attribute support. ----- //

  /* eventWeight is being added so we can track the number of times
     a set of events are found in the dataset. this can be many more
     than the actual number of instances in a dataset. -kap */

  /** The weight of the item set's events. @serial */
  private long eventWeight;

  /** Used to look up actual value of an item using its 
      integer representation. */
  private Hashtable numberHash;

  /** The name of time sequence event attributes */
  private Hashtable nameHash;

  /** Used to look up the real begin and end times of
   * the events in this item set. */
  private Hashtable realHash;

  /** Used to look up the relative begin and end times of
   * the events in this item set. */
  private Hashtable relativeHash;

  /** The relative time labels and corresponding real time
   * values */
  private Vector relativeVector;

  /** A vector of vectors. Each vector contains a list of
   * the time sequence event item labels corresponding to
   * the relative time as denoted by eventVector's index.
   * Used for relabeling when changes occur. */
  private Vector eventVector;

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: output of methods
   * 3: all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Sets the number hash to use for looking up values
   * in this item set
   * This method is needed to support time sequence attributes.
   *
   * @param h
   * the number hash
   */

  public void setNumberHash ( Hashtable h )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "Set number hash: size = " + h.size() );
    } // end switch debug -kap

    numberHash = h;

  } // END public void setNumberHash ( Hashtable h )

//********************************************************//

  /**
   * Returns the number hash used for looking up values
   * in this item set.
   * This method is needed to support time sequence attributes.
   *
   * @return the number hash
   */

  public Hashtable getNumberHash()
  {
    return ( numberHash );

  } // END public Hashtable getNumberHash()

//********************************************************//

  /**
   * Returns the begin item label used for looking up
   * begin times for time sequence events in this
   * item set.
   * This method is needed to support time sequence attributes.
   *
   * @param item
   * the integer representation of an item
   *
   * @return the begin item label
   */

  public static String beginLabel ( int item )
  {
    return ( item + "b" );

  } // END public static String beginLabel ( int item )

//********************************************************//

  /**
   * Returns the begin item label used for looking up
   * begin times for time sequence events in this
   * item set.
   * This method is needed to support time sequence attributes.
   *
   * @param item
   * the integer representation of an item
   *
   * @return the begin item label
   */

  public static String beginLabel ( Integer item )
  {
    return ( item + "b" );

  } // END public static String beginLabel ( Integer item )

//********************************************************//

  /**
   * Returns the end item label used for looking up end times
   * for time sequence events in this item set
   * This method is needed to support time sequence attributes.
   *
   * @param item
   * the integer representation of an item
   *
   * @return the end item label
   */

  public static String endLabel ( int item )
  {
    return ( item + "e" );

  } // END public static endLabel ( int item )

//********************************************************//

  /**
   * Returns the end item label used for looking up end times
   * for time sequence events in this item set
   * This method is needed to support time sequence attributes.
   *
   * @param item
   * the integer representation of an item
   *
   * @return the end item label
   */

  public static String endLabel ( Integer item )
  {
    return ( item + "e" );

  } // END public static endLabel ( Integer item )

//********************************************************//

  /**
   * Returns the item integer used to represent items
   * based on the begin or end label for the item.
   * This method is needed to support time sequence attributes.
   *
   * @param label
   * the begin or end label of an item
   *
   * @return the item integer
   */

  public static int itemInt ( String label )
  {
    // just return the int value of stripping the last character
    return ( Integer.parseInt ( label.substring ( 0, label.length() - 1 ) ) );

  } // END public static int itemInt ( String label )

//********************************************************//

  /**
   * Constructs and returns a string representing the
   * value of an item. If the item is an event and 
   * contained in this itemset, the value will be specific
   * to this itemset. The relative times of the event will
   * be represented.
   *
   * @return a string value of the item specified
   */

  public String itemValue ( int item )
  {
    Integer itemInt = new Integer ( item );

    // if the event is not an item (or not in this
    // itemset) just return the value from number hash
    if ( nameHash.get ( itemInt ) == null )
    {
      return ( (String)numberHash.get ( itemInt ) );

    } // end if item not in this itemset and/or not an event

    // guess its an event

    // need to build the string to return, start with the name
    StringBuffer sb = 
      new StringBuffer ( (String)nameHash.get ( itemInt ) );

    // add the equal sign thingy
    sb.append ( "=" );

    // and the begin time
    sb.append ( relativeHash.get ( beginLabel ( item ) ) );

    // and the sequence delimiter
    sb.append ( TSUtility.TIME_SEQUENCE_DELIMITER );

    // and the end time of course
    sb.append ( relativeHash.get ( endLabel ( item ) ) );

    // all done
    return ( sb.toString() );

  } // END public String itemValue ( int item )
   
//********************************************************//

  /**
   * Tests if the string is an item begin label.
   * This method is needed to support time sequence attributes.
   *
   * @param label
   * the label to test
   *
   * @return true if it's a begin label
   */

  public static boolean isBeginLabel ( String label )
  {
    return ( label.endsWith ( "b" ) );

  } // END public static boolean isBeginLabel ( String label )

//********************************************************//

  /**
   * Tests if the string is an item end label.
   * This method is needed to support time sequence attributes.
   *
   * @param label
   * the label to test
   *
   * @return true if it's a end label
   */

  public static boolean isEndLabel ( String label )
  {
    return ( label.endsWith ( "e" ) );

  } // END public static boolean isEndLabel ( String label )

//********************************************************//

  /**
   * Returns the name of an event. The item specified
   * must be in this itemset and an event or no name will
   * be found.
   *
   * @param item
   * the item to get the name of
   *
   * @return the event name, null if not found
   */

  public String eventName ( Integer item )
  {
    return ( (String)nameHash.get ( item ) );

  } // END public String eventName ( Integer item )

//********************************************************//

  /**
   * Returns the name of an event. The item specified
   * must be in this itemset and an event or no name will
   * be found.
   *
   * @param item
   * the item to get the name of
   *
   * @return the event name, null if not found
   */

  public String eventName ( int item )
  {
    return ( eventName ( new Integer ( item ) ) );

  } // END public String eventName ( int item )

//********************************************************//
  
  /**
   * Returns the relative begin time of the item
   * specified. The item specified must be in this itemset
   * and an event or no value will be returned
   *
   * @param item
   * the item
   *
   * @return the relative begin time
   */

  public Integer relativeBegin ( Integer item )
  {
    return ( (Integer)relativeHash.get ( beginLabel ( item ) ) );

  } // END public Integer relativeBegin ( Integer item )

//********************************************************//
  
  /**
   * Returns the relative end time of the item
   * specified. The item specified must be in this itemset
   * and an event or no value will be returned
   *
   * @param item
   * the item
   *
   * @return the relative end time
   */

  public Integer relativeEnd ( Integer item )
  {
    return ( (Integer)relativeHash.get ( endLabel ( item ) ) );

  } // END public Integer relativeEnd ( Integer item )

//********************************************************//

  /**
   * Returns a copy of the name hash used for looking up
   * attribute names for time sequence events in this item
   * set.
   * This method is needed to support time sequence attributes.
   *
   * @return the name hash
   */

  public Hashtable getNameHash()
  {
    return ( new Hashtable ( nameHash ) );

  } // END public Hashtable getNameHash()

//********************************************************//

  /**
   * Returns a copy of the real hash used for looking up the
   * real times for time sequence events in this item set
   * This method is needed to support time sequence attributes.
   *
   * @return the realtive hash
   */

  public Hashtable getRealHash()
  {
    return ( new Hashtable ( realHash ) );

  } // END public Hashtable getRealHash()

//********************************************************//
  
  /**
   * Returns the number of event items contained in this
   * itemset.
   *
   * @return the number of event items
   */

  public int numEventItems()
  {
    // we have 2 real times for each event item
    return ( realHash.size() / 2 );

  } // END public int numEventItems()

//********************************************************//

  /**
   * Returns a copy of the relative hash used for looking
   * up the relative times for time sequence events in this
   * item set.
   * This method is needed to support time sequence attributes.
   *
   * @return the realtive hash
   */

  public Hashtable getRelativeHash()
  {
    return ( new Hashtable ( relativeHash ) );

  } // END public Hashtable getRelativeHash()

//********************************************************//

  /**
   * Returns a copy of the realtive vector used for
   * looking up the real times associated with relative
   * times for time sequence events in this item set.
   * This method is needed to support time sequence attributes.
   *
   * @return the realtive vector
   */

  public Vector getRelativeVector()
  {
    return ( new Vector ( relativeVector ) );

  } // END public Vector getRelativeVector()

//********************************************************//

  /**
   * Returns a copy of the event vector used for looking
   * up the time sequence event begin and end labels
   * associated with a relative time.
   * This method is needed to support time sequence attributes.
   *
   * @return the event vector
   */

  public Vector getEventVector()
  {
    // this is special since it contains vectors
    // go through and copy each vector
    Vector ev = new Vector ( eventVector.size() );

    int index;

    for ( index = 0; index < eventVector.size(); index++ )
    {
      ev.add ( new Vector ( (Vector)eventVector.get ( index ) ) );
    }

    return ( ev );

  } // END public Vector getEventVector()

//********************************************************//

  /**
   * Creates a new empty item set.
   */

  public ARMinerItemset()
  {
    switch ( debug ) {
    case 3:
      System.out.println ( "Making new ARMinerItemset, empty." );
    } // end switch debug -kap

    capacity = SIZE_INCR;
    set = new int [ capacity ];
    size = 0;
    support = 0;
    weight = 0;
    eventWeight = 0;
    mark = false;
    iterateIndex = 0;

    // Initialize stuff for time sequence support
    nameHash = new Hashtable ( capacity );
    realHash = new Hashtable ( capacity );
    relativeHash = new Hashtable ( capacity );
    relativeVector = new Vector ( capacity );
    eventVector = new Vector ( capacity );

  } // END public ARMinerItemset()

//********************************************************//

  /**
   * Create a new empty item set of specified capacity.
   *
   * @param c  
   * the capacity of the item set
   *
   * @exception IllegalArgumentException  
   * <code>c</code> is negative or zero
   */

  public ARMinerItemset ( int c )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "Making new ARMinerItemset with size " + c + "." );
    } // end switch debug -kap

    if ( c < 1 )
      throw new IllegalArgumentException
	( "ARMinerItemset constructor requires positive argument value." );

    capacity = c;
    set = new int[capacity];
    size = 0;
    support = 0;
    weight = 0;
    eventWeight = 0;
    mark = false;
    iterateIndex = 0;

    // Initialize stuff for time sequence support
    nameHash = new Hashtable ( capacity );
    realHash = new Hashtable ( capacity );
    relativeHash = new Hashtable ( capacity );
    relativeVector = new Vector ( capacity );
    eventVector = new Vector ( capacity );

  } // END public ARMinerItemset ( int c )

//********************************************************//

  /**
   * Create a new item set by copying a given one.
   *
   * @param itemset  
   * the itemset to be copied
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null
   */

  public ARMinerItemset ( ARMinerItemset itemset )
  {
    if ( itemset == null )
    {
      throw new IllegalArgumentException 
	( "ARMinerItemset ( itemset ) constructor: itemset null." );
    } // end if itemset null

    capacity = itemset.capacity;
    set = new int[capacity];
    size = itemset.size;
    support = itemset.support;
    weight = itemset.weight;
    eventWeight = itemset.eventWeight;
    mark = itemset.mark;
    sum = itemset.sum;

    iterateIndex = 0;

    // Initialize stuff for time sequence support
    // be sure to use the access methods so we don't
    // have to remember to make copies of stuff that
    // we need to make copies of
    numberHash = itemset.getNumberHash();
    nameHash = itemset.getNameHash();
    relativeHash = itemset.getRelativeHash();
    realHash = itemset.getRealHash();
    relativeVector = itemset.getRelativeVector();
    eventVector = itemset.getEventVector();
    
    // copy items
    System.arraycopy ( itemset.set, 0, set, 0, size );

    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "ARMinerItemset constructor: " +
			   "Making copy of " + itemset + " " +
			   "their events = " + itemset.eventVector + " " +
			   "our events = " + eventVector );
    } // end switch debug -kap

  } // END public ARMinerItemset ( ARMinerItemset itemset )

//********************************************************//

  // This could/should be made into a constructor -kap 2003-12-02

  /**
   * Builds an ARMinerItemset based on the specified
   * itemset indexes.
   *
   * @param itemset
   * an itemset from which to get items
   *
   * @param indexes
   * the indexes
   */

  static public ARMinerItemset buildItemsetFromIndexes 
    ( ARMinerItemset itemset, int [] indexes )
  {
    ARMinerItemset result;
    int i;

    switch ( debug ) {
    case 3: case 2: case 1: //case 0:
      System.out.print ( "ARMinerItemset.buildItemsetFromIndexes: " +
			 "itemset: " + itemset +
			 " indexes = " );
      for ( i = 0; i < indexes.length; i++ ) {
	System.out.print ( indexes [ i ] + " " );
      } // end for each index
      System.out.println ( "" );
    } // end switch debug -kap

    result = new ARMinerItemset ( indexes.length );
    result.setNumberHash ( itemset.getNumberHash() );
    
    // now add the items
    for ( i = 0; i < indexes.length; i++ )
    {
      result.addItem ( itemset.getItem ( indexes [ i ] ), itemset );
      
    } // end for each index

    // all done
    return ( result );

  } // END private ARMinerItemset buildItemsetFromIndexes
  // ( ARMinerItemset itemset, int [] indexes )

//********************************************************//

  // This could/should be made into a constructor -kap 2003-12-02

  /**
   * Builds an ARMinerItemset based on the specified
   * itemset item list.
   *
   * @param itemset
   * an itemset from which to get items
   *
   * @param items
   * the items
   */

  static public ARMinerItemset buildItemsetFromItemList 
    ( ARMinerItemset itemset, int [] items )
  {
    ARMinerItemset result;
    int i;

    result = new ARMinerItemset ( items.length );
    result.setNumberHash ( itemset.getNumberHash() );
    
    // now add the items
    for ( i = 0; i < items.length; i++ )
    {
      result.addItem ( items [ i ], itemset );
      
    } // end for each index

    // all done
    return ( result );

  } // END private ARMinerItemset buildItemsetFromIndexes
  // ( ARMinerItemset itemset, int [] items )

//********************************************************//


  // This could/should be made into a constructor -kap 2003-12-02

  /**
   * Builds an ARMinerItemset based on the specified
   * itemset item list.
   *
   * @param itemset
   * an itemset from which to get items
   *
   * @param items
   * the items
   */

  static public ARMinerItemset buildItemsetFromItemList 
    ( ARMinerItemset itemset, Integer [] items )
  {
    ARMinerItemset result;
    int i;

    result = new ARMinerItemset ( items.length );
    result.setNumberHash ( itemset.getNumberHash() );
    
    // now add the items
    for ( i = 0; i < items.length; i++ )
    {
      result.addItem ( items [ i ].intValue(), itemset );
      
    } // end for each index

    // all done
    return ( result );

  } // END private ARMinerItemset buildItemsetFromIndexes
  // ( ARMinerItemset itemset, Integer [] items )

//********************************************************//

  /**
   * Return support of itemset.
   */

  public float getSupport()
  {
    return ( support );

  } // END public float getSupport()

//********************************************************//

  /**
   * Return weight of itemset.
   */

  public long getWeight()
  {
    return ( weight );

  } // END public long getWeight()

//********************************************************//

  /**
   * Return event weight of itemset.
   */

  public long getEventWeight()
  {
    return ( eventWeight );

  } // END public long getEventWeight()

//********************************************************//

  /**
   * Return i-th item in set.
   *
   * @param i  
   * the index of the item to get
   *
   * @exception IndexOutOfBoundsException  
   * <code>i</code> is an invalid index
   *
   * @return the <code>i</code>-th item
   */

  public int getItem ( int i )
  {
    if ( i < 0 || i >= size )
    {
      throw new IndexOutOfBoundsException 
	( "ARMinerItemset.getItem ( i ) invalid index" );

    } // end if invalid index

    return ( set [ i ] );

  } // END public int getItem ( int i )

//********************************************************//

  /**
   * Sets the iteration to the first item.
   */

  public void startItemIteration()
  {
    iterateIndex = 0;

  } // END public void startItemIteration()

//********************************************************//

  /**
   * Return first item in set.
   *
   * @exception IndexOutOfBoundsException  
   * there is no first item
   *
   * @return first item
   */

  public int getFirstItem()
  {
    iterateIndex = 0;

    if ( iterateIndex < 0 || iterateIndex >= size )
    {
      throw new IndexOutOfBoundsException 
	( "ARMinerItemset.getFirstItem() no first item" );
    }

    return ( set [ iterateIndex++ ] );

  } // END public int getFirstItem()

//********************************************************//

  /**
   * Return next item in set.
   *
   * @exception IndexOutOfBoundsException  
   * there is no next item
   *
   * @return next item
   */

  public int getNextItem()
  {
    if ( iterateIndex < 0 || iterateIndex >= size )
    {
      throw new IndexOutOfBoundsException
	( "ARMinerItemset.getNextItem() no next item" );
    }

    return ( set [ iterateIndex++ ] );

  } // END public int getNextItem()

//********************************************************//

  /**
   * Return true if there are more items in the itemset. You can call
   * this method to find out whether you can call getNext without
   * raising an exception.
   *
   * @return true if there are more items, false if not
   */

  public boolean hasMoreItems() 
  { 
    if ( iterateIndex < 0 || iterateIndex >= size ) 
    {
      return ( false ); 
    }
    else 
    {
      return ( true ); 
    }
  } // END public boolean hasMoreItems() 

//********************************************************//
  
  /**
   * Return size of itemset.
   *
   * @return size of itemset
   */

  public int size()
  {
    return ( size );

  } // END public int size()

//********************************************************//

  /**
   * Return true if this itemset has the set of attributes
   * with <code>itemset</code>.
   *
   * @param itemset
   * a vector of item sets where only the attribute-item pairs
   * for the attribute exist
   *
   * @exception IllegalArgumentException
   * <code>itemset</code> is null
   *
   * @return
   * true if <code>itemset</code> contains items of this 
   * itemset, false otherwise.
   */

  public boolean hasAttributes ( Vector attributes )
  {
    boolean has = true;
    
    for ( int i = 0; i < attributes.size() && has; i++ )
    {
      has = has && doesIntersect ( (ARMinerItemset)attributes.
				   elementAt ( i ) );
    }
    
    return ( has );

  } // END public boolean hasAttributes ( Vector attributes )

//********************************************************//

  /**
   * Return true if this itemset has items in common
   * with <code>itemset</code>.
   *
   * @param itemset  
   * the itemset with which we compare
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null
   *
   * @return  
   * true if <code>itemset</code> contains items of this itemset,
   * false otherwise.
   */

  public boolean doesIntersect ( ARMinerItemset itemset )
  {
    if ( itemset == null )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.doesIntersect ( itemset ) itemset is null" );
    }

    int i = 0;
    int j = 0;

    for ( ; i < size && j < itemset.size; )
    {
      // if elements are equal, return true
      if ( set [ i ] == itemset.set [ j ] )
      {
	return ( true );
      }
      // if the element in this ARMinerItemset is bigger then
      // we need to move to the next item in itemset.
      else
      {
	// we should now check if the item is an event -kap
	String name = itemset.eventName ( itemset.set [ j ] );
	if ( name != null )
	{
	  // do we have one of these?
	  if ( nameHash.contains ( name ) )
	  {
	    // then we intersect, yay!
	    return ( true );

	  } // end if we have an event item with the same name

	} // end if item is an event

	if ( set [ i ] > itemset.set [ j ] )
	{
	  j++;
	}
	// the element in this ARMinerItemset does not appear
	// in item set so we need to add it to result
	else
	{
	  i++;
	}
      } // end else

    } // end for
    
    return ( false );

  } // END public boolean doesIntersect ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Checks if the item set contains the specified item.
   *
   * @param item
   * the item to check for
   *
   * @return true is the item set contains the item,
   * false otherwise
   */

  public boolean contains ( int item )
  {
    return ( Arrays.binarySearch ( set, item ) > -1 );

  } // END public boolean contains ( int item )

//********************************************************//

  /**
   * Return a new ARMinerItemset that contains only those
   * items that do not appear in <code>itemset</code>.
   *
   * @param itemset  
   * the itemset whose items we want to subtract
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null
   *
   * @return
   * an ARMinerItemset containing only those items of this
   * ARMinerItemset that do not appear in <code>itemset</code>.
   */

  public ARMinerItemset subtract ( ARMinerItemset itemset )
  {
    if ( itemset == null )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.subtract ( itemset ) itemset null." );
    }

    ARMinerItemset result = new ARMinerItemset ( capacity );

    // set the number hash to support time sequence attributes -kap
    result.setNumberHash ( getNumberHash() );
    
    int i = 0;
    int j = 0;

    for ( ; i < size && j < itemset.size; ) 
    {
      // if elements are equal, move to next ones
      if ( set [ i ] == itemset.set [ j ] ) 
      {
	i++;
	j++;
      } 
      else if ( set [ i ] > itemset.set [ j ] )
	// if the element in this ARMinerItemset is bigger then
	// we need to move to the next item in itemset.
	j++;

      // the element in this ARMinerItemset does not appear
      // in itemset so we need to add it to result
      else
      {
	// just add it using the addItem method so the time
	// sequence stuff is accounted for... plus it is more
	// object oriented and stuff... albeit a bit slower -kap
	//	result.sum += set [ i ];
	//	result.set [ result.size++ ] = set [ i++ ];

	// actually, let's use the addItem method that uses an itemset
	// to short cut much of the parsing of item value and stuff
	//	result.addItem ( set [ i++ ] );

	result.addItem ( set [ i ], this );
	i++;
      }
    }
    
    // copy any remaining items from this ARMinerItemset 
    while ( i < size ) 
    {
      // just add it using the addItem method so the time
      // sequence stuff is accounted for... plus it is more
      // object oriented and stuff... albeit a bit slower -kap
      //      result.sum += set[i];
      //      result.set[result.size++] = set[i++];

	// actually, let's use the addItem method that uses an itemset
	// to short cut much of the parsing of item value and stuff
	//	result.addItem ( set[i++] );
	result.addItem ( set[i], this );
	i++;
    }

    // NOTE: the size of the resulting itemset
    // has been automatically updated
    return ( result );
    
  } // END public ARMinerItemset subtract ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Return a new ARMinerItemset that contains all those
   * items that appear in this ARMinerItemset and in
   * <code>itemset</code>.
   *
   * @param itemset
   * the itemset whose items we want to add
   *
   * @exception IllegalArgumentException
   * <code>itemset</code> is null
   *
   * @return
   * an ARMinerItemset containing all those items that appear
   * in this ARMinerItemset and in <code>itemset</code>.
   */

  public ARMinerItemset add ( ARMinerItemset itemset )
  {
    if ( itemset == null )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.add ( itemset ) itemset null" );
    }

    ARMinerItemset result = new ARMinerItemset ( capacity );

    // set time sequence stuff
    // just need number hash cause we'll use the addItem method -kap
    result.setNumberHash ( getNumberHash() );

    int i = 0;
    int j = 0;

    for ( ; i < size && j < itemset.size; ) 
    {
      // the following changed to use the
      // simpler although less efficient addItem method
      // this will ensure time sequence stuff is taken care of -kap

      // let addItem track the sum
      //      result.sum = sum + itemset.sum;  // double-counts equal elements

      // if elements are equal, copy then move to next ones
      if ( set [ i ] == itemset.set [ j ] ) 
      {
	//	result.sum -= set [ i ];  // because it was counted twice
	//	result.set [ result.size++ ] = set [ i++ ];
	//	result.addItem ( set [ i++ ] );
	// actually, let's use the addRegularItem and addeventItem methods
	// that short cut much of the parsing of item value and stuff
	result.addItem ( set [ i ], this );
	i++;
	j++;
      }

      // if the element in this ARMinerItemset is bigger then
      // we need to copy from itemset then move to the next item.
      else if ( set [ i ] > itemset.set [ j ] )
      {
	//	result.set [ result.size++ ] = itemset.set [ j++ ];
	//	result.addItem ( itemset.set [ j++ ] );
	result.addItem ( itemset.set [ j++ ], itemset );
      }
      // else we need to copy from this ARMinerItemset
      else
      {
	//	result.set [ result.size++ ] = set [ i++ ];
	//	result.addItem ( set [ i++ ] );
	result.addItem ( set [ i++ ], this );
      }
    }

    // copy any remaining items from this ARMinerItemset 
    while ( i < size )
    {
      //      result.set [ result.size++ ] = set [ i++ ];
      //      result.addItem ( set [ i++ ] );
      result.addItem ( set [ i++ ], this );
    }

    // copy any remaining items from itemset 
    while ( j < itemset.size )
    {
      //      result.set [ result.size++ ] = set [ j++ ];
      //      result.addItem ( itemset.set [ j++ ] );
      result.addItem ( itemset.set [ j++ ], itemset );
    }

    // NOTE: the size of the resulting itemset
    // has been automatically updated
    return ( result );

  } // END public ARMinerItemset add ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Adds a new item to the itemset using the information
   * provided in the specified itemset rather than the number
   * hash.
   *
   * @param item
   * the item to be added
   *
   * @param itemset
   * the itemset to get item info from
   *
   * @exception IllegalArgumentException   <code>item</code> is <= 0
   *
   * @return   true if item was added, false if it wasn't added (was
   * already there!)
   */

  protected boolean addItem ( int item, 
			    ARMinerItemset itemset )
  {
    Integer itemInt = new Integer ( item );

    // if it's an event, add the event info
    if ( itemset.nameHash.get ( itemInt ) != null )
    {
      addEventItem ( item, 
		     (String)itemset.nameHash.get ( itemInt ),
		     (Double)itemset.realHash.get ( beginLabel ( item ) ),
		     (Double)itemset.realHash.get ( endLabel ( item ) ) );
    } // end if item is an event

    // add the regular item stuff and return if that was
    // successful
    return ( addRegularItem ( item ) );

  } // END private boolean addItem ( int item, 
  //			    ARMinerItemset itemset )

//********************************************************//

  /**
   * Adds a new item to the itemset while not taking care
   * of any time sequence event issues. If the item is an
   * event addEventItem should also be called with this item.
   *
   * @param item   the item to be added
   * @exception IllegalArgumentException   <code>item</code> is <= 0
   * @return   true if item was added, false if it wasn't added (was
   * already there!)
   */

  protected boolean addRegularItem ( int item )
  {
    if ( item <= 0 )
    {
      throw new IllegalArgumentException 
	( "ARMinerItemset.addItem ( " + item +
	  " ) negative value for item not allowed" );
    }

    if ( size == 0 )
    {
      set [ 0 ] = item;
    }
    else 
    {
      // look for place to insert item
      int index;
      for ( index = 0; index < size && item > set [ index ]; index++ )
	;
      
      // if item was already in itemset then return
      if ( index < size && item == set [ index ] )
      {
	return ( false );
      }

      // if set is full then allocate new array
      if ( size == capacity ) 
      {
	capacity = size + SIZE_INCR;
	int[] a = new int [ capacity ];
	
	System.arraycopy ( set, 0, a, 0, index );
	
	a [ index ] = item;
	System.arraycopy ( set, index, a, index + 1, size - index );
	
	set = a;
      } 
      else 
      {
	// otherwise make place and insert new item
	System.arraycopy ( set, index, set, index+1, size - index );
	
	set [ index ] = item;
      }

    } // end else we have to insert the item
    
    // update size
    size++;
    sum += item;
    
    return ( true );

  } // END protected boolean addRegularItem ( int item )

//********************************************************//

  /**
   * Add a new item to the itemset.
   *
   * @param item   the item to be added
   * @exception IllegalArgumentException   <code>item</code> is <= 0
   * @return   true if item was added, false if it wasn't added (was
   * already there!)
   */

  public boolean addItem ( int item )
  {
    // try adding the regular stuff
    if ( addRegularItem ( item ) )
    {
      // now we can add the time sequence stuff
      // please note that this debug switch cannot be simply removed
      // since the method call to add an event item is contained within
      switch ( debug ) {
      case 3: case 2:
	System.out.println ( "ARMinerItemset.addItem: " +
			     "added item " + item + " = " +
			     numberHash.get ( new Integer ( item ) ) + ", " +
			     "is it an event? " + 
			     addEventItem ( item ) );
	break;
      default:
	addEventItem ( item );
	break;
      } // end switch debug -kap

      return ( true );

    } // end if adding the regular item stuff worked

    // I guess it didn't work
    return ( false );

  } // END public boolean addItem ( int item )

//********************************************************//

  /**
   * Add a new event item information to the item set.
   * This method is needed to support time sequence attributes.
   *
   * @param item 
   * the item to be added
   *
   * @exception IllegalArgumentException  
   * <code>item</code> is <= 0
   *
   * @return  
   * true if item was added, false if it wasn't added 
   * (was not an event or we didn't have a number hash.)
   */

  private boolean addEventItem ( int item )
  {
    // now we set the name, begin and end times if it's a time
    // sequence event
    
      // make sure we have a number hash
    if ( numberHash == null )
    {
      System.err.println ( "ARMinerItemset.addEventItem: " +
			   "itemset " + this + " " +
			   "does NOT have a number hash. " +
			   "Time sequence attributes will not be supported. " +
			   "This is not good, no no no." );
      return ( false );

    } // end if we don't have a number hash

    // get the value of this item
    Object o = numberHash.get ( new Integer ( item ) );
    
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "ARMinerItemset.addEventItem: " +
			   item + ": Object = " + o );
    } // end switch debug -kap
    
    // now we have to separate the value from the attribute
    // value pair
    StringTokenizer st = new StringTokenizer ( (String)o, "=" );
    
    String eventName = "";
    
    if ( st.hasMoreTokens() )
    {
      // keep first token, if this turns out to be
      // a time sequence event we will want to put this
      // in the name hash
      eventName = st.nextToken();
    }
    
    String s ="";
    
    if ( st.hasMoreTokens() )
    {
      // get the token after =
      s = st.nextToken();
    }
    
    // is it a time sequence event?
    if ( TSUtility.isTimeSequenceEvent ( s ) )
    {
      // first check if we already have these labels
      if ( relativeHash.get ( beginLabel ( item ) ) != null && 
	   relativeHash.get ( endLabel ( item ) ) != null )
      {
	// then something odd may be going on, print out message
	System.err.println ( "ARMinerItemset.addEventTime: " +
			     "something may be amiss. Labels " +
			     beginLabel ( item ) + " and " + 
			     endLabel ( item ) + " " +
			     "are already in this itemset." );
	return ( false );
	
      } // end if we already have these labels

      // then get the values
      double [] v = TSUtility.getValues ( s );
      
      addEventItem ( item, eventName, v [ 0 ], v [ 1 ] );

      // it worked, yay!
      return ( true );
      
    } // end if it's a time sequence event
    
    // I guess it wasn't an event
    return ( false );
    
  } // END private boolean addEventItem ( int item )

//********************************************************//

  /**
   * Add new event item information to the item set with
   * the specified name, begin and end times.
   * 
   * This method is needed to support time sequence attributes.
   *
   * @param item 
   * the item to be added
   *
   * @param name
   * the name of the event item
   *
   * @param begin
   * the event begin time
   *
   * @param end
   * the event end time
   *
   * @exception IllegalArgumentException  
   * <code>item</code> is <= 0
   *
   * @return  
   * true if item was added, false if it wasn't added 
   * (was not an event or we didn't have a number hash.)
   */

  protected void addEventItem ( int item, String name, 
				double begin, double end )
  {
    addEventItem ( item, name, new Double ( begin ),
		   new Double ( end ) );

  } // END protected void addEventItem ( int item, String name, 
  //				double begin, double end )

//********************************************************//

  /**
   * Add new event item information to the item set with
   * the specified name, begin and end times.
   * 
   * This method is needed to support time sequence attributes.
   *
   * @param item 
   * the item to be added
   *
   * @param name
   * the name of the event item
   *
   * @param begin
   * the event begin time
   *
   * @param end
   * the event end time
   *
   * @exception IllegalArgumentException  
   * <code>item</code> is <= 0
   *
   * @return  
   * true if item was added, false if it wasn't added 
   * (was not an event or we didn't have a number hash.)
   */

  protected void addEventItem ( int item, String name, 
				Double begin, Double end )
  {
    // make sure none of these are null -kap 2004-04-02
    if ( name == null )
    {
      System.err.println ( "ARMinerItemsets.addEventItem: name null." );
      return;
    } // end if item null
    if ( begin == null )
    {
      System.err.println ( "ARMinerItemsets.addEventItem: begin null." );
      return;
    } // end if item null
    if ( end == null )
    {
      System.err.println ( "ARMinerItemsets.addEventItem: end null." );
      return;
    } // end if item null

    // get begin and end labels to avoid multiple calls
    String ib = beginLabel ( item );
    String ie = endLabel ( item );
    
    // add the times
    realHash.put ( ib, begin );
    
    // the second is the end time
    realHash.put ( ie, end );
    
    // don't forget to save the name
    // (storing extra keys for the same value costs little memory and
    //  avoids a lot of processing)
    nameHash.put ( new Integer ( item ), name );
    nameHash.put ( ib, name );
    nameHash.put ( ie, name );
    
    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "ARMinerItemset.addEventItem: " +
			   "Added event: " +
			   nameHash.get ( ib ) +
			   ": begin: " +
			   realHash.get ( ib ) +
			   " end: " +
			   realHash.get ( ie ) );
    } // end switch debug -kap
    
      // now we have to add the relative time
    addRelativeTime ( item );

  } // END protected void addEventItem ( int item, String name, 
  //				 double begin, double end )

//********************************************************//

  /** 
   * Adds the relative begin and end times of an item to
   * this item set.
   * This method is needed to support time sequence attributes.
   *
   * @param item
   */

  private void addRelativeTime ( int item )
  {
    // get the begin and end labels for this item
    String ib = beginLabel ( item );
    String ie = endLabel ( item );

    // keep track of where we have to relabel things from
    // if we have to that is

    // did the size of the relative vector change (so we have to relabel)
    int oldSize = relativeVector.size();
    int relabel = 0;

    // start at relative time 0
    int insertedAt1 = addRelativeTime 
      ( ib, (Double)realHash.get ( ib ), 0 ) ;

    // did the size of the relative vector change?
    if ( relativeVector.size() != oldSize )
    {
      relabel = 1;
    }
    
    // start at the last place we inserted
    int insertedAt2 = addRelativeTime 
      ( ie, (Double)realHash.get ( ie ), insertedAt1 );

    // did the size of the relative vector change? and do we care...
    if ( ( relabel == 0 ) && ( relativeVector.size() != oldSize ) )
    {
      relabel = 2;
    }

    // do we have to relabel? 
    // Only if the size of the relative vector has changed
    switch ( relabel )
    {
    case 1:
      // start relabeling one relative time ahead of where insert was made
      labelRelativeTimes ( insertedAt1 + 1 );
      break;
    case 2:
      // start relabeling one relative time ahead of where insert was made
      labelRelativeTimes ( insertedAt2 + 1 );
      break;
    } // end switch relabel
    
    switch ( debug ) {
    case 3:
      System.out.println ( "ARMinerItemset.addRelativeTime: " +
			   "relative vector = " + relativeVector );
      System.out.println ( "ARMinerItemset.addRelativeTime: " +
			   "event vector = " + eventVector );
    case 2:
      System.out.println ( "ARMinerItemset.addRelativeTime: " +
			   "added item = " + item + " " +
			   "real time = " + 
			   realHash.get ( ib ) + ":" +
			   realHash.get ( ie ) + " " +
			   "relative time = " + relativeHash.get ( ib ) + ":" +
			   relativeHash.get ( ie ) );
    } // end switch debug -kap
    
  } // END private void addRelativeTime ( int item )

//********************************************************//

  /** 
   * Adds the relative begin and end times of an item to
   * this item set.
   * This method is needed to support time sequence attributes.
   *
   * @param item
   * the item label to be added
   *
   * @param time
   * the real time of the item label
   *
   * @param index
   * the index from which to start scanning for a place
   * to add the item (also the relative time)
   *
   * @ return the relative time of the item.
   */

  private int addRelativeTime ( String item, Double time, int index )
  {
    // we'll need a temp vector for stuff
    Vector tempVector;

    // we have to scan the relative vector for the time
    int relativeTime;

    // remember to start at the specified index
    for ( relativeTime = index; relativeTime < relativeVector.size(); 
	  relativeTime++ )
    {
      if ( ((Double)relativeVector.elementAt ( relativeTime )).doubleValue()
	   > time.doubleValue() )
      {
	break;
      }
      
    } // end for elements in relative vector

    // can we reuse the previous relative time (they have the same real time)
    // or do we have to insert a new one?
    // make sure we are not at the beginning of course
    if ( ( relativeTime > 0 ) &&
	 ( ((Double)relativeVector.elementAt ( relativeTime -1 )).
	   doubleValue() == time.doubleValue() ) )
    {
      relativeTime--;

      // if this item is already in this vector, do not add
      if ( ! ((Vector)eventVector.elementAt ( relativeTime )).
	   contains ( item ) )
      {
	// great, just add to event vector
	((Vector)eventVector.elementAt ( relativeTime )).addElement ( item );
      }
      else
      {
	// this should not happen so print an error message
	System.err.println ( "ARMinerItemset.addRelativeTime: " +
			     "eventVector already contains " + item );
      } // end else it was in the event vector already
      
      // if this item is already in the relative hash, do not add
      if ( relativeHash.get ( item ) == null )
      {
	// assign the time
	relativeHash.put ( item, new Integer ( relativeTime ) );
      }
      else
      {
	// this should not happen so print an error message
	System.err.println ( "ARMinerItemset.addRelativeTime: " +
			     "relativeHash already contains " + item );

      } // end else it was in the relative hash already

      // all done
      return ( relativeTime );

    } // end if we can reuse the time label, yay!    

    // if relative time is at the end of the vector we simply add 
    // (this takes care of when the list is empty too)
    if ( relativeTime == relativeVector.size() )
    {
      // add the item label time to relative time
      relativeVector.addElement ( time );

      // add the relative time labels to the event vector
      tempVector = new Vector();
      tempVector.addElement ( item );
      eventVector.addElement ( tempVector );
      
      // add the time to the relative hash
      relativeHash.put ( item, new Integer ( relativeTime ) );
      
      // all done, yay!
      return ( relativeTime );

    } // end if we stopped at end of relative vector

    // have to insert here I guess
    // make a new vector of vector for event vector
    tempVector = new Vector();
    tempVector.addElement ( item );
    eventVector.insertElementAt ( tempVector, relativeTime );
    
    // insert time in relative vector
    relativeVector.insertElementAt ( time, relativeTime );
    
    // add item key and relative time to hash
    relativeHash.put ( item, new Integer ( relativeTime ) );

    // all done... return the index where we inserted
    return ( relativeTime );

  } // END private int addRelativeTime ( String item, Double time, int index )

//********************************************************//

  /**
   * Relables the relative times of the events in this
   * item set. New values are stored in the relative hash.
   * This method is needed to support time sequence attributes.
   *
   * @param i
   * the reltive time to start relabling items
   */

  private void labelRelativeTimes ( int i )
  {
    // check we have a valid index/time
    if ( ( i < 0 ) || ( i > eventVector.size() ) )
    {
      // then nothing to relabel
      // print out error message and be done with it
      System.err.println ( "ARMinerItemset.labelRelativeTimes ( " +
			   i + " ): " +
			   "time index out of bounds ( 0 - " +
			   eventVector.size() + " )" );
      return;
    } // end if i is out of bounds

    // need an enumeration
    Enumeration e;

    // ok, simply go through the lists of item time labels
    // in event Vector and set the values in relative hash
    // to the index of event Vector under which the item labels
    // are found... it's really quite simple
    for ( int index = i; index < eventVector.size(); index++ )
    {
      e = ((Vector)eventVector.elementAt ( index )).elements();

      while ( e.hasMoreElements() )
      {
	relativeHash.put ( e.nextElement(), new Integer ( index ) );

      } // end while e has more

    } // end for each list of item labels

  } // END private void labelRelativeTimes ( int i )

//********************************************************//

  /**
   * Removes a given item from the itemset.
   *
   * @param item  
   * the item to remove
   *
   * @exception IllegalArgumentException  
   * <code>item</code> is <= 0
   *
   * @return  
   * true if item was removed,
   * false if it wasn't removed (was not found in itemset!)
   */

  public boolean removeItem ( int item ) 
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "ARMinerItemset.removeItem: " + item );
    } // end switch debug -kap

    if ( item <= 0 )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.removeItem ( " + item +
	  " ) negative value for item not allowed" );
    } // end if item was negative
    
    int index;

    for ( index = 0; index < size && item != set [ index ] ; index++ )
      ;
    
    if ( item == set[index] ) 
    {
      System.arraycopy ( set, index + 1, set, index, size - index - 1 ); 
      
      size--;
      sum -= item;

      // remove from time sequence stuff too -kap
      removeEventItem ( item );

      return ( true );
    } 
    else 
    {
      return ( false );
    }

  } // END public boolean removeItem ( int item ) 

//********************************************************//

  /**
   * Removes a given event item from the time sequence
   * specific areas of this itemset.
   * This method is needed to support time sequence attributes.
   *
   * @param item  
   * the item to remove
   */

  private void removeEventItem ( int item ) 
  {
    // if we get null back from name hash we can assume
    // the item is not an event and skip the rest
    if ( nameHash.remove ( new Integer ( item ) ) != null )
    {
      // save labels to avoid repeatedly creating them
      String ib = beginLabel ( item );
      String ie = endLabel ( item );

      nameHash.remove ( ib );
      nameHash.remove ( ie );

      // now the relative time stuff... this is trickier so call the
      // special method for it... 
      // (remember to do this before removing the real time entries
      //  cause we need it for look up in this method)
      removeRelativeTime ( item );

      // now we can remove the real time entries
      realHash.remove ( ib );
      realHash.remove ( ie );

    } // end if item is an event

    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "ARMInerItemset.removeEventItem: " +
			   "removed events for item " + item + " " +
			   "events = " + eventVector + " " +
			   "relative times = " + relativeVector );
    } // end switch debug -kap

  } // END private void removeEventItem ( int item ) 

//********************************************************//

  /** 
   * Removes the relative begin and end times of an item
   * from this item set.
   * This method is needed to support time sequence attributes.
   *
   * @param item
   * the item to remove relative times for
   */

  private void removeRelativeTime ( int item )
  {
    // since this is a private method we can assume
    // that is it only called from removeEventItem when
    // item is an event.

    // get the labels to remove
    String ib = beginLabel ( item );
    String ie = endLabel ( item );

    // keep track of where we have to relabel things from
    // if we have to that is

    // did the size of the relative vector change (so we have to relabel)
    int oldSize = relativeVector.size();
    int relabel = 0;

    // start at relative time 0 and remove begin time
    int removedFrom1 = removeRelativeTime 
      ( ib, (Double)realHash.get ( ib ), 0 );

    // if the index returned is the same as old size
    // then the begin label was not found and that
    // shouldn't happen
    if ( removedFrom1 == oldSize )
    {
      System.err.println ( "ARMinerItemset.removeRelativeTime: " +
			   "label = " + ib + " not found." );
    } // end if item label not found

    // did the size of the relative vector change?
    if ( relativeVector.size() != oldSize )
    {
      relabel = 1;
    }
    
    // start at the last place we removed from and remove end time
    int removedFrom2 = removeRelativeTime 
      ( ie, (Double)realHash.get ( ie ), removedFrom1 );

    // if the index returned is the same as old size
    // then the end label was not found and that
    // shouldn't happen
    if ( removedFrom2 == oldSize )
    {
      System.err.println ( "ARMinerItemset.removeRelativeTime: " +
			   "label = " + ie + " not found." );
    } // end if item label not found

    // did the size of the relative vector change? and do we care...
    if ( ( relabel == 0 ) && ( relativeVector.size() != oldSize ) )
    {
      relabel = 2;
    }

    // do we have to relabel? 
    // Only if the size of the relative vector has changed
    switch ( relabel )
    {
    case 1:
      // start relabeling at the time the removal was made
      labelRelativeTimes ( removedFrom1 );
      break;
    case 2:
      // start relabeling at the time the removal was made
      labelRelativeTimes ( removedFrom2 );
      break;
    } // end switch relabel
    
    switch ( debug ) {
    case 3:
      System.out.println ( "ARMinerItemset.removeRelativeTime: " +
			   "relative vector = " + relativeVector );
    case 2:
      System.out.println ( "ARMinerItemset.removeRelativeTime: " +
			   "removed item = " + item + " " +
			   "real time = " + 
			   realHash.get ( ib ) + ":" +
			   realHash.get ( ie ) + " " +
			   "relative time = " + relativeHash.get ( ib ) + ":" +
			   relativeHash.get ( ie ) + " - " +
			   "(the time values should be missing.)" );
    } // end switch debug -kap
    
  } // END private void removeRelativeTime ( int item )

//********************************************************//

  /** 
   * Removes the relative time and label of an item
   * from this item set.
   * This method is needed to support time sequence attributes.
   *
   * @param item
   * the item label to be removed
   *
   * @param time
   * the real time of the item label
   *
   * @param index
   * the index from which to start scanning for the item
   * to remove (also the relative time)
   *
   * @ return the relative time of the item removed
   */

  private int removeRelativeTime ( String item, Double time, int index )
  {
    // we have to scan the relative vector for the time
    int relativeTime;

    // remember to start at the specified index
    for ( relativeTime = index; relativeTime < relativeVector.size(); 
	  relativeTime++ )
    {
      if ( ((Double)relativeVector.elementAt ( relativeTime )).doubleValue()
	   == time.doubleValue() )
      {
	break;
      }
      
    } // end for elements in relative vector

    // if relative time is at the end of the vector we simply return
    // (- this takes care of when the list is empty too)
    if ( relativeTime == relativeVector.size() )
    {
      // all done, yay!
      return ( relativeTime );

    } // end if we stopped at end of relative vector

    // are any other items associated with this relative time?
    // if NOT we can remove the elements stored at relative time
    // in both event and relative vectors
    if ( ((Vector)eventVector.elementAt ( relativeTime )).size() < 2 )
    {
      relativeVector.removeElementAt ( relativeTime );
      eventVector.removeElementAt ( relativeTime );

    } // end if this item was the only one using that relative time
    else
    {
      // we should only remove the label from event vector
      ((Vector)eventVector.elementAt ( relativeTime )).remove ( item );

    } // else we should only remove the item label from event vector
    
    // remove item key and relative time from hash
    relativeHash.remove ( item );

    // all done... return the index from which we removed stuff
    return ( relativeTime );

  } // END private int removeRelativeTime ( String item, Double time )

//********************************************************//

  /**
   * Removes last item (which has the greatest value) from the itemset. 
   *
   * @return  
   * true if item was removed,
   * false if it wasn't removed (the itemset was empty)
   */

  public boolean removeLastItem()
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "ARMinerItemset.removeLastItem: " );
    } // end switch debug -kap

    if ( size > 0 ) 
    {
      // remove from time sequence stuff first -kap
      removeEventItem ( set [ size - 1 ] );

      sum -= set [ size-1 ];
      size--;

      return ( true );
    } 
    else
    {
      return ( false );
    }

  } // END public boolean removeLastItem()

//********************************************************//

  /**
   * Set the support of the itemset.
   *
   * @param newSupport
   * the support of the itemset
   *
   * @exception IllegalArgumentException  
   * <code>newSupport</code> is < 0 or > 100
   */

  public void setSupport ( float newSupport )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "ARMinerItemset.setSupport: " + newSupport );
    } // end switch debug -kap

    if ( newSupport < 0 || newSupport > 1 )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.setSupport ( newSupport ): " +
	  "support must be between 0 and 1" );
    }

    support = newSupport;

  } // END public void setSupport ( float newSupport )

//********************************************************//

  /**
   * Set the weight of the itemset.
   *
   * @param newWeight 
   * the weight of the itemset
   *
   * @exception IllegalArgumentException  
   * <code>newWeight</code> is < 0
   */

  public void setWeight ( long newWeight )
  {
    if ( newWeight < 0 )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.setWeight ( newWeight ) weight must be positive" );
    }

    weight = newWeight;

  } // END public void setWeight ( long newWeight )

//********************************************************//

  /**
   * Set the event weight of the itemset.
   *
   * @param newWeight 
   * the event weight of the itemset
   *
   * @exception IllegalArgumentException  
   * <code>newWeight</code> is < 0
   */

  public void setEventWeight ( long newWeight )
  {
    if ( newWeight < 0 )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.setEventWeight ( newWeight ) " +
	  "weight must be positive" );
    }

    eventWeight = newWeight;

  } // END public void setEventWeight ( long newWeight )

//********************************************************//

  /**
   * Increment the weight of the itemset.
   */

  public void incrementWeight()
  {
    weight++;

  } // END public void incrementWeight()

//********************************************************//

  /**
   * Increment the event weight of the itemset.
   *
   * @param addWeight
   * how much to increment the event weight
   *
   * @exception IllegalArgumentException  
   * <code>addWeight</code> is < 0
   */

  public void incrementEventWeight ( long addWeight )
  {
    if ( addWeight < 0 )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.incrementEventWeight ( addWeight ) " +
	  "weight must be positive" );
    }

    eventWeight += addWeight;

  } // END public void incrementEventWeight ( long addWeight )

//********************************************************//

  /* This seems to be used only in 
     ARMinerItemset.pruneDuplicates. I don't think we have
     to do anything special to support time sequences here.
     Of course we should if we want to be thorough.
     It would force us to declare whether event attribute
     type and relative time should be used to decide equality
     instead of event attribute type and real time, or using
     just items.
     --kap */

  /**
   * Checks equality with a given itemset.
   *
   * @param itemset
   * the itemset against which we test for equality
   * 
   * @exception IllegalArgumentException
   * <code>itemset</code> is null
   */

  public boolean isEqualTo ( ARMinerItemset itemset )
  {
    if ( itemset == null )
    {
      throw new IllegalArgumentException 
	( "ARMinerItemset.isEqualTo ( itemset ) itemset null" );
    }

    // if they don't have the same number of items they can't be equal
    if ( size != itemset.size() )
    {
      return ( false );
    }

    // if the sum (of integer representations) of the items differ
    // then they can't be equal
    if ( sum != itemset.sum() ) 
    {
      return ( false );
    }

    // guess we'll have to compare each item
    for ( int i = 0; i < size; i++ )
    {
      if ( set [ i ] != itemset.set [ i ] )
      {
	return ( false );
      }

    } // end for each item check equality

    return ( true );

  } // END public boolean isEqualTo ( ARMinerItemset itemset )

//********************************************************//

  /* This method is used to see if an instance can be counted 
   * toward the support of a candidate itemset. It has been
   * changed for time sequence events. (kap)
   */

  /**
   * Checks if this itemset is included in the given
   * itemset. 
   *
   * @param itemset  
   * the itemset against which we test for inclusion
   * (so we test if the items we have are in itemset)
   *
   * @return true if the itemset is found, false otherwise
   *
   * @exception IllegalArgumentException
   * <code>itemset</code> is null
   */

  public boolean isIncludedIn ( ARMinerItemset itemset )
  {
    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "ARMinerItemset.isIncludedIn: " +
			   "Check inclusion of: " + this + 
			   " with nameHash: " + nameHash );
      System.out.print ( "We have items: " );
      for ( int d = 0; d < set.length; d ++ )
      {
	System.out.print ( set [ d ] + " name - " +
			   nameHash.get ( new Integer ( set [ d ] ) ) );
      } // end for each item
      System.out.println ( " Thank you." );
    case 2: case 1: //case 0:
      System.out.println ( "ARMinerItemset.isIncludedIn: " +
			   "Checking ( " + this +
			   " ) for inclusion in ( " + itemset + " )" );
    } // end switch debug -kap

    // need a real itemset
    if ( itemset == null )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.isIncludedIn ( itemset ) itemset null." );

    } // end if itemset is stupid

    // the itemset must have the same or more items
    if ( itemset.size() < size )
    {
      switch ( debug ) {
      case 3: case 2: //case 0:
	System.out.println ( "ARMinerItemset.isIncludedIn: " +
			     "No, itemset not big enough." );
      } // end switch debug -kap

      return ( false );

    } // end if itemset does not have enough items

    // the item set's sum (of internal integer representation of
    // the items) should be the same or higher
    if ( itemset.sum < sum )
    {
      // this might not hold if any of the items are events
      // let's simply check the event vector to see if we
      // have any events. if not we can assume itemset is
      // not included -kap
      if ( eventVector.isEmpty() )
      {
	switch ( debug ) {
	case 3: case 2: //case 0:
	  System.out.println ( "ARMinerItemset.isIncludedIn: " +
			       "No - sums don't match and no events" );
	} // end switch debug -kap

	return ( false );

      } // end if there are no events

    } // end if sums don't match

    int i, j;

    switch ( debug ) {
    case 3:
      System.out.println ( "Gonna check for each item now" ); 
    } // end switch debug -kap

    // compare each item, ignoring events
    // execpt we don't want the constraint of item number value restricting
    // what we look at.
    //    for ( i = 0, j = 0; i < size && j < itemset.size() &&
    // set [ i ] >= itemset.set [ j ]; )

    // have to make so many integer objects
    Integer itemInt;

    for ( i = 0, j = 0; i < size && j < itemset.size() ; )
    {
      itemInt = new Integer ( set [ i ] );

      // if itemInt is an event, ignore it
      if ( nameHash.get ( itemInt ) != null )
      {
	// but first make sure itemset has an event item with the same name
	if ( ! itemset.nameHash.contains ( nameHash.get ( itemInt ) ) )
	{
	  return ( false );

	} // end if itemset does not have the name

	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.isIncludedIn: " +
			       "ignoring this event = " + 
			       numberHash.get ( itemInt ) );
	} // end switch debug -kap

	i++;
	
      } // end if itemInt (set [ i ]) is an event
      else
      {
	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.isIncludedIn: " +
			       "itemset.nameHash: " + itemset.nameHash );
	} // end switch debug -kap

	// if itemset.set [ j ] is an event, ignore it
	if ( itemset.nameHash.get ( new Integer ( itemset.set [ j ] ) ) 
	     != null )
	{
	  switch ( debug ) {
	  case 3:
	    System.out.println ( "ARMinerItemset.isIncludedIn: " +
				 "ignoring that event" +
				 numberHash.get ( new Integer 
						  ( set [ j ] ) ) );
	  } // end switch debug -kap

	  j++;
	  
	} // end if itemset.set [ j ] is an event
	else
	{
	  if ( set [ i ] == itemset.set [ j ] )
	  {
	    switch ( debug ) {
	    case 3: //case 0:
	      System.out.println ( "ARMinerItemset.isIncludedIn: " +
				   "Items are the same baby!" );
	    } // end switch debug -kap

	    // ok, we found item i, go to the next item
	    i++;

	  } // end if the items are the same

	  // go to the next j item regardless if it matched item i
	  // or not. if it matched then great, if not we should try 
	  // the next item
	  j++;

	} // end else check if items are same

      } // end else they might be events
      
    } // end for each item

    // if we didn't make it through all our items we are not included
    // check if we are past all the i items
    if ( i != size )
    {
      // ok, even though it says we didn't make it through the list
      // of items it is possible that there are only event items
      // left in the item set which would be ignored and all the 
      // regular items had been found. -kap 2003-11-13
      
      // simply check if the item i points to is an event, if so
      // we can assume all the items left in the item set are events.
      // we can do this because event items in the itemsets being mined
      // are guaranteed to be numbered higher than regular items.
      // -kap 2003-11-13

      itemInt = new Integer ( set [ i ] );
      
      // if itemInt is an event, ignore it
      if ( nameHash.get ( itemInt ) != null )
      {
	// but first make sure itemset has an event item with the same name
	if ( ! itemset.nameHash.contains ( nameHash.get ( itemInt ) ) )
	{
	  return ( false );
	  
	} // end if itemset does not have the name

      } // end if item is an event
      else
      {
	switch ( debug ) {
	case 3: case 2: //case 0:
	  System.out.println ( "ARMinerItemset.isIncludedIn: No - " +
			       "Didn't find all our regular items ( " +
			       i + "/" + size + ")." );
	} // end switch debug -kap
	
	return ( false );
      } // end else item is a regular item and we did not find it
      // in the itemset we were looking to be included in

    } // end if we didn't find all our regular items

    switch ( debug ) {
    case 3: 
      System.out.println ( "ARMinerItemset.isIncludedIn: " +
			   "event vector = " + eventVector +
			   "nameHash = " + nameHash );
    } // end switch debug -kap

    // do we have any events?
    if ( eventVector.size() > 0 )
    {
      // if we have only one event we can simply check that itemset
      // contains it and skip the relative time nonsense
      if ( numEventItems() == 1 )
      {
	// check that an event with the same name exists in itemset
	// of course we already checked this when going through all
	// our items... so just return true, isn't that cool?
	return ( true );

      } // end if we have only one event

      // make sure itemset has enough relative times
      if ( eventVector.size() <= itemset.eventVector.size() )
      {
	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.isIncludedIn: " +
			       "We have items baby, oh yeah!" );
	} // end switch debug -kap

	// it all depends on the events and their relative time now
	return ( checkRelativeTime ( itemset ) ); 

      } // end if itemset has enough reltive times
      else
      {
	return ( false );

      } // end else itemset didn't have enough relative times
    } // end if we have events

    // if there are no events then we're done and we found all the items
    return ( true );

  } // END public boolean isIncludedIn ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Compares the relative times of the events contained
   * in the itemset specified with this itemset. Tests if
   * all of this itemset's events can be found in the same
   * relative time order in the specified event.
   *
   * @param itemset
   * the itemset to check for relative times in
   *
   * @return
   * true if we find all the events with the same relative
   * temporal relationships, false if not
   */

  private boolean checkRelativeTime ( ARMinerItemset itemset )
  {
    // temporary hashtable
    Hashtable eventLabelMap;

    // our first event begin label has name of
    String firstLabelName = (String)nameHash.
      get ( (String)((Vector)eventVector.get ( 0 )).get ( 0 ) );

    String tempName;

    Vector labels;
    int label;

    // check each possible starting point in itemset's relative
    // times
    for ( int offset = 0; 
	  offset <= itemset.relativeVector.size() - relativeVector.size();
	  offset++ )
    {
      switch ( debug ) {
      case 3:
	System.out.println ( "ARMinerItemset.checkRelativeTime: " +
			     "offset= " + offset );
      } // end switch debug

      // let's try to find a possible starting point...
      // that is, find the first relative time that contains at least
      // the first event begin label
      
      // get the list of event labels from time offset in itemset
      labels = (Vector)itemset.eventVector.get ( offset );

      for ( label = 0; label < labels.size(); label++ )
      {
	tempName = (String)itemset.nameHash.
	  get ( (String)labels.get ( label ) );
	
	if ( firstLabelName.equals ( tempName ) )
	{
	  // try the offset then
	  if ( checkRelativeTime ( itemset, offset ) != null )
	  {
	    // ok, we found one, we're done
	    return ( true );

	  } // end if this offset worked

	} // end if temp name matches

      } // end for labels from time offset

    } // end for each valid offset

    // looks like we didn't find any matches
    return ( false );

  } // END private boolean checkRelativeTime ( ARMinerItemset itemset )
  
//********************************************************//

  /**
   * Compares the relative times of the events contained
   * in the itemset specified with this itemset. If all
   * of this itemsets events can be found in the same
   * relative order in the specified event, a map of event
   * labels in this itemset to those in the itemset sepcified
   * is returned.
   *
   * @param itemset
   * the itemset to check for relative times in
   *
   * @param offset
   * the relative time to start checking in itemset 
   *
   * @return
   * a map of event labels if we find all the relative times, 
   * null otherwise
   */

  private Hashtable checkRelativeTime ( ARMinerItemset itemset,
					int offset )
  {
    // compare the relative begin and end times for the events
    Vector thisEvents, thatEvents;
    
    // relative time keeper
    int relativeTime;
        
    // might have to skip relative times in itemset
    int skip = 0;
    
    // was the item label found
    boolean found;
    
    // indexes for event begin and end labels
    int thisLabel, thatLabel;
    
    // keep track of the relative time we use to match us in itemset
    int[] usedTimes = new int [ relativeVector.size() ];

    for ( relativeTime = 0; relativeTime < relativeVector.size();
	  relativeTime++ )
    {
      // note itemset's relative time that we are trying to use
      usedTimes [ relativeTime ] = relativeTime + offset + skip;

      // get the vector of items for this time
      thisEvents = (Vector)eventVector.elementAt ( relativeTime );
      
      // get a copy of the vector of items from that itemset
      thatEvents = new 
	Vector ( (Vector)(itemset.eventVector.
			  elementAt ( relativeTime + offset + skip ) ) );      
      
      for ( thisLabel = 0; thisLabel < thisEvents.size(); thisLabel++ )
      {
	// find the item label in that with the same name as this
	found = false;

	String thisName = 
	  (String)(nameHash.get ( thisEvents.elementAt ( thisLabel ) ) );
	
	for ( thatLabel = 0; thatLabel < thatEvents.size(); thatLabel++ )
	{      
	  // check if there are enough labels in that
	  // should save checking each label on many occassions
	  if ( thisEvents.size() - thisLabel > thatEvents.size() - thatLabel )
	  {
	    break;
	  } // end if not enough labels left in that

	  // if the attribute name matches and they are both begin
	  // labels we found it
	  if ( thisName.equals ( (String)(itemset.nameHash.
					  get ( thatEvents.
						elementAt ( thatLabel ) ) ) )
	       &&
	       isBeginLabel ( (String)thisEvents.get ( thisLabel ) ) ==
	       isBeginLabel ( (String)thatEvents.get ( thatLabel ) ) )
	  {
	    // yay! we found it
	    found = true;

	    // remove it so we don't look at it again
	    thatEvents.removeElementAt ( thatLabel );
	    
	    // we're all done with this
	    break;
	  } // end if names match
	  
	} // end for each that event label
	
	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.checkRelativeTime: " +
			       "event label = " + thisName + " " +
			       "was found = " + found );	  
	} // end switch debug -kap
	
	// was the item found?
	if ( ! found )
	{
	  // see if we can increment skip
	  if ( itemset.relativeVector.size() - 
	       relativeTime - offset - skip - 1 < 
	       relativeVector.size() - relativeTime )
	  {
	    // we can look no further, so give up
	    return ( null );
	    
	  } // end if we have run out of space to look
	  else
	  {
	    // increment skip
	    skip++;
	    
	    // set the relative time back 1 so we recheck
	    relativeTime--;
	    
	    switch ( debug ) {
	    case 3:
	      System.out.println ( "ARMinerItemset.checkRelativeTime: " +
				   "skip= " + skip + " " +
				   "offset= " + offset + " " +
				   "relativeTime= " + relativeTime );
	    } // end switch debug
	    
	    // remember to break since we want to retry looking
	    // for event labels for this relative time
	    break;

	  } // else offset can be incremented

	} // end if not found

      } // end for each event label for this relative time

      // we may want to check if the used times mapping we have so
      // far let's us check for matching begin and end times yet.
      // we could do this by noting the end time for begin labels
      // we find. when the end time is mapped we then check to see
      // that the matching label is found. we might save a lot of time.
      // it would also make this method even more complicated looking

    } // end for all the relative times

    switch ( debug ) {
    case 3:
      System.out.print ( "ARMinerItemset.checkRelativeTime: " +
			   "usedTimes= " );
      for ( int index = 0; index < usedTimes.length; index++ )
      {
	System.out.print ( " " + index + ":" + usedTimes [ index ] + ":" + 
			     itemset.eventVector.get ( usedTimes [ index ] ) );
      }
      System.out.println ( "" );
    } // end switch debug

    // We (sh/c)ould make sure the pairs of labels match.
    // That is, each time we used a begin label to match one of ours
    // we should make sure the end label we used to match is from the
    // same item.

    // We would have to go through the
    // used times removing pairs of begin-end item labels
    // as we matched them to our own. If we got to the end of our
    // label list (stored in eventVector) without running out of
    // used labels then we have a true match.
    
    // well, here goes...
    return ( checkRelativeTime ( itemset, usedTimes ) );

  } // END private Hashtable checkRelativeTime ( ARMinerItemset itemset,
//				      int offset )
  
//********************************************************//

  /**
   * Compares the relative times of the events contained
   * in the itemset specified with this itemset. If all
   * of this itemsets events can be found in the same
   * relative order using the usedTimes in the specified
   * itemset, a map of event labels in itemset to ours
   * is returned.
   *
   * @param itemset
   * the itemset to check for relative times in
   *
   * @param usedTimes
   * the relative times used in itemset to match our 
   * relative times
   *
   * @ return
   * a map of event labels if we find all the relative times, null otherwise
   */

  private Hashtable checkRelativeTime ( ARMinerItemset itemset,
					int [] usedTimes )
  {
    // get a copy of itemset's event vector
    Vector thatEventVector = itemset.getEventVector();

    switch ( debug ) {
    case 3:
      System.out.println ( "ARMinerItemset.checkRelativeTimes: " +
			   "thatEventVector= " + thatEventVector );
    } // end switch debug

    Vector thisEvents, thatEvents;
    int thisLabel, thatLabel, endTime;
    boolean found = false;

    String thisName, thisBeginLabel, thatBeginLabel, 
      thisEndLabel, thatEndLabel;

    // a hash table mapping event labels in itemset to ours
    Hashtable eventLabelMap = new Hashtable();

    // now go through using the usedTimes as the index of thatEventVector
    for ( int relativeTime = 0; relativeTime < eventVector.size();
	  relativeTime++ )
    {
      // get the vector of items for this time
      thisEvents = (Vector)eventVector.elementAt ( relativeTime );
      
      // get a ref to that using the used Times index
      thatEvents = ( (Vector)thatEventVector.
		     elementAt ( usedTimes [ relativeTime ] ) );
    
      for ( thisLabel = 0; thisLabel < thisEvents.size(); thisLabel++ )
      {
	// find the item label in that with the same name as this,
	// and has a matching end label
	found = false;

	thisBeginLabel = (String)thisEvents.elementAt ( thisLabel );

	// only check for begin labels... then we make sure we have a
	// matching end label
	if ( isBeginLabel ( thisBeginLabel ) )
	{	
	  thisName = (String)nameHash.get ( thisBeginLabel );
	  
	  for ( thatLabel = 0; thatLabel < thatEvents.size(); thatLabel++ )
	  {      
	    thatBeginLabel = (String)thatEvents.get ( thatLabel );

	    // if the attribute name matches and that is a begin
	    // label we can check the end times... check the begin label
	    // condition first since it takes less time
	    if ( isBeginLabel ( thatBeginLabel ) &&
		 thisName.equals ( (String)(itemset.nameHash.
					    get ( thatBeginLabel ) ) ) )
	    {
	      // now we have to check the end times of these and see if
	      // they match... according to the usedTimes mapping that is
	      thisEndLabel = endLabel ( itemInt ( thisBeginLabel ) );
	      
	      endTime = 
		((Integer)relativeHash.get ( thisEndLabel )).intValue();
	      
	      thatEndLabel = endLabel ( itemInt ( thatBeginLabel ) );
	      
	      if ( usedTimes [ endTime ] == 
		   ((Integer)itemset.relativeHash.
		    get ( thatEndLabel )).intValue() ) 
	      {
		// yay! we found it
		found = true;
		
		// put the labels in the label map
		
		// the begin label
		eventLabelMap.put ( thatBeginLabel,
				    thisBeginLabel );
		// the end label
		eventLabelMap.put ( thatEndLabel,
				    thisEndLabel );
		
		// remove that begin and that end label so we don't look
		// at them again
		thatEvents.removeElementAt ( thatLabel );
		
		((Vector)thatEventVector.get ( usedTimes [ endTime ] )).
		  remove ( thatEndLabel );

		// we're all done with this
		break;
		
	      } // end if the end times match
	      
	    } // end if names match
	    
	  } // end for each that event label
	
	  switch ( debug ) {
	  case 3:
	    System.out.println ( "ARMinerItemset.checkRelativeTime: " +
				 "thatEventVector= " + thatEventVector );
	  } // end switch debug
	  
	  // was the item found?
	  if ( ! found )
	  {
	    // we can look no further, so give up
	    return ( null );
	    
	  } // end if not found
	  
	} // end if this begin label
	
      } // end for each event label for this relative time
      
    } // end for all the relative times
    
    switch ( debug ) {
    case 2:
      System.out.println ( "ARMinerItemset.checkRelativeTime: " +
			   "eventLabelMap= " + eventLabelMap );
    } // end switch debug

    // I guess we found everything :)
    return ( eventLabelMap );

  } // END private Hashtable checkRelativeTime ( ARMinerItemset itemset,
  //				      int [] timeMap )
  
//********************************************************//

  /* This method can be used to see if an instance can be
   * counted toward the support of a candidate itemset.
   */

  /**
   * Checks how many times this itemset is included in the
   * given itemset. This uses a different method for
   * determining the number of inclusions based on events
   * than the findAllMaps method. The findAllMaps is the
   * prefered method for the counting purpose.
   *
   * @param itemset  
   * the itemset against which we test for inclusion
   * (so we test if the items we have are in itemset)
   *
   * @return the number times the itemset is found
   *
   * @exception IllegalArgumentException
   * <code>itemset</code> is null
   */

  public int numIncludedIn ( ARMinerItemset itemset )
  {
    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "ARMinerItemset.numIncludedIn: " +
			   "Check inclusion of: " + this + 
			   " with nameHash: " + nameHash );
      System.out.print ( "We have items: " );
      for ( int d = 0; d < set.length; d ++ )
      {
	System.out.print ( set [ d ] + " name - " +
			   nameHash.get ( new Integer ( set [ d ] ) ) );
      } // end for each item
      System.out.println ( " Thank you." );
    case 2: case 1: //case 0:
      System.out.println ( "ARMinerItemset.numIncludedIn: " +
			   "Checking ( " + this +
			   " ) for inclusion in ( " + itemset + " )" );
    } // end switch debug -kap

    // need a real itemset
    if ( itemset == null )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.numIncludedIn ( itemset ) itemset null." );

    } // end if itemset is stupid

    // the itemset must have the same or more items
    if ( itemset.size() < size )
    {
      switch ( debug ) {
      case 3: case 2: //case 0:
	System.out.println ( "ARMinerItemset.numIncludedIn: " +
			     "No, itemset not big enough." );
      } // end switch debug -kap

      return ( 0 );

    } // end if itemset does not have enough items

    // the item set's sum (of internal integer representation of
    // the items) should be the same or higher
    if ( itemset.sum < sum )
    {
      // this might not hold if any of the items are events
      // let's simply check the event vector to see if we
      // have any events. if not we can assume itemset is
      // not included -kap
      if ( eventVector.isEmpty() )
      {
	switch ( debug ) {
	case 3: case 2: //case 0:
	  System.out.println ( "ARMinerItemset.numIncludedIn: " +
			       "No - sums don't match and no events" );
	} // end switch debug -kap

	return ( 0 );

      } // end if there are no events

    } // end if sums don't match

    int i, j;

    switch ( debug ) {
    case 3:
      System.out.println ( "Gonna check for each item now" ); 
    } // end switch debug -kap

    // compare each item, ignoring events
    // execpt we don't want the constraint of item number value restricting
    // what we look at.
    //    for ( i = 0, j = 0; i < size && j < itemset.size() &&
    // set[i] >= itemset.set[j]; )

    // have to make so many integer objects
    Integer itemInt;

    for ( i = 0, j = 0; i < size && j < itemset.size() ; )
    {
      itemInt = new Integer ( set [ i ] );

      // if itemInt is an event, ignore it
      if ( nameHash.get ( itemInt ) != null )
      {
	// but first make sure itemset has an event item with the same name
	if ( ! itemset.nameHash.contains ( nameHash.get ( itemInt ) ) )
	{
	  return ( 0 );

	} // end if itemset does not have the name

	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.numIncludedIn: " +
			       "ignoring this event = " + 
			       numberHash.get ( itemInt ) );
	} // end switch debug -kap

	i++;
	
      } // end if itemInt (set [ i ]) is an event
      else
      {
	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.numIncludedIn: " +
			       "itemset.nameHash: " + itemset.nameHash );
	} // end switch debug -kap

	// if itemset.set [ j ] is an event, ignore it
	if ( itemset.nameHash.get ( new Integer ( itemset.set [ j ] ) ) 
	     != null )
	{
	  switch ( debug ) {
	  case 3:
	    System.out.println ( "ARMinerItemset.numIncludedIn: " +
				 "ignoring that event" +
				 numberHash.get ( new Integer 
						  ( set [ j ] ) ) );
	  } // end switch debug -kap

	  j++;
	  
	} // end if itemset.set [ j ] is an event
	else
	{
	  if ( set [ i ] == itemset.set [ j ] )
	  {
	    switch ( debug ) {
	    case 3: //case 0:
	      System.out.println ( "ARMinerItemset.numIncludedIn: " +
				   "Items are the same baby!" );
	    } // end switch debug -kap

	    // ok, we found item i, go to the next item
	    i++;

	  } // end if the items are the same

	  // go to the next j item regardless if it matched item i
	  // or not. if it matched then great, if not we should try 
	  // the next item
	  j++;

	} // end else check if items are same

      } // end else they might be events
      
    } // end for each item

    // if we didn't make it through all our items we are not included
    // check if we are past all the i items
    if ( i != size )
    {
      // ok, even though it says we didn't make it through the list
      // of items it is possible that there are only event items
      // left in the item set which would be ignored and all the 
      // regular items had been found. -kap 2003-11-13
      
      // simply check if the item i points to is an event, if so
      // we can assume all the items left in the item set are events.
      // we can do this because event items in the itemsets being mined
      // are guaranteed to be numbered higher than regular items.
      // -kap 2003-11-13

      itemInt = new Integer ( set [ i ] );
      
      // if itemInt is an event, ignore it
      if ( nameHash.get ( itemInt ) != null )
      {
	// but first make sure itemset has an event item with the same name
	if ( ! itemset.nameHash.contains ( nameHash.get ( itemInt ) ) )
	{
	  return ( 0 );
	  
	} // end if itemset does not have the name

      } // end if item is an event
      else
      {
	switch ( debug ) {
	case 3: case 2: //case 0:
	  System.out.println ( "ARMinerItemset.numIncludedIn: No - " +
			       "Didn't find all our regular items ( " +
			       i + "/" + size + ")." );
	} // end switch debug -kap
	
	return ( 0 );
      } // end else item is a regular item and we did not find it
      // in the itemset we were looking to be included in

    } // end if we didn't find all our regular items

    switch ( debug ) {
    case 3: 
      System.out.println ( "ARMinerItemset.numIncludedIn: " +
			   "event vector = " + eventVector +
			   "nameHash = " + nameHash );
    } // end switch debug -kap

    // so far we are included
    int included = 1;

    // do we have any events?
    if ( eventVector.size() > 0 )
    {
      // if we have only one event we can simply check that itemset
      // contains it and skip the relative time nonsense
      if ( numEventItems() == 1 )
      {
	// check that an event with the same name exists in itemset
	// of course we already checked this when going through all
	// our items... so just return true, isn't that cool?
	//	return ( true );

	// ok, instead of that let's see how many of that single
	// event there are in itemset

	// we cannot use the count from name hash keys directly
	// because the name hash includes not only the name for
	// an event but the begin and end labels for the event as
	// well. This creates three keys for every event and our
	// count would be wrong. we must divide our count by three to
	// be accurate. -kap 2003-10-02
	
	// get the list of keys from itemset's name hash
	Enumeration e = itemset.nameHash.keys();

	switch ( debug ) {
	case 3: //case 0:
	  System.out.println ( "ARMinerItemset.numIncludedIn: " +
			       "itemset namehash:" + 
			       itemset.nameHash );
	} // end switch debug -kap 2003-10-02
	
	// the number we find
	int numFound = 0;
	
	// name string
	String name = eventName 
	  ( itemInt 
	    ( (String)((Vector)eventVector.
		       firstElement()).firstElement() ) );

	while ( e.hasMoreElements() )
	{
	  if ( name.
	       equals ( (String)(itemset.nameHash.get ( e.nextElement() )) ) )
	  {
	    numFound++;

	  } // end if this key has the same name

	} // end while e has more keys

	// we have to divide by three to get an accurate count.
	// see comment above for details -kap 2003-10-02
	numFound = numFound / 3;

	switch ( debug ) {
	case 3: //case 0:
	  System.out.println ( "ARMinerItemset.numIncludedIn: " +
			       "numFound = " + numFound );
	} // end switch debug

	// return the count baby
	return ( numFound );

      } // end if we have only one event

      // make sure itemset has enough relative times
      if ( eventVector.size() <= itemset.eventVector.size() )
      {
	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.numIncludedIn: " +
			       "We have items baby, oh yeah!" );
	} // end switch debug -kap
	
	// it all depends on the events and their relative time now
	included = mapRelativeTime ( itemset ).size(); 

      } // end if itemset has enough reltive times
      else
      {
	return ( 0 );

      } // end else itemset didn't have enough relative times
    } // end if we have events

    // return our answer
    return ( included );

  } // END public int numIncludedIn ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Compares the relative times of the events contained
   * in the itemset specified with this itemset. If all
   * of this itemset's events can be found in the same
   * relative time order in the specified event, a list of
   * the possible mappings of event labels in the specified
   * itemset to ours is returned.
   *
   * @param itemset
   * the itemset to check for relative times in
   *
   * @return
   * a list of maps of event labels if we find all the relative times, 
   * an empty list if none are found
   */

  private Vector mapRelativeTime ( ARMinerItemset itemset )
  {
    // the list of hash table maps to return
    Vector eventLabelMapVector = new Vector ();

    // temporary hashtable
    Hashtable eventLabelMap;

    // our first event begin label has name of
    String firstLabelName = (String)nameHash.
      get ( (String)((Vector)eventVector.get ( 0 )).get ( 0 ) );

    String tempName;

    Vector labels;
    int label;

    // check each possible starting point in itemset's relative
    // times
    for ( int offset = 0; 
	  offset <= itemset.relativeVector.size() - relativeVector.size();
	  offset++ )
    {
      switch ( debug ) {
      case 3:
	System.out.println ( "ARMinerItemset.mapRelativeTime: " +
			     "offset= " + offset );
      } // end switch debug

      // let's try to find a possible starting point...
      // that is, find the first relative time that contains at least
      // the first event begin label
      
      // get the list of event labels from time offset in itemset
      labels = (Vector)itemset.eventVector.get ( offset );

      for ( label = 0; label < labels.size(); label++ )
      {
	tempName = (String)itemset.nameHash.
	  get ( (String)labels.get ( label ) );
	
	if ( firstLabelName.equals ( tempName ) )
	{
	  // try the offset then
	  eventLabelMap = mapRelativeTime ( itemset, offset );
	  
	  // did it work?
	  if ( eventLabelMap != null )
	  {
	    // ok, we found one, add it to our list
	    eventLabelMapVector.add ( eventLabelMap );

	  } // end if this offset worked
	  else
	  {
	    // no need to try the rest of the labels
	    break;
	  } // end else this offset didn't work

	} // end if temp name matches

      } // end for labels from time offset

    } // end for each valid offset

    // did we find any matches?

    switch ( debug ) {
    case 3:
      if ( eventLabelMapVector.size() > 1 )
      {
	System.out.println ( "ARMinerItemset.mapRelativeTime: " +
			     "num found = " + eventLabelMapVector.size() );
      } // end if num found > 1
    } // end switch debug

    // if we didn't find any matches eventLabelMapVector will be empty,
    // otherwise it will contain valid map(s)
    // so we can just return it
    return ( eventLabelMapVector );

  } // END private Vector mapRelativeTime ( ARMinerItemset itemset )
  
//********************************************************//

  /**
   * Compares the relative times of the events contained
   * in the itemset specified with this itemset. If all
   * of this itemsets events can be found in the same
   * relative order in the specified event, a map of event
   * labels in this itemset to those in the itemset sepcified
   * is returned.
   *
   * @param itemset
   * the itemset to check for relative times in
   *
   * @param offset
   * the relative time to start checking in itemset 
   *
   * @return
   * a map of event labels if we find all the relative times, 
   * null otherwise
   */

  private Hashtable mapRelativeTime ( ARMinerItemset itemset,
					int offset )
  {
    // compare the relative begin and end times for the events
    Vector thisEvents, thatEvents;
    
    // relative time keeper
    int relativeTime;
        
    // might have to skip relative times in itemset
    int skip = 0;
    
    // was the item label found
    boolean found;
    
    // indexes for event begin and end labels
    int thisLabel, thatLabel;
    
    // keep track of the relative time we use to match us in itemset
    int[] usedTimes = new int [ relativeVector.size() ];

    for ( relativeTime = 0; relativeTime < relativeVector.size();
	  relativeTime++ )
    {
      // note itemset's relative time that we are trying to use
      usedTimes [ relativeTime ] = relativeTime + offset + skip;

      // get the vector of items for this time
      thisEvents = (Vector)eventVector.elementAt ( relativeTime );
      
      // get a copy of the vector of items from that itemset
      thatEvents = new 
	Vector ( (Vector)(itemset.eventVector.
			  elementAt ( relativeTime + offset + skip ) ) );      
      
      for ( thisLabel = 0; thisLabel < thisEvents.size(); thisLabel++ )
      {
	// find the item label in that with the same name as this
	found = false;

	String thisName = 
	  (String)(nameHash.get ( thisEvents.elementAt ( thisLabel ) ) );
	
	for ( thatLabel = 0; thatLabel < thatEvents.size(); thatLabel++ )
	{      
	  // check if there are enough labels in that
	  // should save checking each label on many occassions
	  if ( thisEvents.size() - thisLabel > thatEvents.size() - thatLabel )
	  {
	    break;
	  } // end if not enough labels left in that

	  // if the attribute name matches and they are both begin
	  // labels we found it
	  if ( thisName.equals ( (String)(itemset.nameHash.
					  get ( thatEvents.
						elementAt ( thatLabel ) ) ) )
	       &&
	       isBeginLabel ( (String)thisEvents.get ( thisLabel ) ) ==
	       isBeginLabel ( (String)thatEvents.get ( thatLabel ) ) )
	  {
	    // yay! we found it
	    found = true;

	    // remove it so we don't look at it again
	    thatEvents.removeElementAt ( thatLabel );
	    
	    // we're all done with this
	    break;
	  } // end if names match
	  
	} // end for each that event label
	
	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.mapRelativeTime: " +
			       "event label = " + thisName + " " +
			       "was found = " + found );	  
	} // end switch debug -kap
	
	// was the item found?
	if ( ! found )
	{
	  // see if we can increment skip
	  if ( itemset.relativeVector.size() - 
	       relativeTime - offset - skip - 1 < 
	       relativeVector.size() - relativeTime )
	  {
	    // we can look no further, so give up
	    return ( null );
	    
	  } // end if we have run out of space to look
	  else
	  {
	    // increment skip
	    skip++;
	    
	    // set the relative time back 1 so we recheck
	    relativeTime--;
	    
	    switch ( debug ) {
	    case 3:
	      System.out.println ( "ARMinerItemset.mapRelativeTime: " +
				   "skip= " + skip + " " +
				   "offset= " + offset + " " +
				   "relativeTime= " + relativeTime );
	    } // end switch debug
	    
	    // remember to break since we want to retry looking
	    // for event labels for this relative time
	    break;

	  } // else offset can be incremented

	} // end if not found

      } // end for each event label for this relative time

      // we may want to check if the used times mapping we have so
      // far let's us check for matching begin and end times yet.
      // we could do this by noting the end time for begin labels
      // we find. when the end time is mapped we then check to see
      // that the matching label is found. we might save a lot of time.
      // it would also make this method even more complicated looking

    } // end for all the relative times

    switch ( debug ) {
    case 3:
      System.out.print ( "ARMinerItemset.mapRelativeTime: " +
			   "usedTimes= " );
      for ( int index = 0; index < usedTimes.length; index++ )
      {
	System.out.print ( " " + index + ":" + usedTimes [ index ] + ":" + 
			     itemset.eventVector.get ( usedTimes [ index ] ) );
      }
      System.out.println ( "" );
    } // end switch debug

    // We (sh/c)ould make sure the pairs of labels match.
    // That is, each time we used a begin label to match one of ours
    // we should make sure the end label we used to match is from the
    // same item.

    // We would have to go through the
    // used times removing pairs of begin-end item labels
    // as we matched them to our own. If we got to the end of our
    // label list (stored in eventVector) without running out of
    // used labels then we have a true match.
    
    // well, here goes...
    return ( mapRelativeTime ( itemset, usedTimes ) );

  } // END private Hashtable mapRelativeTime ( ARMinerItemset itemset,
//				      int offset )
  
//********************************************************//

  /**
   * Compares the relative times of the events contained
   * in the itemset specified with this itemset. If all
   * of this itemsets events can be found in the same
   * relative order using the usedTimes in the specified
   * itemset, a map of event labels in itemset to ours
   * is returned.
   *
   * @param itemset
   * the itemset to check for relative times in
   *
   * @param usedTimes
   * the relative times used in itemset to match our 
   * relative times
   *
   * @ return
   * a map of event labels if we find all the relative times, null otherwise
   */

  private Hashtable mapRelativeTime ( ARMinerItemset itemset,
					int [] usedTimes )
  {
    // get a copy of itemset's event vector
    Vector thatEventVector = itemset.getEventVector();

    switch ( debug ) {
    case 3:
      System.out.println ( "ARMinerItemset.mapRelativeTimes: " +
			   "thatEventVector= " + thatEventVector );
    } // end switch debug

    Vector thisEvents, thatEvents;
    int thisLabel, thatLabel, endTime;
    boolean found = false;

    String thisName, thisBeginLabel, thatBeginLabel, 
      thisEndLabel, thatEndLabel;

    // a hash table mapping event labels in itemset to ours
    Hashtable eventLabelMap = new Hashtable();

    // now go through using the usedTimes as the index of thatEventVector
    for ( int relativeTime = 0; relativeTime < eventVector.size();
	  relativeTime++ )
    {
      // get the vector of items for this time
      thisEvents = (Vector)eventVector.elementAt ( relativeTime );
      
      // get a ref to that using the used Times index
      thatEvents = ( (Vector)thatEventVector.
		     elementAt ( usedTimes [ relativeTime ] ) );
    
      for ( thisLabel = 0; thisLabel < thisEvents.size(); thisLabel++ )
      {
	// find the item label in that with the same name as this,
	// and has a matching end label
	found = false;

	thisBeginLabel = (String)thisEvents.elementAt ( thisLabel );

	// only check for begin labels... then we make sure we have a
	// matching end label
	if ( isBeginLabel ( thisBeginLabel ) )
	{	
	  thisName = (String)nameHash.get ( thisBeginLabel );
	  
	  for ( thatLabel = 0; thatLabel < thatEvents.size(); thatLabel++ )
	  {      
	    thatBeginLabel = (String)thatEvents.get ( thatLabel );

	    // if the attribute name matches and that is a begin
	    // label we can check the end times... check the begin label
	    // condition first since it takes less time
	    if ( isBeginLabel ( thatBeginLabel ) &&
		 thisName.equals ( (String)(itemset.nameHash.
					    get ( thatBeginLabel ) ) ) )
	    {
	      // now we have to check the end times of these and see if
	      // they match... according to the usedTimes mapping that is
	      thisEndLabel = endLabel ( itemInt ( thisBeginLabel ) );
	      
	      endTime = 
		((Integer)relativeHash.get ( thisEndLabel )).intValue();
	      
	      thatEndLabel = endLabel ( itemInt ( thatBeginLabel ) );
	      
	      if ( usedTimes [ endTime ] == 
		   ((Integer)itemset.relativeHash.
		    get ( thatEndLabel )).intValue() ) 
	      {
		// yay! we found it
		found = true;
		
		// put the labels in the label map
		
		// the begin label
		eventLabelMap.put ( thatBeginLabel,
				    thisBeginLabel );
		// the end label
		eventLabelMap.put ( thatEndLabel,
				    thisEndLabel );
		
		// remove that begin and that end label so we don't look
		// at them again
		thatEvents.removeElementAt ( thatLabel );
		
		((Vector)thatEventVector.get ( usedTimes [ endTime ] )).
		  remove ( thatEndLabel );

		// we're all done with this
		break;
		
	      } // end if the end times match
	      
	    } // end if names match
	    
	  } // end for each that event label
	
	  switch ( debug ) {
	  case 3:
	    System.out.println ( "ARMinerItemset.mapRelativeTime: " +
				 "thatEventVector= " + thatEventVector );
	  } // end switch debug
	  
	  // was the item found?
	  if ( ! found )
	  {
	    // we can look no further, so give up
	    return ( null );
	    
	  } // end if not found
	  
	} // end if this begin label
	
      } // end for each event label for this relative time
      
    } // end for all the relative times
    
    switch ( debug ) {
    case 2:
      System.out.println ( "ARMinerItemset.mapRelativeTime: " +
			   "eventLabelMap= " + eventLabelMap );
    } // end switch debug

    // I guess we found everything :)
    return ( eventLabelMap );

  } // END private Hashtable mapRelativeTime ( ARMinerItemset itemset,
  //				      int [] timeMap )
  
//********************************************************//

  /**
   * Finds all the possible mappings between the event items
   * in this item set and the event items in the specified
   * item set. It is assumed that regular items are not of
   * interest. Inclusion of this itemset in the specified
   * itemset is not determined by this method. Please use
   * isIncludedIn for that purpose.
   *
   * @param itemset
   * the itemset containing the event items to map to
   *
   * @return a list of maps between event items
   */

  public Vector findAllMaps ( ARMinerItemset itemset )
  {
    // the list of maps to return
    Vector mapVector = new Vector();

    // ok, this itemset must have event items and
    // the specified itemset must have the same or more
    // event items than we do
    if ( numEventItems() < 1 ||
	 numEventItems() > itemset.numEventItems() )
    {
      switch ( debug ) {
      case 3: //case 0:
	System.out.println ( "ARMinerItemset.findAllMaps(itemset): " +
			     "Not enough event items in itemset." );
      } // end switch debug

      // nothing to do then
      return ( mapVector );

    } // end if no event items or not enough in other itemset

    // get a list of the event items we have to map
    // into the other item set
    Vector eventItemVector = new Vector();

    int i;

    for ( i = 0; i < size; i++ )
    {
      // just break when we find the first event item
      // we can do this because we assume the event items in
      // the itemsets are number higher than the regular items
      // if this changes so must the logic used here
      if ( eventName ( getItem ( i ) ) != null )
      {
	break;
      } // end if item is an event

    } // end for all items in this item set

    // ok, get the first event item we have to map
    // into the specified item set.
    // event items are selected by the order they begin
    // on the relative time line

    Integer eventItem = new Integer ( getItem ( i ) );

    // the name of the event so we don't have to retrieve it each time
    String eventString = eventName ( eventItem );

    // increment i so we don't have that event item twice
    i++;

    // ok, all the items from i on are events, add them to
    // our list
    for ( ; i < size; i++ )
    {
      eventItemVector.add ( new Integer ( getItem ( i ) ) );
			    
    } // end for the event items
      
    // get a list of the event items in the other itemset
    // that have the same type as this event item
    // put the rest in another list
    Vector otherEventVector = new Vector();
    Vector mapEventVector = new Vector();

    Enumeration eventItemKeys = itemset.nameHash.keys();

    Object tempObject;

    while ( eventItemKeys.hasMoreElements() )
    {
      tempObject = eventItemKeys.nextElement();

      // we want the event item key if it is an Integer
      while ( ( ! eventItem.getClass().isInstance ( tempObject ) ) &&
	      eventItemKeys.hasMoreElements() )
      {
	tempObject = eventItemKeys.nextElement();

      } // end while we do not have an Integer object

      // we might have run out event item keys and tempObject
      // might still not be an integer
      if ( ! eventItem.getClass().isInstance ( tempObject ) )
      {
	break;

      } // end if we ran out of event item keys and still
      // don't have an Integer

      // add this event item to the other event list if it is the 
      // same event type as our subject event item
      if ( eventString.equals ( itemset.eventName ( (Integer)tempObject ) ) )
      {
	mapEventVector.add ( (Integer)tempObject );
	
      } // end if they have the same event type
      
      // make a list of the event items regardless of type
      otherEventVector.add ( (Integer)tempObject );
      
    } // end while more event items

    // now we have a list of all the possible maps (sort of)
    // between this first event item in this item set and
    // all the event items of the same type in the other item
    // set

    // create a hash table to store each map and provide
    // it as a partial map to our recursive find all maps method

    Hashtable map;

    for ( i = 0; i < mapEventVector.size(); i++ )
    {
      map = new Hashtable ();

      map.put ( eventItem, mapEventVector.get ( i ) );

      // we must remove the event item used temporarily
      otherEventVector.remove ( mapEventVector.get ( i ) );

      // the recursive findAllMaps method takes as arguments:
      // - the other itemset
      // - the partial map
      // - the list of event items left to map into the other itemset
      // - the list of event items left in the other itemset

      // we add the results from this to our list of maps to return
      mapVector.addAll 
	( findAllMaps ( itemset, map, eventItemVector, otherEventVector ) );

      // put back the event item we removed
      otherEventVector.add ( mapEventVector.get ( i ) );

    } // end for all possible maps to first event item

    // all done
    return ( mapVector );

  } // END public Vector findAllMaps ( ARMinerItemset itemset )
  
//********************************************************//

  /**
   * Finds all the possible mappings between the event items
   * in this item set and the event items in the specified
   * item set. It is assumed that regular items are not of
   * interest. Inclusion of this itemset in the specified
   * itemset is not determined by this method. Please use
   * isIncludedIn for that purpose.
   *
   * @param itemset
   * the itemset containing the event items to map to
   *
   * @param maps
   * a list of partial map of event items
   *
   * @return a list of partial maps between event items
   */
  
  public Vector findAllMaps ( ARMinerItemset itemset,
			      Vector maps )
  {
    // the list of maps to return
    Vector fullMapVector = new Vector();

    for ( int i = 0; i < maps.size(); i++ )
    {
      fullMapVector.addAll ( findAllMaps 
			     ( itemset, (Hashtable)maps.get ( i ) ) );

    } // end for each partial map

    return ( fullMapVector );

  } // END public Vector findAllMaps ( ARMinerItemset itemset,
  //			      Vector maps )
  
//********************************************************//

  /**
   * Each partial map in the map list is tested to see if
   * it can be expanded to a valid map of the entire
   * item set. The count of partial maps that can be expanded
   * is returned. It is assumed that regular items are not of
   * interest. Inclusion of this itemset in the specified
   * itemset is not determined by this method. Please use
   * isIncludedIn for that purpose.
   *
   * @param itemset
   * the itemset containing the event items to map to
   *
   * @param maps
   * a list of partial map of event items
   *
   * @return the number of partial maps that can be 
   * expanded to a full map
   */
  
  public int countMapExtensions ( ARMinerItemset itemset,
				  Vector maps )
  {
    // the count of full maps
    int numFullMaps = 0;

    Vector mapVector;

    for ( int i = 0; i < maps.size(); i++ )
    {
      // we only need to find one extension for each map
      // specify this when calling finallmaps
      mapVector = findAllMaps ( itemset, (Hashtable)maps.get ( i ), 1 );

      if ( mapVector.size() > 0 )
      {
	numFullMaps++;

	switch ( debug ) {
	case 3: //case 0:
	  System.out.println ( "ARMinerItemset.countMapExtensions: " +
			       "maps of: " + this + "\nto: " + itemset +
			       "\nare: (" + mapVector.size() + ") " + 
			       mapVector );
	} // end switch debug

      } // end if full maps were found

    } // end for each partial map

    return ( numFullMaps );

  } // END public int countMapExtensions ( ARMinerItemset itemset,
  //			                   Vector maps )
  
//********************************************************//

  /**
   * Finds all the possible mappings between the event items
   * in this item set and the event items in the specified
   * item set. It is assumed that regular items are not of
   * interest. Inclusion of this itemset in the specified
   * itemset is not determined by this method. Please use
   * isIncludedIn for that purpose.
   *
   * @param itemset
   * the itemset containing the event items to map to
   *
   * @param map
   * a partial map of event items
   *
   * @return a list of maps between event items
   */
  
  public Vector findAllMaps ( ARMinerItemset itemset,
			      Hashtable map )
  {
    return ( findAllMaps ( itemset, map, 0 ) );

  } // END public Vector findAllMaps ( ARMinerItemset itemset,
  //			      Hashtable map )
  
//********************************************************//

  /**
   * Finds all the possible mappings between the event items
   * in this item set and the event items in the specified
   * item set. It is assumed that regular items are not of
   * interest. Inclusion of this itemset in the specified
   * itemset is not determined by this method. Please use
   * isIncludedIn for that purpose.
   *
   * @param itemset
   * the itemset containing the event items to map to
   *
   * @param map
   * a partial map of event items
   *
   * @param numMaps
   * the maps found so far are returned once this number
   * of maps are found. A value or 0 (zero) or less indicates 
   * find all maps possible.
   *
   * @return a list of maps between event items
   */
  
  public Vector findAllMaps ( ARMinerItemset itemset,
			      Hashtable map, int numMaps )
  {
    // all we have to do is create a list of event items
    // that have not been mapped yet and a list of event
    // items from the specified item set that are not 
    // mapped to yet. then we call the findAllMaps method
    // that takes these as input

    // get a list of the event items we have to map
    // into the other item set
    Vector eventItemVector = new Vector();

    int i;

    for ( i = 0; i < size; i++ )
    {
      // just break when we find the first event item
      // we can do this because we assume the event items in
      // the itemsets are number higher than the regular items
      // if this changes so must the logic used here
      if ( eventName ( getItem ( i ) ) != null )
      {
	break;
      } // end if item is an event

    } // end for all items in this item set

    Integer tempInteger;

    // ok, all the items from i on are events, add them to
    // our list
    for ( ; i < size; i++ )
    {
      tempInteger = new Integer ( getItem ( i ) );

      // make sure this event item is not mapped already
      if ( ! map.containsKey ( tempInteger ) )
      {
	eventItemVector.add ( new Integer ( getItem ( i ) ) );

      } // end if event item is not mapped
	    
    } // end for the event items
     
    // now make the list of event items in the other itemset
    // note we cannot assume that event items are numbered 
    // higher than regular items
    
    Vector otherEventVector = new Vector();

    for ( i = 0; i < itemset.size(); i++ )
    {
      if ( itemset.eventName ( itemset.getItem ( i ) ) != null )
      {
	// make sure event item is not mapped
	tempInteger = new Integer ( itemset.getItem ( i ) );

	if ( ! map.contains ( tempInteger ) )
	{
	  otherEventVector.add ( tempInteger );

	} // end if event item not mapped

      } // end if item is an event

    } // end for each item in other item set

    return ( findAllMaps ( itemset, map, eventItemVector, 
			   otherEventVector, numMaps ) );

  } // END public Vector findAllMaps ( ARMinerItemset itemset,
  //			      Hashtable map, int numMaps )

//********************************************************//

  /**
   * Finds all the possible mappings between the event items
   * in this item set and the event items in the specified
   * item set. It is assumed that regular items are not of
   * interest. Inclusion of this itemset in the specified
   * itemset is not determined by this method. Please use
   * isIncludedIn for that purpose.
   *
   * @param itemset
   * the itemset containing the event items to map to
   *
   * @param map
   * a partial map of event items
   *
   * @param eventItemVector
   * list of event items to map into the specified item set
   *
   * @param otherEventVector
   * list of event items not already mapped to in the 
   * specified item set
   *
   * @return a list of maps between event items
   */
  
  private Vector findAllMaps ( ARMinerItemset itemset,
			       Hashtable map, 
			       Vector eventItemVector,
			       Vector otherEventVector )
  {
    return ( findAllMaps ( itemset, map, eventItemVector,
			   otherEventVector, 0 ) );

  } // END private Vector findAllMaps ( ARMinerItemset itemset,
  //			       Hashtable map, 
  //			       Vector eventItemVector,
  //			       Vector otherEventVector )

//********************************************************//

  /**
   * Finds all the possible mappings between the event items
   * in this item set and the event items in the specified
   * item set. It is assumed that regular items are not of
   * interest. Inclusion of this itemset in the specified
   * itemset is not determined by this method. Please use
   * isIncludedIn for that purpose.
   *
   * @param itemset
   * the itemset containing the event items to map to
   *
   * @param map
   * a partial map of event items
   *
   * @param eventItemVector
   * list of event items to map into the specified item set
   *
   * @param otherEventVector
   * list of event items not already mapped to in the 
   * specified item set
   *
   * @param numMaps
   * the maps found so far are returned once this number
   * of maps are found. A value or 0 (zero) or less indicates 
   * find all maps possible.
   *
   * @return a list of maps between event items
   */
  
  private Vector findAllMaps ( ARMinerItemset itemset,
			       Hashtable map, 
			       Vector eventItemVector,
			       Vector otherEventVector,
			       int numMaps )
  {
    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "ARMinerItemset.findAllMaps" +
			   "(itemset,map,eventItemVector,otherEventVector): " +
			   "itemset: " + itemset +
			   " map: " + map +
			   " eventItemVector: " + eventItemVector +
			   " otherEventVector: " + otherEventVector );
    } // end switch debug

    // if numMaps is less than 0 set it to zero so we can for
    // a single value
    if ( numMaps < 0 )
    {
      numMaps = 0;

    } // end if num maps should be set to 0

    // the list of maps to return
    Vector mapVector = new Vector ();

    // are there event items left to map?
    if ( eventItemVector.size() < 1 )
    {
      // well, then let's see if it is a valid map
      if ( isMapValid ( itemset, map ) )
      {
	// add it to the map list then
	mapVector.add ( map );

      } // end if the mapping is included in the item set

      return ( mapVector );

    } // end if no more event items to map

    // get the next event item to map and remove it from list
    Integer eventItem = (Integer)eventItemVector.remove ( 0 );

    // the name of the event so we don't have to retrieve it each time
    String eventString = eventName ( eventItem );

    // get a list of all the possible event items to map to in the
    // other list
    int i;

    Vector mapEventVector = new Vector();

    for ( i = 0; i < otherEventVector.size(); i++ )
    {
      if ( eventString.equals 
	   ( itemset.eventName ( (Integer)otherEventVector.get ( i ) ) ) )
      {
	mapEventVector.add ( (Integer)otherEventVector.get ( i ) );

      } // end if this event has same type

    } // end for each event item in other list

    // now go through the map list and make new maps for each
    // possibility
    Hashtable newMap;

    for ( i = 0; i < mapEventVector.size(); i++ )
    {
      newMap = new Hashtable ( map );

      // add the new mapping
      newMap.put ( eventItem, mapEventVector.get ( i ) );

      // remove the event item mapped to temporarily
      otherEventVector.remove ( mapEventVector.get ( i ) );

      // recursive time baby!
      mapVector.addAll 
	( findAllMaps ( itemset, newMap, eventItemVector, otherEventVector ) );

      switch ( numMaps )
      {
      case 0:
	break;
      default:
	if ( mapVector.size() >= numMaps )
	{
	  return ( mapVector );

	} // end if there's enough maaps to return
	break;

      } // end switch numMaps

      // put the event item back
      otherEventVector.add ( mapEventVector.get ( i ) );

    } // end for each map

    // ok, all done
    return ( mapVector );

  } // END private Vector findAllMaps ( ARMinerItemset itemset,
  //			       Hashtable map, 
  //			       Vector eventItemVector,
  //			       Vector otherEventVector,
  //			       int numMaps )

//********************************************************//

  /**
   * Tests if the map of event items to the specified
   * item set creates an item set which is included in
   * this item set. If the map is valid this is true.
   * This assumes regular items are not of interest.
   *
   * @param itemset
   * item set mapped to
   *
   * @param map
   * the map of event items to the event items in the 
   * specified item set
   *
   * @return true if the map is valid, false otherwise
   */

  public boolean isMapValid ( ARMinerItemset itemset,
			      Hashtable map )
  {
    // get a list of values from the map, these are the
    // event items from itemset that our event items are
    // mapped to
    Integer [] mappedEvents = 
      (Integer[])( map.values().toArray ( new Integer [ map.size() ] ) );

    ARMinerItemset mappedItemset = 
      ARMinerItemset.buildItemsetFromItemList ( itemset, mappedEvents );

    return ( checkRelativeTime ( mappedItemset ) );

  } // END public boolean isMapValid ( ARMinerItemset itemset,
  //				   Hashtable map )

//********************************************************//

  /**
   * Mark the itemset.
   *
   * @return
   * true if itemset was already marked, false otherwise
   */

  public boolean mark()
  {
    boolean old_mark = mark;
    mark = true;
    return ( old_mark );

  } // END public boolean mark()

//********************************************************//

  /**
   * Unmark the itemset.
   *
   * @return  
   * true if itemset was marked, false otherwise
   */

  public boolean unmark()
  {
    boolean old_mark = mark;
    mark = false;
    return ( old_mark );

  } // END public boolean unmark()

//********************************************************//

  /**
   * Return itemset mark.
   *
   * @return  
   * true if itemset is marked, false otherwise
   */

  public boolean isMarked()
  {
    return ( mark );

  } // END public boolean isMarked()

//********************************************************//

  /**
   * Return a String representation of the ARMinerItemset.
   *
   * @return  
   * String representation of ARMinerItemset
   */

  public String toString()
  {
    StringBuffer sb = new StringBuffer ( "{" );

    // put all but the last item
    int i;
    for ( i = 0; i < size - 1; i++ )
    {
      sb.append ( set [ i ] );
      sb.append ( " " );
    }

    // now we add the last item, no space of course
    sb.append ( set [ i ] );

    sb.append ( "}/[" );
    sb.append ( support );
    sb.append ( "/" );
    sb.append ( weight );
    sb.append ( " event weight: " );
    sb.append ( eventWeight );
    sb.append ( "] (" );
    sb.append ( size );
    sb.append ( ") \nevents: " );
    sb.append ( eventVector );

    return ( sb.toString() );

  } // END public String toString()

//********************************************************//

  /**
   * Check whether two item sets can be combined. 
   * Two item sets can be combined if they differ only in
   * the last item. 
   * To support time sequences a mapping of event labels in
   * itemset to the itemset specified will be returned
   * if the itemsets can be combined. If there are no
   * events an empty map will be returned.
   *
   * @param itemset  
   * itemset with which to combine
   *
   * @exception IllegalArgumentException 
   * <code>itemset</code> is null
   *
   * @return  
   * a map of labels if the item sets can be combined, 
   * null otherwise
   */

  public Hashtable canCombineWith ( ARMinerItemset itemset )
  {
    if ( itemset == null )
    {
      throw new IllegalArgumentException 
	( "ARMinerItemset.canCombineWith ( itemset ): itemset null");
    }

    // they have to be the same size
    if ( size != itemset.size )
    {
      return ( null );
    }

    // we have to have at least one item to combine
    if ( size == 0 )
    {
      return ( null );
    }

    // the sums of all but the last item should be the same
    if ( ( sum - set [ size - 1 ] ) != 
	 ( itemset.sum - itemset.set [ size - 1 ] ) )
    {
      // this might not hold if any of the items are events
      // let's simply check the event vector to see if we
      // have any events. if not we can assume itemset is
      // not included -kap
      if ( eventVector.isEmpty() )
      {
	switch ( debug ) {
	case 3: case 2:
	  System.out.println ( "ARMinerItemset.canCombineWith: " +
			       "No - sums don't match and no events" );
	} // end switch debug -kap

	return ( null );
      } // end if there are no events

    } // end if sum for all but the last items do not match

    // have to check each item I guess
    // compare each item, ignoring events
    int i, j;

    Integer itemI;
    Object nameObject;

    for ( i = 0, j = 0; i < size - 1 && j < size - 1; )
    {
      itemI = new Integer ( set [ i ] );

      // if itemI (set [ i ]) is an event, ignore it
      if ( ( nameObject = nameHash.get ( itemI ) ) != null )
      {
	// but first make sure itemset has the same name
	if ( ! itemset.nameHash.contains ( (String)nameObject ) )
	{
	  return ( null );

	} // end if itemset does not have the name
	
	// ignore it
	i++;
      }
      else
      {
	// if itemset.set [ j ] is an event, ignore it
	if ( itemset.nameHash.get ( new Integer ( itemset.set [ j ] ) ) 
	     != null )
	{
	  j++;
	} // end if itemset.set [ j ] is an event
	else
	{
	  if ( set [ i ] != itemset.set [ j ] )
	  {
	    return ( null );

	  } // end if items don't match

	  i++;
	  j++;

	} // else check if items are the same

      } // end else check itemset.set [ j ]

    } // end for each regular item compare

    // if we didn't make it through all but our last item we cannot
    // combine with itemset
    if ( i != size - 1 )
    {
      switch ( debug ) {
      case 3: case 2:
	System.out.println ( "ARMinerItemset.canCombineWith: " +
			     "No - Did not match all the regular items... " +
			     "besides the last one of course. Stopped on " +
			     i + " out of " + size + " items.");
      } // end switch debug

      return ( null );

    } // end if we didn't find all but our last regular item

    // make sure that the last items are not the same.
    // even though the two items may be events with the same name
    // we still allow the itemsets to combine. this is how multiple
    // types of an event can end up in a rule
    if ( itemset.set [ itemset.size - 1 ] == set [ size - 1 ] )
    {
      // we can't do that... silly people
      return ( null );

    } // end if last items are the same

    // make copies of each itemset, removing the last item
    ARMinerItemset thisCopy = new ARMinerItemset ( this );
    ARMinerItemset thatCopy = new ARMinerItemset ( itemset );

    thisCopy.removeItem ( set [ size - 1 ] );
    thatCopy.removeItem ( itemset.set [ size - 1 ] );

    // one thing to note, if this item set is combined with
    // the specified item set we must ensure that the item
    // which would get added does not already exist in the item set
    // 2003-12-05 -kap
    if ( thisCopy.contains ( itemset.set [ itemset.size - 1 ] ) )
    {
      return ( null );

    } // end if item to add already contained in item set

    // they should have the same number of relative times
    if ( thisCopy.relativeVector.size() != thatCopy.relativeVector.size() )
    {
      // then we can't combine
      return ( null );

    } // end if they don't have the same number of relative times

    // if we made it this far all we have to do is check 
    // the relative times of events. To do this we will
    // be tricky and call the mapRelativeTime method that
    // takes an itemset and a mapping of relative times used to match.
    // Our mapping will be simply 0-0, 1-1, 2-2, etc. This will
    // return false without going through extra work looking
    // for a good starting point

    // prepare the usedTime array
    int [] usedTimes = new int [ thisCopy.relativeVector.size() ];

    for ( i = 0; i < usedTimes.length; i++ )
    {
      usedTimes [ i ] = i;

    } // end for each relative time mapping

    // changed this call from checkRelativeTime to mapRelativeTime
    // since counting event weight is no longer done during support
    // counting. I might want to change this to the new 
    // checkRelativeTime though -2004-01-03 -kap
    return ( thisCopy.mapRelativeTime ( thatCopy, usedTimes ) );

  } // END public Hashtable canCombineWith ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Combine two item sets into a new one that will contain all the
   * items in the first itemset plus the last item in the second
   * itemset. To support time sequence events this has been changed
   * to return a list of itemsets instead of just one.
   *
   * @param itemset  
   * itemset with which to combine
   *
   * @param eventLabelMap
   * A map of event labels from itemset to ours.
   * The map can be obtained from the
   * canCombineWith method. The map can be null if no
   * events are present in itemset.
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null
   *
   * @return  
   * a list of itemsets that combines the two item sets as
   * described above 
   */

  public Vector combineWith ( ARMinerItemset itemset, 
			      Hashtable eventLabelMap ) 
  {
    // need a real itemset to combine with
    if ( itemset == null )
    {
      throw new IllegalArgumentException
	( "ARMinerItemset.combineWith ( itemset ) itemset null." );
    }

    switch ( debug ) {
    case 3: case 2: case 1: //case 0:
      System.out.println ( "ARMinerItemset.combineWith: " );
      System.out.println ( "us  = " + this );
      System.out.println ( "them= " + itemset );
      System.out.println ( "eventLabelMap= " + eventLabelMap );
    } // end switch debug

    // our list of itemsets to return
    Vector result = new Vector();

    // the itemset(s) we will make
    ARMinerItemset is;

    // the item from itemset we have to add
    int item = itemset.set [ itemset.size - 1 ];

    // get the event name of the item we have to add from itemset
    String eventName = itemset.eventName ( item );

    // if we have no events or the item to add is not an event,
    // we can more or less add the item
    // without worrying about the event stuff
    if ( ( eventVector.size() < 1 ) || ( eventName == null ) )
    {
      // only need one itemset to return
      is = new ARMinerItemset ( this );
      is.support = 0;
      is.weight = 0;
      is.eventWeight = 0;
      
      // it will be slightly faster to check if the item to add
      // is an event here since we have access to itemset's 
      // nameHash. If we called the public addItem the value
      // of the item would have to be parsed

      // add it as a regular item first
      is.addRegularItem ( item );

      // if the item is an event then call the event adding stuff
      if ( eventName != null )
      {
	// make sure to call the add event item that takes the name,
	// begin and end time parameters to save having to parse strings
	// for all that info
	is.addEventItem ( item, eventName,
			  (Double)itemset.realHash.get ( beginLabel ( item ) ),
			  (Double)itemset.realHash.get ( endLabel ( item ) ) );

      } // end if adding an event

      switch ( debug ) {
      case 3: case 2:
	System.out.println ( "ARMinerItemset.combineWith: " +
			     "created itemset = " + is + " " +
			     "with events = " + is.eventVector + ", " +
			     "relative time = " + is.relativeVector + " " +
			     "by adding item " + item );
      } // end switch debug
      
      // all done
      result.add ( is );
      return ( result );

    } // end if we have no events or the item to add is not an event

    // ok, so we have one or more events and the item we are adding
    // is an event

    // make lists for begin and end times of the possible 13
    double [] begintimes = new double [ 13 ];
    double [] endtimes = new double [ 13 ];
    
    // how many itemsets we gonna make?
    int numItemsets = 0;

    // if we have just one event then there will be 13 itemsets
    // created and we know what they are
    if ( numEventItems() == 1 )
    {
      // make all 13 possible itemsets
      numItemsets = 13;
      
      // fill in the begin and end time arrays
      fillTimeArrays ( begintimes, endtimes );
      
    } // end if we only have one event
    else
    {
      // if we don't have a valid map we shouldn't do this
      if ( eventLabelMap == null )
      {
	// just print an error and return the empty list
	System.err.println
	  ( "ARMinerItemset.combineWith: Event label map is null. Since " +
	    "the itemsets being combined contain events a map is needed. " +
	    "A valid map can be obtained from the canCombineWith method." );

	return ( result );
      } // end if map not valid

      // get an array of ranges for possible relative times
      int [] range = findRange ( item, itemset, eventLabelMap );

      switch ( debug ) {
      case 3: //case 0:
	System.out.print ( "ARMinerItemset.combineWith: " +
			   "range= " );
	for ( int i = 0; i < range.length; i++ )
	{
	  System.out.print ( range [ i ] + " " );
	}
      System.out.println ( "" );
      } // end switch debug
      
      // now we can fill the time array using the ranges
      numItemsets = fillTimeArrays ( begintimes, endtimes, range );

    } // end else have to figure out the times

    switch ( debug ) {
    case 3:
      System.out.println ( "ARMinerItemset.combineWith: " );
      System.out.print ( "begintimes= " );
      for ( int i = 0; i < begintimes.length; i++ )
      {
	System.out.print ( begintimes [ i ] + " " );
      }
      System.out.println ( "" );
      System.out.print ( "endtimes  = " );
      for ( int i = 0; i < endtimes.length; i++ )
      {
	System.out.print ( endtimes [ i ] + " " );
      }    
      System.out.println ( "" );
    } // end switch debug

    // now simply go through and add itemsets for each 
    // time pair
    for ( int index = 0; index < numItemsets; index++ )
    {
      // make a new itemset just like us
      is = new ARMinerItemset ( this );
      is.support = 0;
      is.weight = 0;
      is.eventWeight = 0;

      // add the regular item info
      is.addRegularItem ( item );

      // add the event info
      is.addEventItem ( item, eventName,
			begintimes [ index ],
			endtimes [ index ] );

      switch ( debug ) {
      case 3:
	System.out.println ( "ARminerItemset.combineWith: " +
			     "added event= " + eventName + ":" +
			     begintimes [ index ] + ":" +
			     endtimes [ index ] );
      } // end switch debug

      // add this itemset to our result list
      result.add ( is );

    } // end for each itemset to make

    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "ARMinerItemset.combineWith: " +
			   "result= " + result );
    } // end switch debug

    return ( result );

  } // END public Vector combineWith ( ARMinerItemset itemset ) 

//********************************************************//
  
  /**
   * Finds the range of possible relative times for the
   * begin and end of an event being added based on common
   * events with anothert itemset. Used by the combineWith
   * method.
   *
   * @param item
   * the item to find the range for
   *
   * @param itemset
   * the itemset sharing events in common
   *
   * @param eventLabelMap
   * a map of event labels from itemset ours
   *
   * @return
   * an array of size 4 with:
   * [0] lower begin bound
   * [1] upper begin bound
   * [2] lower end bound
   * [3] upper end bound
   */

  private int [] findRange ( int item, ARMinerItemset itemset, 
			     Hashtable eventLabelMap )
  {
    // the array to return
    int [] range = new int [ 4 ];

    // find the range for the begin and end times
    int [] beginRange = findRange ( beginLabel ( item ), itemset,
				    eventLabelMap );
    int [] endRange = findRange ( endLabel ( item ), itemset,
				  eventLabelMap );

    // combine them
    range [ 0 ] = beginRange [ 0 ];
    range [ 1 ] = beginRange [ 1 ];
    range [ 2 ] = endRange [ 0 ];
    range [ 3 ] = endRange [ 1 ];

    return ( range );

  } // END private int [] findRange ( int item, ARMinerItemset itemset )

//********************************************************//
  
  /**
   * Finds the range of possible relative times for the
   * event label specified based on common
   * events with anothert itemset. 
   *
   * @param eventLabel
   * the event label to find the range for
   *
   * @param itemset
   * the itemset sharing events in common
   *
   * @param eventLabelMap
   * a map of event labels from itemset to ours
   *
   * @return
   * an array of size 2 with:
   * [ 0 ] lower bound
   * [ 1 ] upper bound
   */

  private int [] findRange ( String eventLabel, 
			     ARMinerItemset itemset,
			     Hashtable eventLabelMap )
  {
    // the array to return
    int [] range = new int [ 2 ];
    
    // what is the relative time of the event label?
    int relativeTime = 
      ((Integer)itemset.relativeHash.get ( eventLabel )).intValue();

    // fill in the ranges we can determine for now
    // if it's a begin label
    if ( isBeginLabel ( eventLabel ) )
    {
      // the lower bound could be 2 times before the start
      range [ 0 ] = -2;
      // the upper bound could be 1 more than the end
      range [ 1 ] = relativeVector.size();
    } // end if it's a begin label
    else
    {
      // it must be an end label

      // the lower bound could be 1 time before the start
      range [ 0 ] = -1;
      // the upper bound could be 2 more than the end
      range [ 1 ] = relativeVector.size() + 1;

    } // end else it's an end label
	 
    // try using common events
    
    // any in the same relative time?
    // get the event vector for that relative time
    Vector v = (Vector)itemset.eventVector.get ( relativeTime );
    
    Integer ourTime;

    if ( v.size() > 1 )
    {
      // find an item that isn't the one we are adding
      for ( int index = 0; index < v.size(); index++ )
      {
	// Please note that this debug switch cannot be simply
	// removed. It executes a vital piece. When debug level
	// is high information not normally available is displayed.
	// this uses memory and processor time, hence the switch.
	switch ( debug ) 
	{
	case 3:
	  // save the strings so we can print them
	  String l1 = (String)v.get ( index );
	  String l2 = (String)eventLabelMap.get ( l1 );

	  // only print if something is going wrong
	  if ( relativeHash == null || l1 == null || l2 == null )
	  {
	    System.err.println ( "ARMinerItemset.findRange: " +
				 "relativeHash= " + relativeHash + " " +
				 "l1= " + l1 + " l2= " + l2 + " " +
				 "this should never happen here." );
	    System.err.println ( "us     = " + this );
	    System.err.println ( "itemset= " + itemset );
	    System.err.println ( "eventLabelMap= " + eventLabelMap );
	    System.err.println ( "us.nameHash= " + nameHash );
	    System.err.println ( "itemset.nameHash= " + itemset.nameHash );
	    System.err.println ( "us.realHash= " + realHash );
	    System.err.println ( "itemset.realHash= " + itemset.realHash );

	  } // end if relative Hash is null, this should never happen here
	  
	  // use the mapping of event labels
	  ourTime = (Integer)relativeHash.get ( l2 );
	  break;

	default:
	  // use the mapping of event labels
	  ourTime = (Integer)relativeHash.
	    get ( eventLabelMap.get ( (String)v.get ( index ) ) );
	  
	} // end switch debug
	
	if ( ourTime != null )
	{
	  range [ 0 ] = ourTime.intValue();
	  range [ 1 ] = range [ 0 ];
	  break;
	} // end if we found an event label that we have...
	
      } // end for each event label
      
    } // end if there are other events with same time
    else
    {
      // we may have to skip some times to find a reference point
      int offset = 0;

      // did we find what we were looking for?
      boolean found = false;

      // are there event labels before?
      while ( relativeTime - offset > 0 )
      {	
	// get the vector of events
	v = (Vector)itemset.eventVector.get ( relativeTime - offset - 1 );

	switch ( debug ) {
	case 3:
	  System.out.println ( "ARMinerItemset.findRange: " +
			       "itemset.eventVector= " + 
			       itemset.eventVector + " - " +
			       "relativeTime= " + relativeTime + " - " +
			       "events at " + (relativeTime - offset - 1) + 
			       "= " + v + " - " +
			       "map from event to us= " + 
			       eventLabelMap.get ( v.get ( 0 ) ) );
	} // end switch ( debug )

	// try event labels until we find one we have a mapping for
	for ( int index = 0; index < v.size(); index++ )
	{
	  if ( eventLabelMap.get ( (String)v.get ( index ) ) != null )
	  {
	    // now, we want the relative after this remember so add 1
	    range [ 0 ] = 
	      ((Integer)relativeHash.
	       get ( (String)eventLabelMap.
		     get ( (String)v.get ( index ) ) )).intValue() + 1;
	    
	    switch ( debug ) {
	    case 2: //case 0:
	      System.out.println ( "ARMinerItemset.findRange: " +
				   "eventLabel= " + eventLabel + " " +
				   "lower bound= " + range [ 0 ] + " " +
				   "relativeTime= " + relativeTime + " " +
				   "offset= " + offset );
	    } // end switch debug

	    // all done
	    found = true;
	    break;
	  } // end if we found a label with mapping
	  
	} // end for each event label
	
	// if we found what we were looking for, we done baby, yeah!
	if ( found )
	{
	  break;
	} // end if found
	else
	{
	  // try the next offset (since we are subtracting offset from
	  // relativeTime we want to increment it here to try the next one)
	  offset++;

	} // end else keep looking

      } // end while there are event labels before begin

      // reset the offset and found
      offset = 0;
      found = false;
      
      // are there event labels after?
      while ( relativeTime + offset < itemset.relativeVector.size() - 1 )
      {
	switch ( debug ) {
	case 2: //case 0:
	  System.out.println ( "ARMinerItemset.findRange: " +
			       "upper bound: eventLabel= " + eventLabel + " " +
			       "relativeTime= " + relativeTime + " " +
			       "offset= " + offset );
	} // end switch debug

	// get the vector of events
	v = (Vector)itemset.eventVector.get ( relativeTime + offset + 1 );
	
	// try event labels until we find one we have a mapping for
	for ( int index = 0; index < v.size(); index++ )
	{
	  if ( eventLabelMap.get ( (String)v.get ( index ) ) != null )
	  {
	    // now, we want the relative before this remember so subtract 1
	    range [ 1 ] = 
	      ((Integer)relativeHash
	       .get ( (String)eventLabelMap.
		      get ( (String)v.get ( index ) ) )).intValue() - 1;

	    switch ( debug ) {
	    case 2: //case 0:
	      System.out.println ( "ARMinerItemset.findRange: " +
				   "upper bound= " + range [ 1 ] );
	    } // end switch debug

	    // all done
	    found = true;
	    break;
	  } // end if we found a label with mapping
	  
	} // end for each event label
	
	// if we found what we were looking for, we done baby, yeah!
	if ( found )
	{
	  break;
	} // end if found
	else
	{
	  // try the next offset
	  offset++;

	} // end else keep looking

      } // end while there are events labels after
      
    } // end else no other event labels for relative time
    
    return ( range );

  } // END private int [] findRange ( String eventLabel,
  //                            ARMinerItemset itemset )

//********************************************************//
  
  /**
   * Fills in the arrays of begin and end times specified
   * for the combineWith method.
   *
   * @param begin
   * the begin time array
   *
   * @param end
   * the end time array
   *
   * @param range
   * the range limiting the 13 possible begin and end times
   *
   * @return the number of begin and end time pairs
   */

  private int fillTimeArrays ( double[] begin, double[] end, 
			       int [] range )
  {
    // indexes
    int beginIndex, endIndex;

    // the start and stop index for end might depend on begin
    int endStart, endStop;

    // the begin time to use in the end time loop
    double beginTime = Double.NEGATIVE_INFINITY;

    // the number of itemsets we have so far
    int numItemsets = 0;

    // it is possible for new relative times to be valid here
    // when there are more than one relative time in a range we
    // should create a new begin and end time between each adjacent
    // pair.

    // this will toggle the creation of an inbetween time during loops
    // remember to initialize them to false in the beginning of a loop
    boolean beginInbetween, endInbetween;

    for ( beginInbetween = false, beginIndex = range [ 0 ]; 
	  beginIndex <= range [ 1 ]; beginIndex++ )
    {
      // if we just inserted a new time we should decrement the
      // begin index for the existing time
      if ( beginInbetween == true )
      {
	beginIndex--;
      } // end if last time a new time was inserted
      
      // shall we insert a new time before the current index?
      if ( 
	  // only if we didn't last time
	  beginInbetween == false &&
	  // and we are not before the beginning of time
	  beginIndex >= 0 &&
	  // and we are not beyond the end of time
	  beginIndex < relativeVector.size() &&
	  // and we don't know exactly where the begin is
	  range [ 0 ] != range [ 1 ]
	  ) // end if we should insert a new time
      {
	beginInbetween = true;
      } // end if we should insert a new time before the current index
      else
      {
	beginInbetween = false;
      } // end else we should not insert a new time before the current index

      // we have to take into consideration what the begin index
      // is when deciding when to start and stop the end index
      
      // normally the start and stop points are those specified
      // in the range
      endStart = range [ 2 ];
      endStop = range [ 3 ];

      // but the end index must come after the begin index
      if ( beginIndex >= endStart )
      {
	// check if this is an inbetween time for begin
	// if it is then both begin and end can have the same
	// index... if not end should start one space after begin
	if ( beginInbetween )
	{
	  endStart = beginIndex;
	} // end if begin inbetween
	else
	{
	  endStart = beginIndex + 1;
	} // end else begin a regular time

      } // end if begin is bigger than endy

      // the end index should only go one space after the end of
      // time if the begin time is not after the end of time
      if ( beginIndex < relativeVector.size() )
      {
	if ( endStop > relativeVector.size() )
	{
	  endStop = relativeVector.size();
	} // end if end stop beyond the end edge of time 
      } // end if begin index not after end of time

      switch ( debug ) {
      case 3: //case 0:
	System.out.println ( "ARMinerITemset.fillTimeArrays: " );
	System.out.println ( "     beginIndex= " + beginIndex + " " +
			     "beginInbetween= " + beginInbetween );
      } // end switch debug

      // ok, first we assign the begin time
      if ( beginInbetween )
      {
	// it may be that beginTime = 0 in which case there is no previous
	// relative time to use. Simply subtract 1 from the first time
	// instead
	if ( beginIndex == 0 )
	{
	  beginTime =
	    ((Double)relativeVector.get ( beginIndex )).doubleValue() - 1.0;

	} // end if begin index 0
	else
	{
	  beginTime =
	    ( ((Double)relativeVector.get ( beginIndex )).doubleValue() +
	      ((Double)relativeVector.get ( beginIndex - 1 )).doubleValue() )
	    / 2.0;

	} // end else begin index more than 0

      } // end if inbetween begin time
      else
      { 
	// if begin index is before the start of time assign a value
	// of the first time plus begin index (which is negative)
	if ( beginIndex < 0 )
	{
	  beginTime =
	    ((Double)relativeVector.get ( 0 )).doubleValue() + beginIndex;

	  switch ( debug ) {
	  case 3: 
	    System.out.println ( "ARMinerITemset.fillTimeArrays: adding " +
				 "begin time before start of relative time. " +
				 "beginTime = " + beginTime + " = " +
				 relativeVector.get ( 0 ) + " + " +
				 beginIndex );
	  } // end switch debug

	} // end if begin before start of time
	else
	{
	  // if the begin index is during time assigning a value is easy
	  if ( beginIndex < relativeVector.size() )
	  {
	    beginTime = 
	      ((Double)relativeVector.get ( beginIndex )).doubleValue();
	  }  // end if begin after the end of time
	  else
	  {
	    // begin must be after the end of time,
	    // assign a value of the last time plus the amount
	    // begin index goes past the end
	    beginTime = 
	      ((Double)relativeVector.lastElement()).doubleValue() +
	      ( beginIndex - ( relativeVector.size() - 1 ) );
	  } // end else begin after the end of time
	  
	} // end else begin after start of time
	
      } // end else not inbetween begin time
      
      for ( endInbetween = false, endIndex = endStart; 
	    endIndex <= endStop; endIndex++ )
      {
	// if we just inserted a new time we should decrement the
	// end index for the existing time
	if ( endInbetween == true )
	{
	  endIndex--;
	} // end if last time a new time was inserted

	// shall we insert a new time before the current index?
	if ( 
	    // only if we didn't last time
	    endInbetween == false &&
	    // and we are not before the beginning of time
	    endIndex >= 0 &&
	    // and we are not beyond the end of time
	    endIndex < relativeVector.size() &&
	    // and we dont know exactly where the end is
	    range [ 2 ] != range [ 3 ]
	    )  // end if we should insert a new time
	{
	  endInbetween = true;
	}
	else
	{
	  endInbetween = false;
	} // end else we should not insert a new time

	switch ( debug ) {
	case 3: //case 0:
	  System.out.println ( "ARMinerITemset.fillTimeArrays: " );
	  System.out.println ( "     beginIndex= " + beginIndex + " " +
			       "beginInbetween= " + beginInbetween + " " +
			       "     endIndex= " + endIndex + " " +
			       "endInbetween= " + endInbetween + " " +
			       "     numItemsets= " + numItemsets );
	} // end switch debug

	// first set the begin time in the array
	begin [ numItemsets ] = beginTime;

	// now we assign the end time
	if ( endInbetween )
	{
	  // this is a little tricky.
	  // it might be that the begin index is the same as end index
	  // and is inbetween as well. This would result in
	  // the begin timer and end time pair being equal. We can't have
	  // a begin and end occur during the same real time so we have
	  // to treat this as a special case.
	  if ( beginIndex == endIndex )
	  {
	    // instead of using the time before end index
	    // we will use the value of begin time
	    end [ numItemsets ] =
	      ( ((Double)relativeVector.get ( endIndex )).doubleValue() +
		beginTime ) / 2.0;
	  } // end if begin and end indexes the same and begin is an
	  // inbetween
	  else
	  {
	    // it may be the case that endIndex == 0 in which case
	    // beginIndex must be less than zero... we should
	    // use beginIndex time instead of relative time -1 to
	    // insert the new time
	    if ( endIndex == 0 )
	    {
	      end [ numItemsets ] =
		( ((Double)relativeVector.get ( endIndex )).doubleValue() +
		  beginTime ) / 2.0;

	    } // end if end index 0
	    else
	    {
	      end [ numItemsets ] =
		( ((Double)relativeVector.get ( endIndex )).doubleValue() +
		  ((Double)relativeVector.get ( endIndex - 1 )).doubleValue() )
		/ 2.0;

	    } // end else end index bigger than zero

	  } // end else end inbetween is simple

	} // end if end inbetween
	else
	{
	  // if end index is before the start of time assign a value
	  // of the first time plus end index (which is negative)
	  if ( endIndex < 0 )
	  {
	    end [ numItemsets ] =
	      ((Double)relativeVector.get ( 0 )).doubleValue() + endIndex;
	  } // end if end before start of time
	  else
	  {
	    // if the end index is during time assigning a value is easy
	    if ( endIndex < relativeVector.size() )
	    {
	      end [ numItemsets ] = 
		((Double)relativeVector.get ( endIndex )).doubleValue();
	    }  // end if end after the end of time
	    else
	    {
	      // end must be after the end of time,
	      // assign a value of the last time plus the amount
	      // end index goes past the end
	      end [ numItemsets ] = 
		((Double)relativeVector.lastElement()).doubleValue() +
		( endIndex - ( relativeVector.size() - 1 ) );
	    } // end else end after the end of time
	    
	  } // end else begin after start of time
	  
	} // end else end not inbetween

	// remember to increment the number of itemsets
	// which will be created by adding an events to an itemset
	// with the begin and end times specified by these loops
	numItemsets++;

      } // end for end times

      // if begin index started before -1 we can skip to -1
      // we do this because the only reason begin index would start
      // there is if end also started before the start of time.
      // we only need one iteration with begin index < 0 to generate
      // all the times for which begin starts before relative time 0
      if ( beginIndex < -1 )
      {
	beginIndex = -1;
      }

    } // end for begin times
    
    // all done
    return ( numItemsets );

  } // END private int fillTimeArrays ( double[] begin, double[] end, 
  //				int range )

//********************************************************//
  
  /**
   * Fills in the arrays of begin and end times specified
   * for the combineWith method.
   *
   * @param begin
   * the begin time array
   *
   * @param end
   * the end time array
   */

  private void fillTimeArrays ( double[] begin, double[] end )
  {
    // now, some of the math may look like we are making assumptions
    // about what the begin and end times are. This is because we are.
    // To avoid extra computation we are taking advantage of the knowledge
    // that the events being used in itemsets NOT representing the 
    // instances of a data set all start out with real times begin = 0,
    // end = 1. If this ever changes this assumption should be removed
    // from this method.
    
    // begin = 0
    // end = 1
    
    // 1 : begin time before our begin time
    //     end time before our begin time
    begin [ 0 ] = -2;
    end [ 0 ] = -1;
    
    // 2 : begin time before our begin time
    //     end time same as our begin time
    begin [ 1 ] = -2;
    end [ 1 ] = 0;
    
    // 3 : begin time before our begin time
    //     end time after our begin time but before our end time
    begin [ 2 ] = -2;
    end [ 2 ] = 0.5;
    
    // 4 : begin time before our begin time
    //     end time same as our end time
    begin [ 3 ] = -2;
    end [ 3 ] = 1;
    
    // 5 : begin time before our begin time
    //     end time after our end time
    begin [ 4 ] = -2;
    end [ 4 ] = 2;
    
    // 6 : begin time same as our begin time
    //     end time after our begin time and before our end time
    begin [ 5 ] = 0;
    end [ 5 ] = 0.5;

    // 7 : begin time same as our begin time
    //     end time after same as our end time
    begin [ 6 ] = 0;
    end [ 6 ] = 1;

    // 8 : begin time same as our begin time
    //     end time after our end time
    begin [ 7 ] = 0;
    end [ 7 ] = 2;

    // 9 : begin time after our begin time and before our end time
    //     end time before our end time
    begin [ 8 ] = 0.5;
    end [ 8 ] = 0.75;

    // 10 : begin time after our begin time and before our end time
    //      end time same as our end time
    begin [ 9 ] = 0.5;
    end [ 9 ] = 1;

    // 11 : begin time after our begin time and before our end time
    //      end time after our end time
    begin [ 10 ] = 0.5;
    end [ 10 ] = 2;

    // 12 : begin time same as our end time
    //      end time after our end time
    begin [ 11 ] = 1;
    end [ 11 ] = 2;

    // 13 : begin time after our end time
    //      end time after our end time
    begin [ 12 ] = 2;
    end [ 12 ] = 3;

  } // END private void fillTimeArrays ( double[] begin, double[] end )

//********************************************************//

  /**
   * Gets the sum of the items in the itemset
   *
   * @return
   * the sum of the items in the itemset
   */

  public int sum() 
  {
    return ( sum );

  } // END public int sum() 

//********************************************************//

  /**
   * Remove all non-maximal item sets from the vector v
   *
   * @param v  
   * the collection of item sets
   */

  public static synchronized void pruneNonMaximal ( Vector v )
  {
    int i, j;
    int size = v.size();

    for ( i = 0; i < size; i++ )
      {
	// see if anything is included in itemset at index i
	for ( j = i + 1; j < size; j++ )
	  if ( ((ARMinerItemset)v.
		get ( j )).isIncludedIn ((ARMinerItemset)v.get ( i )) )
	      {
		// replace this element with last, delete last,
		// and don't advance index
		v.set ( j, v.lastElement () );
		v.remove ( --size );
		j--;
	      }

	// see if itemset at index i is included in another itemset
	for ( j = i + 1; j < size; j++ )
	  if ( ((ARMinerItemset)v.
		get ( i )).isIncludedIn ((ARMinerItemset)v.get ( j )) )
	      {
		// replace this element with last, delete last,
		// and don't advance index
		v.set ( i, v.lastElement() );
		v.remove ( --size );
		i--;
		break;
	      }
      }

  } // END public static synchronized void pruneNonMaximal ( Vector v )

//********************************************************//

  /**
   * Remove all duplicate item sets from the vector v
   *
   * @param v  
   * the collection of item sets
   */

  public static synchronized void pruneDuplicates ( Vector v )
  {
    int i, j;
    int size = v.size();

    for ( i = 0; i < size; i++ )
    {
      // see if anything is equal to itemset at index i
      for ( j = i + 1; j < size; j++ )
      {
	if ( ((ARMinerItemset)v.get ( j )).
	     isEqualTo((ARMinerItemset)v.get ( i )) )
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  v.set ( j, v.lastElement() );
	  v.remove ( --size );
	  j--;
	}
      }
    }
    
  } // END public static synchronized void pruneDuplicates ( Vector v )

//********************************************************//

  /** 
   * for testing purposes only !!! 
   */

  public static void main ( String[] args )
  {
    // tired of checking things worked by hand... output
    // only on error... or if things are being worked on -kap

    System.out.println ( "This main method is for testing only.\n" +
			 "If any errors occur in the expected output it " +
			 "will be noted.\nOtherwise everything " +
			 "is working, more or less. 8^) -kap\n" );

    ARMinerItemset is1 = new ARMinerItemset();
    ARMinerItemset is2 = new ARMinerItemset();

    // the name hash
    Hashtable nh = new Hashtable();
    is1.setNumberHash ( nh );
    is2.setNumberHash ( nh );

    nh.put ( new Integer ( 7 ), "event-a=7:14" );
    is1.addItem ( 7 );
    nh.put ( new Integer ( 3 ), "event-a=3:6" );
    is1.addItem ( 3 );
    nh.put ( new Integer ( 15 ), "event-b=15:30" );
    is1.addItem ( 15 );
    nh.put ( new Integer ( 5 ), "event-a=5:10" );
    is1.addItem ( 5 );
    nh.put ( new Integer ( 12 ), "event-b=12:24" );
    is1.addItem ( 12 );

    is2.addItem ( 12 );
    is2.addItem ( 15 );
    is2.addItem ( 7 );
    is2.addItem ( 5 );
    is2.addItem ( 3 );
    nh.put ( new Integer ( 8 ), "event-a=8:16" );
    is2.addItem ( 8 );

    System.out.println ( "is1= " + is1 );
    System.out.println ( "is2= " + is2 );

    // --- is1 and is2 should share items --- //
    if ( is1.doesIntersect ( is2 ) ) { System.out.print ( "worked - " ); }
    else { System.out.print ( "BROKEN - " ); }

    System.out.print ( "is1 and is2 intersect: " 
		       + is1.doesIntersect ( is2 ) + "\n" );

    if ( is2.doesIntersect ( is1 ) ) { System.out.print ( "worked - " ); }
    else { System.out.print ( "BROKEN - " ); }

    System.out.print ( "is2 and is1 intersect: " 
		       + is2.doesIntersect ( is1 ) + "\n" );
    // END is1 and is2 should share items --- //

    ARMinerItemset is3 = is1.subtract ( is2 );

    
    System.out.println ( "is3 <= subtracting is2 from is1:" + is3 );
    
    System.out.println ( "do is1 and is3 share items: " 
			 + is1.doesIntersect ( is3 ) );
    System.out.println ( "do is3 and is1 share items: " 
			 + is3.doesIntersect ( is1 ) );
    
    is3 = is2.subtract ( is1 );
    System.out.println ( "is3 <= subtracting is1 from is2:" + is3 );
    
    System.out.println ( "do is1 and is3 share items: " 
			 + is1.doesIntersect ( is3 ) );
    System.out.println ( "do is3 and is1 share items: " 
			 + is3.doesIntersect ( is1 ) );
    
    System.out.println ( "do is3 and is2 share items: " 
			 + is3.doesIntersect ( is2 ) );
    System.out.println ( "do is2 and is3 share items: " 
			 + is2.doesIntersect ( is3 ) );
    
    System.out.println ( "adding is2 to is1:" + is1.add ( is2 ) );
    System.out.println ( "adding is1 to is2:" + is2.add ( is1 ) );
 
    System.out.println ( "\nis1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );

    System.out.println ( "is1 equal to is2: " + is1.isEqualTo ( is2 ) );
    System.out.println ( "is1 included in is2: " + is1.isIncludedIn ( is2 ) );
    System.out.println ( "is2 included in is1: " + is2.isIncludedIn ( is1 ) );
    
    is1.addItem ( 8 );

    System.out.println ( "\nis1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );

    System.out.println ( "is1 equal to is2: " + is1.isEqualTo ( is2 ) );
    System.out.println ( "is1 included in is2: " + is1.isIncludedIn ( is2 ) );
    System.out.println ( "is2 included in is1: " + is2.isIncludedIn ( is1 ) );

    nh.put ( new Integer ( 1 ), "event-a=1:2" );
    is1.addItem ( 1 );

    System.out.println ( "\nis1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );
 
    System.out.println ( "is1 equal to is2: " + is1.isEqualTo ( is2 ) );
    System.out.println ( "is1 included in is2: " + is1.isIncludedIn ( is2 ) );
    System.out.println ( "is2 included in is1: " + is2.isIncludedIn ( is1 ) );

    nh.put ( new Integer ( 50 ), "event-b=50:100" );
    is1.addItem ( 50 );
 
    System.out.println ( "\nis1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );

    System.out.println ( "is1 equal to is2: " + is1.isEqualTo ( is2 ) );
    System.out.println ( "is1 included in is2: " + is1.isIncludedIn ( is2 ) );
    System.out.println ( "is2 included in is1: " + is2.isIncludedIn ( is1 ) );

    nh.put ( new Integer ( 100 ), "event-b=0:100" );
    is1.addItem ( 100 );

    System.out.println ( "\nis1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );

    System.out.println ( "is1 equal to is2: " + is1.isEqualTo ( is2 ) );
    System.out.println ( "is1 included in is2: " + is1.isIncludedIn ( is2 ) );
    System.out.println ( "is2 included in is1: " + is2.isIncludedIn ( is1 ) );

    nh.put ( new Integer ( 70 ), "event-b=70:140" );
    System.out.println ( "adding 70 to is2: " + is2.addItem ( 70 ) );
    System.out.println ( "adding 70 to is2: " + is2.addItem ( 70 ) + "\n" );

    System.out.println ( "is1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );

    System.out.println ( "is1 equal to is2: " + is1.isEqualTo ( is2 ) );
    System.out.println ( "is1 included in is2: " + is1.isIncludedIn ( is2 ) );
    System.out.println ( "is2 included in is1: " + is2.isIncludedIn ( is1 ) );

    System.out.println ( "removing 1 from is1: " + is1.removeItem ( 1 ) );
    System.out.println ( "removing 1 from is1: " + is1.removeItem ( 1 ) );
    System.out.println ( "is1: " + is1);
    System.out.println ( "removing 50 from is1: " + is1.removeItem ( 50 ) );
    System.out.println ( "is1: " + is1);
    System.out.println ( "removing 70 from is2: " + is2.removeItem ( 70 ) );
    System.out.println ( "is2: " + is2 );

    System.out.print ( "going through items of is1:" );
    
    while ( is1.hasMoreItems() )
    {
      System.out.print ( " " + is1.getNextItem() );
    }

    System.out.print ( "\ngoing through items of is2:" );
    
    while ( is2.hasMoreItems() )
    {
      System.out.print ( " " + is2.getNextItem() );
    }
    
    System.out.println ( "\nis1 first item: " + is1.getFirstItem() );
    System.out.println ( "is1 next item: " + is1.getNextItem() );
    System.out.println ( "is2 first item: " + is2.getFirstItem() );
    System.out.println ( "is2 next item: " + is2.getNextItem() + "\n" );

    System.out.println ( "is1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );

    Hashtable eventLabelMap = is1.canCombineWith ( is2 );
    
    System.out.println ( "Can we combine is1 and is2 = " +
			 "map= " + eventLabelMap + "(null means false)" );

    System.out.println ( "Adding item 50 to is2: " + is2.addItem ( 50 ) + 
			 "\n" );

    System.out.println ( "is1= " + is1 );
    System.out.println ( "is2= " + is2 + "\n" );

    eventLabelMap = is1.canCombineWith ( is2 );
    System.out.println ( "Can we combine is1 and is2 = " +
			 eventLabelMap + " ( null = false )\n" );
    Enumeration e =  is1.combineWith ( is2, eventLabelMap ).elements();
    System.out.println ( "" );

    while ( e.hasMoreElements() )
    {
      ARMinerItemset combined = (ARMinerItemset)e.nextElement();
      
      System.out.println ( "Combining is1 and is2 = " + combined );
      System.out.println ( "combined.realHash = " + combined.getRealHash() + 
			   "\n" );
    } // end while e has more
      
    System.out.println ( "Removing item 50 to is2: " + is2.removeItem ( 50 ) );
    ARMinerItemset copy =  new ARMinerItemset ( is1 );
    System.out.println ( "\nis1  = " + is1 + ", " );
    System.out.println ( "copy = " + copy + ", \n" );
    System.out.println ( "is1.relativeVector  = " + 
			 is1.relativeVector + ", " );
    System.out.println ( "copy.getRelativeVector() = " + 
			 copy.getRelativeVector() + ", \n" );
    System.out.println ( "is1.relativeHash  = " + 
			 is1.relativeHash + ", " );
    System.out.println ( "copy.relativeHash = " + 
			 copy.relativeHash + ", " + "\n" );

    System.out.println ( "is1.realHash  = " + is1.realHash + ", " );
    System.out.println ( "copy.realHash = " + copy.realHash + ", " + "\n" );
    System.out.println ( "is1.nameHash  = " + is1.nameHash );
    System.out.println ( "copy.nameHash = " + copy.nameHash + "\n" );

    System.out.println ( "is2: " + is2 );
    System.out.println ( "removing all items from is2" );
    while ( is2.removeLastItem() )
    {
      ;
    }

    System.out.println ( "is2: " + is2 );

    System.out.println ( "mark is1, previous state: " + is1.mark() );
    System.out.println ( "mark is1, previous state: " + is1.mark() );
    System.out.println ( "is1 mark state: " + is1.isMarked() );
    System.out.println ( "unmark is1, previous state: " + is1.unmark() );
    System.out.println ( "unmark is1, previous state: " + is1.unmark() );

    ARMinerItemset is4 = new ARMinerItemset ( is1 );
    is4.addItem ( 50 );
    System.out.println ( is4 );

    // --- contrived combineWith test --- //
    System.out.println ( "\nContrived combineWith test\n" );
    is1 = new ARMinerItemset();
    is1.setNumberHash ( nh );

    is1.addItem ( 3 );
    is2 = new ARMinerItemset ( is1 );

    is1.addItem ( 100 );
    is2.addItem ( 7 );

    eventLabelMap = is1.canCombineWith ( is2 );
    System.out.println ( "is1= " + is1 );
    System.out.println ( "is2= " + is2 );
    System.out.println ( "is1.canCombineWith ( is2 )= " + eventLabelMap + 
			 "\n" );

    e = is1.combineWith ( is2, eventLabelMap ).elements();

    while ( e.hasMoreElements() )
    {
      ARMinerItemset combined = (ARMinerItemset)e.nextElement();
      System.out.println ( "itemset= " + combined );
      System.out.println ( "times= " + combined.getRelativeVector() );
    } 
    // END contrived combineWith test --- //

    // trouble shooting test

    nh.clear();

    nh.put ( new Integer ( 7 ), "just-a-plain-old-item" );
    nh.put ( new Integer ( 11764 ), "a=0:1" );
    nh.put ( new Integer ( 11799 ), "b=0:3" );
    nh.put ( new Integer ( 11763 ), "a=0:1" );
    nh.put ( new Integer ( 11800 ), "b=0:3" );

    is1 = new ARMinerItemset();
    is1.setNumberHash ( nh );

    is1.addItem ( 11764 );

    is2 = new ARMinerItemset ( is1 );

    is1.addItem ( 7 );
    is1.addItem ( 11799 );

    is2.addItem ( 11763 );
    is2.addItem ( 11764 );
    is2.addItem ( 11800 );

    System.out.println ( "\nTrouble shooting problem...\n" );
    System.out.println ( "is1= " + is1 );
    System.out.println ( "is2= " + is2 );

    eventLabelMap = is1.canCombineWith ( is2 );

    System.out.println ( "eventLabelMap= " + eventLabelMap );

    System.out.println ( "combined= " + is1.combineWith ( is2, 
							  eventLabelMap ) );

  } // END public static void main ( String[] args )

//********************************************************//

} // END public class ARMinerItemset implements java.io.Serializable
