/*
ARMiner - Association Rules Miner
Copyright (C) 2000  UMass/Boston - Computer Science Department

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The ARMiner Server was written by Dana Cristofor and Laurentiu
Cristofor.

The ARMiner Client was written by Abdelmajid Karatihy, Xiaoyong Kuang,
and Lung-Tsung Li.

The ARMiner package is currently maintained by Laurentiu Cristofor
(laur@cs.umb.edu).
*/

/*
   This file is a part of the ARMiner project.
   
   (P)1999-2001 by ARMiner Server Team:

   Dana Cristofor
   Laurentiu Cristofor

   Modified by Keith A . Pray, (kap@wpi.edu) 2002
*/

package wpi.associations.arminerSequence;

import java.util.*;

/*
  Maintenance log started on April 3rd, 2001 by Laurentiu Cristofor

  Apr. 3rd, 2001    - modified the HashTree so that it can change
                      dynamically its HashNode size in order to
                      index large numbers of itemsets.

  June 1, 2002      - Made slight modification to incorporate this
                      newer version of HashTree into Weka.
		      Includes package declaration and Itemset - 
		      ARMinerItemset substitution. -kap

  June 3, 2002      - cleaned up formatting and commented code -kap
*/ 

/**
 * A special data structure that is used to index a Vector of
 * ARMinerItemset objects for more efficient processing.
 *
 * @author Dana Cristofor
 * @author Laurentiu Cristofor
 * @author Keith A. Pray (kap@wpi.edu)
*/

public class HashTree
{

//********************************************************//
// ----- Stuff -----
//********************************************************//  

  private static final int LIST_NODE = 1;
  private static final int HASH_NODE = 2;

  private static final int DEFAULT_LIST_SIZE = 20;
  private static final int DEFAULT_HASH_SIZE = 40;

  private int LIST_SIZE;
  private int HASH_SIZE;

//********************************************************//
// ----- Inner Classes -----
//********************************************************//  

  private static class Node
  {
    public int type; // LIST_NODE or HASH_NODE

  } // END private static class Node

  //********************************************************//  

  private class ListNode extends Node
  {
    public int[] indexes;   // index in Vector of ARMinerItemsets
    public int size;        // how many indexes we keep in above array
    public boolean visited; // have we seen this node?

    //--------------------------------------------------------//  

    public ListNode()
    {
      type = LIST_NODE;
      indexes = new int [ LIST_SIZE ];
      size = 0;
      visited = false;

    } // END public ListNode()

    //--------------------------------------------------------//  

  } // END private class ListNode extends Node

  //********************************************************//

  private class HashNode extends Node
  {
    public MyHashtable children;

    //--------------------------------------------------------//  

    public HashNode()
    {
      type = HASH_NODE;
      children = new MyHashtable ( HASH_SIZE );

    } // END public HashNode()

    //--------------------------------------------------------//  

  } // END private class HashNode extends Node

  //********************************************************//

  private static class MyHashtable
  {
    public Node[] contents;

    //--------------------------------------------------------//  

    public MyHashtable ( int size )
    {
      contents = new Node [ size ];

    } // END public MyHashtable ( int size )

    //--------------------------------------------------------//  

    public void put ( int key, Node n )
    {
      int index = key % contents.length;
      contents [ index ] = n;

    } // END public void put ( int key, Node n )

    //--------------------------------------------------------//  

    public Node get ( int key )
    {
      int index = key % contents.length;

      return ( contents [ index ] );

    } // END public Node get ( int key )

    //--------------------------------------------------------//  

    public Enumeration elements()
    {
      return new Enumeration()
	{
	  int i = 0;

	  //--------------------------------------------------------//

	  public boolean hasMoreElements()
	  {
	    while ( i < contents.length && contents [ i ] == null )
	    {
	      i++;
	    } // end while
	    
	    if ( i >= contents.length )
	    {
	      return ( false );
	    } // end if
	    else
	    {
	      return ( true );
	    } // end else

	  }  // end public boolean hashMoreElements()

	  //--------------------------------------------------------//

	  public Object nextElement()
	  {
	    while ( i < contents.length && contents [ i ] == null )
	    {
	      i++;
	    } // end while

	    if ( i >= contents.length )
	    {
	      throw new NoSuchElementException();
	    } // end if
	    else
	    {
	      return ( contents [ i++ ] );
	    } // end else

	  } // end public Object nextElement()

	  //--------------------------------------------------------//

	}; // end return new Enumeration()

    } // END public Enumeration elements()

    //--------------------------------------------------------//  

  } // END private static class MyHashtable

  //********************************************************//
  
  private static class HashTreeOverflowException 
    extends RuntimeException
  {

  } // END private static class HashTreeOverflowException 
  //    extends RuntimeException

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** used for some computations */
  private int counter;

  /** keeps all leaves of the HashTree */
  private Vector leaves;

  /** from the getCandidate method: */
  private Vector itemsets;

  /** the root of the HashTree */
  private Node theRoot;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Sets each leaf to unvisited.
   */

  private void unvisitLeaves()
  {
    for ( int i = 0; i < leaves.size(); i++ )
    {
      ((ListNode)leaves.get ( i )).visited = false;
    }

  } // END private void unvisitLeaves()

//********************************************************//

  /**
   * Create a new HashTree. The <code>listSize</code> parameter determines
   * after how many inserts in a ListNode we have to change it to a 
   * HashNode (i.e. perform a split). The <code>hashSize</code> parameter
   * can be specified to improve the efficiency of the structure.
   *
   * @param listSize  
   * the size of the internal lists in the list nodes
   *
   * @param hashSize  
   * the size of the internal hashtables in the hash nodes
   *
   * @param itemsets  
   * the Vector of ARMinerItemsets that we should index
   *
   * @exception IllegalArgumentException  
   * <code>itemsets</code> is null
   * or <code>listSize <= 0</code> or <code>hashSize <= 0</code>  
   */

  public HashTree ( int listSize, int hashSize, Vector itemsets )
  {
    if ( itemsets == null || listSize <= 0 || hashSize <= 0 )
    {
      throw new IllegalArgumentException ( "HashTree constructor: " +
					   "invalid arguments");
    } // end if arguments stupid

    LIST_SIZE = listSize;
    HASH_SIZE = hashSize;
    this.itemsets = itemsets;

    theRoot = new ListNode();
    leaves = new Vector();

  } // END public HashTree ( int listSize, int hashSize,
  // Vector itemsets )

//********************************************************//

  /**
   * Create a new HashTree. This initializes the HashTree with
   * default parameters.
   *
   * @param itemsets  
   * the Vector of ARMinerItemsets that we should index
   *
   * @exception IllegalArgumentException  
   * <code>itemsets</code> is null
   */

  public HashTree ( Vector itemsets )
  {
    this ( DEFAULT_LIST_SIZE, DEFAULT_HASH_SIZE, itemsets );

  } // END public HashTree ( Vector itemsets )

//********************************************************//

  /**
   * This method should be called before calling update()
   * to gather all leaves of the HashTree for more efficient
   * processing.
   */

  public void prepareForDescent()
  {
    leaves.clear();
    prepare ( theRoot );

  } // END public void prepareForDescent()

//********************************************************//

  /** Prepares the hash tree for desecent recursively. */

  private void prepare ( Node node )
  {
    if ( node.type == HASH_NODE )
    {
      Enumeration e = ((HashNode)node).children.elements();

      while ( e.hasMoreElements() )
      {
	prepare ( (Node)e.nextElement() );

      } // end while e has more nodes

    } // end if hash node type
    else
    {
      // it must be a list node
      leaves.add ( node );

    } // end else it must be a list node

  } // END private void prepare ( Node node )

//********************************************************//

  /**
   * Reset HashTree and readd all itemsets added previously
   * this method may be called by add()
   */

  private void readdAll()
  {
    // increase size of hash nodes
    HASH_SIZE = 2 * HASH_SIZE + 1;

    // first save leaf nodes
    prepareForDescent();

    // then reset HashTree
    theRoot = new ListNode();

    // readd everything that we added before
    for ( int i = 0; i < leaves.size(); i++ ) 
    {
      ListNode ln = (ListNode)leaves.get ( i );

      for ( int j = 0; j < ln.size; j++ )
      {
	add ( ln.indexes [ j ] );
      } // end for

    } // end for 

    // clean up
    leaves.clear();

  } // END private void readdAll()

//********************************************************//

  /**
   * This method indexes in the HashTree the ARMinerItemset at index
   * <code>index</code> from Vector <code>itemsets</code> which
   * was passed to the constructor of this HashTree.
   *
   * @param index  
   * the index of the ARMinerItemset that we need to index in
   * this HashTree.
   */

  public void add ( int index )
  {
    // repeat the operation if it results in a HashTree overflow
    while ( true )
    {
      try 
      {
	theRoot = add ( theRoot, 0, index );
	
	// if we get here then the add() was successful and we can
	// exit the loop
	break;

      } // end try
      catch ( HashTreeOverflowException e )
      {
	// call readdAll to increase the size of hash nodes 
	// and readd all itemsets that were previously added
	readdAll();

      } // end catch 

    } // end while forever

  } // END public void add ( int index )

//********************************************************//

  /**
   * Adds a node recursively.
   *
   * @param node
   *
   * @param level
   *
   * @param index
   */

  private Node add ( Node node, int level, int index )
  {
    if ( node.type == LIST_NODE )
    {
      ListNode ln = (ListNode)node;
      
      if ( ln.size == LIST_SIZE ) // list is full
      {
	// if the level is equal to the itemsets size and we
	// filled the list node, then we overflowed the HashTree.
	if ( ((ARMinerItemset)itemsets.get ( index )).size() == level )
	{
	  throw ( new HashTreeOverflowException() );
	} // end if hash tree overflowed

	// else, must split!
	HashNode hn = new HashNode();
	
	// hash the list elements
	for ( int i = 0; i < LIST_SIZE; i++ )
	{
	  add ( hn, level, ln.indexes [ i ] );

	} // end for itemsets in the list node
	
	// add our node
	add ( hn, level, index );
	
	// return this HashNode to replace old ListNode
	return ( hn );

      } // end if list is full
      else
      {
	// append index at end of list
	ln.indexes [ ln.size++ ] = index;
	
      } // end else list isn't full

    } // end if list node
    else 
    {
      // it must be a hash node
      HashNode hn = (HashNode)node;

      // compute hash key
      ARMinerItemset is = (ARMinerItemset)itemsets.get ( index );

      int key = is.getItem ( level );
      
      // try to get next node
      Node n = hn.children.get ( key );

      if ( n == null ) // no node, must create a new ListNode
      {
	ListNode ln = new ListNode();
	ln.indexes [ ln.size++ ] = index; 
	hn.children.put ( key, ln );

      } // end if there's no node
      else
      {
	// found a node, do a recursive call
	n = add ( n, level + 1, index );

	hn.children.put ( key, n );

      } // end else 

    } // end else hash node

    return ( node );
  } // END private Node add ( Node node, int level, int index )

//********************************************************//

  /**
   * Update the weights of all indexed ARMinerItemsets
   * that are included in <code>row</code>
   *
   * @param row
   * the ARMinerItemset (normally a database row) against
   * which we test for inclusion
   */

  public void update ( ARMinerItemset row )
  {
    update ( theRoot, row, 0 );

    // remember to unvisit the leaves
    unvisitLeaves();

  } // END public void update ( ARMinerItemset row )

//********************************************************//

  /** 
   * Updates the support of itemsets recursively.
   *
   * @param node
   *
   * @param row
   * the ARMinerItemset (normally a database row) against
   * which we test for inclusion
   *
   * @param index
   */

  private void update ( Node node, ARMinerItemset row, int index )
  {
    if ( node.type == LIST_NODE )
    {
      ListNode ln = (ListNode)node;
      
      if ( ln.visited )
      {
	return;
      }

      for ( int i = 0; i < ln.size; i++ )
      {
	ARMinerItemset is = (ARMinerItemset)itemsets.get ( ln.indexes [ i ] );

	if ( is.isIncludedIn ( row ) )
	{
	  is.incrementWeight();
	} // end if itemset was included in row

      } // end for each itemset in list node

      ln.visited = true; // now we've seen this node

    } // end if list node
    else
    {
      // must be a hash node
      HashNode hn = (HashNode)node;

      // the tricky piece of algorithm below doesn't fly for
      // event items. We should be updating itemsets based on
      // attribute name for event items in addition to matching
      // the actual integer ID of an item. -kap
      
      // this is a tricky piece of algorithm that ensures we
      // look for all possible subsets of row
      for ( int i = index; i < row.size(); i++ )
      {
	int key = row.getItem ( i );

	// remember to check if it's an event item -kap

	Node n = hn.children.get ( key );

	if ( n != null )
	{
	  update ( n, row, i + 1 );

	} // end if node exists

      } // end for each item in row

    } // end else hash node

  } // END private void update ( Node node, ARMinerItemset row, int index )

//********************************************************//

// not used in weka version, commenting out -kap

  /**
   * Update the weights of all indexed ARMinerItemsets that are included
   * in <code>row</code> and also update the matrix <code>counts</code>
   *
   * @param row  
   * the ARMinerItemset (normally a database row) against which 
   * we test for inclusion
   *
   * @param counts  
   * a matrix used by some algorithms to speed up
   * computations; its rows correspond to ARMinerItemsets and its columns
   * correspond to items; each value in the matrix tells for how many
   * times had an item appeared together with an itemset in the rows
   * of the database.
   */
  /*
  public void update ( ARMinerItemset row, long[][] counts )
  {
    update ( theRoot, row, 0, counts );
    unvisitLeaves();
  } // END public void update ( ARMinerItemset row, long[][] counts )
  */
//********************************************************//

// not used in Weka version, commenting out -kap

  /** private recursive method */
  /*
  private void update ( Node node, ARMinerItemset row, int index, 
			long[][] counts )
  {
    if ( node.type == LIST_NODE )
    {
      ListNode ln = (ListNode)node;

      if ( ln.visited )
      {
	  return;
      } // end if list node was visited

      for ( int i = 0; i < ln.size; i++ )
      {
	ARMinerItemset is = (ARMinerItemset)itemsets.get ( ln.indexes [ i ] );
	if ( is.isIncludedIn ( row ) > 0 )
	{
	  is.incrementWeight();
	  
	  for ( int j = 0; j < row.size(); j++ )
	    counts [ ln.indexes [ i ] ] [ row.getItem ( j ) - 1 ]++;
	}
      }

      ln.visited = true; // now we've seen this node

    } // end if list node
    else
    {
      // it must be a hash node
      HashNode hn = (HashNode)node;

      // this is a tricky piece of algorithm that ensures we
      // look for all possible subsets of row
      for ( int i = index; i < row.size(); i++ )
      {
	int key = row.getItem ( i );
	Node n = hn.children.get ( key );

	if ( n != null )
	{
	  update ( n, row, i + 1, counts );

	} // end if node exists

      } // end for each row

    } // end else hash node

  } // END private void update ( Node node, ARMinerItemset row, int index, 
  //			long[][] counts )
  */
//********************************************************//

// not used in weka version, commenting out -kap

  /**
   * Count how many frequent ARMinerItemsets (frequent = having weight 
   * greater than a specified minimum weight) are included in 
   * <code>itemset</code>
   *
   * @param itemset  
   * the ARMinerItemset for which we count the subsets
   *
   * @param minWeight  
   * the minimum weight
   */
  /*
  public long countFrequentSubsets ( ARMinerItemset itemset, 
				     long minWeight )
  {
    counter = 0;
    countFrequentSubsets ( theRoot, itemset, 0, minWeight );
    unvisitLeaves();
    return ( counter );

  } // END public long countFrequentSubsets ( ARMinerItemset itemset, 
  //				     long minWeight )
  */
//********************************************************//

// not used in weka version, commenting out -kap

  /** private recursive method */
  /*
  private void countFrequentSubsets ( Node node, 
				      ARMinerItemset itemset,
				      int index, long minWeight )
  {
    if ( node.type == LIST_NODE )
    {
      ListNode ln = (ListNode)node;
      
      if ( ln.visited )
      {
	  return;
      } // end if list node visited

      for ( int i = 0; i < ln.size; i++ )
      {
	ARMinerItemset is = (ARMinerItemset)itemsets.get ( ln.indexes [ i ] );

	if ( is.isIncludedIn ( itemset ) && is.getWeight() >= minWeight > 0 )
	{
	  counter++;

	} // end if included

      } // end for itemsets in list node

	ln.visited = true; // now we've seen this node
    } // end if list node
    else
    {
      // it must be a hash node
      HashNode hn = (HashNode)node;

      // this is a tricky piece of algorithm that ensures we
      // look for all possible subsets of row
      for ( int i = index; i < itemset.size(); i++ )
      {
	int key = itemset.getItem ( i );
	Node n = hn.children.get ( key );

	if ( n != null )
	{
	  countFrequentSubsets ( n, itemset, i + 1, minWeight );

	} // end if node exists

      } // end for items in itemset

    } // end else hash node

  } // END private void countFrequentSubsets ( Node node, 
  //				      ARMinerItemset itemset,
  //				      int index, long minWeight )
  */
//********************************************************//

  /**
   * Count how many ARMinerItemsets are included in <code>itemset</code>
   *
   * @param itemset  
   * the ARMinerItemset for which we count the subsets
   */

  public long countSubsets ( ARMinerItemset itemset )
  {
    counter = 0;
    countSubsets ( theRoot, itemset, 0 );
    unvisitLeaves();
    return ( counter );

  } // END public long countSubsets ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Count how many ARMinerItemsets are included in
   * <code>itemset</code> recursively.
   *
   * @param node  
   *
   * @param itemset  
   * the ARMinerItemset for which we count the subsets
   *
   * @param index  
   */

  private void countSubsets ( Node node, 
			      ARMinerItemset itemset, 
			      int index )
  {
    if ( node.type == LIST_NODE )
    {
      ListNode ln = (ListNode)node;

      if ( ln.visited )
      {
	return;
      } // end if this node was visited already

      for ( int i = 0; i < ln.size; i++ )
      {
	ARMinerItemset is = (ARMinerItemset)itemsets.get ( ln.indexes [ i ] );
	if ( is.isIncludedIn ( itemset ) )
	{
	  counter++;

	} // end if included

      } // end for itemsets in list node

      ln.visited = true; // now we've seen this node
    } // end if list node
    else
    {
      // must be a hash node
      HashNode hn = (HashNode)node;
      
      // this is a tricky piece of algorithm that ensures we
      // look for all possible subsets of row
      for ( int i = index; i < itemset.size(); i++ )
      {
	int key = itemset.getItem ( i );
	Node n = hn.children.get ( key );
	if ( n != null )
	{
	  countSubsets ( n, itemset, i + 1 );

	} // end if node exists

      } // end for items in itemset

    } // end else hash node

  } // END private void countSubsets ( Node node, 
  //			      ARMinerItemset itemset, 
  //			      int index )

//********************************************************//

// not used in weka version, commenting out -kap

  /**
   * Verifies if any of the indexed ARMinerItemsets is not large by checking
   * whether they're included in the frequent itemset <code>itemset</code>.
   * If an ARMinerItemset is not large then it will be marked.
   *
   * @param itemset  
   * the ARMinerItemset we check
   */
  /*
  public void checkLargeness ( ARMinerItemset itemset )
  {
    checkLargeness ( theRoot, itemset, 0 );
    unvisitLeaves();

  } // END public void checkLargeness ( ARMinerItemset itemset )
  */
//********************************************************//

// not used in weka version, commenting out -kap

  /** private recursive method */
  /*
  private void checkLargeness ( Node node, 
				ARMinerItemset itemset, 
				int index )
  {
    if ( node.type == LIST_NODE )
    {
      ListNode ln = (ListNode)node;
      
      if ( ln.visited )
      {
	return;
      } // end if node already visited

      for ( int i = 0; i < ln.size; i++ )
      {
	ARMinerItemset is = (ARMinerItemset)itemsets.get ( ln.indexes [ i ] );
	if ( is.isIncludedIn ( itemset ) > 0 )
	{
	  is.mark();

	} // end if included in, mark itemset

      } // end for itemsets in list node

	ln.visited = true; // now we've seen this node
    } // end if list node
    else
    {
      // must be a hash node
      HashNode hn = (HashNode)node;
      
      // this is a tricky piece of algorithm that ensures we
      // look for all possible subsets of row
      for ( int i = index; i < itemset.size(); i++ )
      {
	int key = itemset.getItem ( i );
	Node n = hn.children.get ( key );

	if ( n != null )
	{
	  checkLargeness ( n, itemset, i + 1 );

	} // end if node exists

      } // end for each item

    } // end else hash node

  } // END private void checkLargeness ( Node node, 
  //				ARMinerItemset itemset, 
  //				int index )
  */
//********************************************************//

} // END public class HashTree
