package wpi.filters;

import java.io.*;
import java.util.*;

import weka.core.*;
import weka.filters.*;

/** 
 * An instance filter that adds a new attribute to
 * the dataset and assigns the given value to all
 * the instances. Uses weka.filters.AddFilter
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version $Revision: 0.9 $
 */

public class AddWithValueFilter extends AddFilter 
  implements OptionHandler
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The value to assign */
  protected String value = "?";

//********************************************************//
// ----- Methods-----
//********************************************************//

  /** 
   * Sets the value to be assigned the attribute
   *
   * @param v
   * the value
   */

  public void setValue ( String v )
  {
    // if it's null, set it to empty string
    if ( v == null )
      {
	value = "";
      }
    else
      {
	value = v;
      }

  } // END public void setValue ( String v )

//********************************************************//

  /** 
   * Returns the value to be assigned the attribute
   *
   * @return
   * the value
   */

  public String getValue()
  {
    return ( value );

  } // END public String getValue()

//********************************************************//

  public String valueTextTip()
  {
    return ( "Value to set the new attribute in each instance." );

  } // END public String valueTextTip()

//********************************************************//

  /**
   * Input an instance for filtering. Filter requires
   * all instances be read before producing output.
   *
   * @param instance the input instance
   *
   * @return true if the filtered instance may now be
   * collected with output().
   *
   * @exception IllegalStateException
   * if no input format has been defined.
   * 
   */

  public boolean input ( Instance instance )
  {
    // make sure input format is set
    if ( getInputFormat() == null ) 
      {
	throw new IllegalStateException ( "No input instance format defined" );
      }

    // if this is a new batch clear the queue and 
    // signify that a batch has begun
    if ( m_NewBatch )
      {
	resetQueue();
	m_NewBatch = false;

      } // end if new batch

    Instance inst = (Instance)instance.copy();

    // First copy string values from input to output
    // Will break if the attribute being created is of type STRING (currently
    // AddFilter only adds NOMINAL or NUMERIC types)
    copyStringValues ( inst, true, inst.dataset(), getOutputFormat() );

    // Insert the new attribute and reassign to output
    inst.setDataset ( null );
    inst.insertAttributeAt ( m_Insert );
    inst.setDataset ( getOutputFormat() );

    // set the value of the new attribute
    // if the value is set that is
    if ( getValue().length() != 0 )
      {
	// if the attribute is numeric then we better
	// try to make the string a number
	if ( getOutputFormat().attribute ( getAttributeName() ).isNumeric() )
	  {
	    try 
	      {
		double d = Double.parseDouble ( getValue() );

		inst.setValue ( getOutputFormat().
				attribute ( getAttributeName() ), 
				d );
	      }
	    catch ( NumberFormatException ex )
	      {
		// just do the regular assignment and let that
		// take care of the errors
		inst.setValue ( getOutputFormat().
				attribute ( getAttributeName() ), 
				getValue() );
	      }
	    
	  } // end if attribute numeric
	else
	  {
	    inst.setValue ( getOutputFormat().
			    attribute ( getAttributeName() ), 
			    getValue() );

	  } // end else attribute nominal || string
      }

    // all set, push her on out
    push ( inst );
  
    // the output can be collected right away
    return ( true );
    
  } // END public boolean input ( Instance instance )

//********************************************************//

  /**
   * Returns a string decribing this filter.
   *
   * @return
   * a description of the filter suitable for displaying
   * in the explorer/experimenter gui
   *
   */

  public String globalInfo()
  {
    return ( "An instance filter that adds a new " +
	     "attribute to the dataset and assigns the " +
	     "given value to all the instances. " +
	     "Please see AddFilter for more details.");

  } // END public String globalInfo()

//********************************************************//

  /**
   * Returns an enumeration describing the available options
   *
   * @return
   * an enumeration of all the available options
   */

  public Enumeration listOptions() 
  {
    Enumeration superOptions = super.listOptions();

    Vector newVector = new Vector();

    while ( superOptions.hasMoreElements() )
      {
	newVector.addElement ( superOptions.nextElement() );	
      }

    newVector.addElement ( new Option (
              "\tSpecify the value to assign the new attribute.\n" +
	      "\tIf the new attribute is nominal the value\n" +
	      "\tmust appear in the given labels.",
              "V", 1, "-V <value>" ) );

    return ( newVector.elements() );

  } // END public Enumeration listOptions() 

//********************************************************//

  /**
   * Parses a list of options for this object. Valid options are:<p>
   *
   * @param options
   * the list of options as an array of strings
   *
   * -V value <br>
   * Specify the value to assign to the new attribute
   * (default unknown)<p>
   *
   * @exception Exception
   * if an option is not supported
   */

  public void setOptions ( String[] options ) throws Exception 
  {
    setValue ( Utils.getOption ( 'V', options ) );

    super.setOptions ( options );

  } // END public void setOptions ( String[] options ) 
  // throws Exception

//********************************************************//

  /**
   * Gets the current settings of the filter.
   *
   * @return
   * an array of strings suitable for passing to setOptions
   */

  public String [] getOptions() 
  {
    // have to return parent's options too
    String [] superOptions = super.getOptions();

    String [] options = new String [ superOptions.length + 2 ];
    int current = 0;

    // copy (sorta) the parent's options
    for ( ; current < superOptions.length; current++ )
      {
	options [ current ] = superOptions [ current ];
      }
    
    // add my own spiffy options
    if ( getValue() != "" && getValue() != null )
    {
      options [ current++ ] = "-V"; options [ current++ ] = getValue();
    } // end if we have an option to add

    // fill anything left over with empty's
    while ( current < options.length ) 
    {
      options [ current++ ] = "";
    }
    
    // all done cause we're good like that
    return ( options );

  } // END public String [] getOptions()

//********************************************************//
  
  /**
   * Main method for testing this class.
   *
   * @param argv
   * should contain arguments to the filter: 
   * use -h for help
   */

  public static void main ( String [] argv ) 
  {
    try 
      {
	if ( Utils.getFlag ( 'b', argv ) ) 
	  {
	    Filter.batchFilterFile ( new AddWithValueFilter(), argv ); 
	  } 
	else 
	  {
	    Filter.filterFile ( new AddWithValueFilter(), argv );
	  }
      } 
    catch ( Exception ex ) 
      {
	System.out.println ( ex.getMessage() );
      }

  } // END public static void main ( String [] argv )

//********************************************************//

} // END public class AddWithValueFilter extends AddFilter 
