package wpi.filters;

import java.io.*;
import java.util.*;

import wpi.timesequence.*;

import weka.core.*;
import weka.filters.*;

/*
  Maintainance log started on February 18, 2000

  February 18, 2003  - fixed minNumValues not showing in GUI. cannot
                       overload set method of options -kap

  February 24, 2003  - made consistent the convention of using a lower case
		       letter to start a method name. -kap
*/

/** 
 * Finds time periods during which events occur in time
 * sequence attribute values. Three events are currently
 * supported: increase, decrease and sustain.
 * A new attribute containing the events found is appended
 * after the last attribute.
 *
 * Valid filter-specific options are:<br>
 *
 * -A attributes <br>
 * Specify which attributes to find events in.
 *
 * -I <br>
 * Find increase events.
 *
 * -D <br>
 * Find decrease events.
 *
 * -S <br>
 * Find sustain events.
 *
 * -T <br>
 * Specify the tolerance to use when including values in events.
 *
 * -N <br>
 * Specify the minimum required number of values in an event
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version $Revision: 0.9 $
 */

public class EventFilter extends Filter 
  implements OptionHandler
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The string to append to an attribute's name for
      increase events. */
  public static String INCREASE_EVENTS = "-increase-events";

  /** The string to append to an attribute's name for
      decrease events. */
  public static String DECREASE_EVENTS = "-decrease-events";

  /** The string to append to an attribute's name for
      decrease events. */
  public static String SUSTAIN_EVENTS = "-sustain-events";

  /* Just so we can switch on event type */
  private final static int INCREASE_INT = 1;
  private final static int DECREASE_INT = 2;
  private final static int SUSTAIN_INT = 3;

  /** The delimiters to use for sets of events */
  public static char SETDELIMITER = '^';
  public static char SETBEGIN = '{';
  public static char SETEND = '}';

  /** The default tolerance */
  public static double TOLERANCE_DEFAULT = 0.10;

  /** The default min number of values in an event */
  public static int MIN_NUM_VALUES_DEFAULT = 2;

  /** list of time sequence attributes to find events in */
  private Vector timeSequenceAttributeVector = null;

  /** The original string for attributes to look for events in */
  protected String attributeString = "";

  /** tolerance for events */
  protected double tolerance = TOLERANCE_DEFAULT;

  /** minimum required number of values for events */
  protected int minNumValues = MIN_NUM_VALUES_DEFAULT;

  /** find increase events? */
  protected boolean increaseEvents = true;

  /** find decrease events? */
  protected boolean decreaseEvents = true;

  /** find sustain events? */
  protected boolean sustainEvents = true;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /** 
   * Sets the attribute to find events in
   *
   * @param a
   * the attribute
   */

  public void setAttributeString ( String a )
  {
    // save original option
    attributeString = a;

    // if it's null or empty, set the list to null
    if ( (a == null) || a.equals ( "" ) )
      {
	timeSequenceAttributeVector = null;
      }
    else
      {
	timeSequenceAttributeVector = new Vector();

	// parse the string for a list
	StringTokenizer stringTokenizer = new StringTokenizer ( a, "," );

	while ( stringTokenizer.hasMoreTokens() )
	{
	  timeSequenceAttributeVector.
	    addElement ( stringTokenizer.nextToken() );

	} // end while more tokens

      } // end else parse option

  } // END public void setAttributeString ( String a )

//********************************************************//

  /** 
   * Returns the time sequence attributes to find events in
   *
   * @return
   * a list of the attributes
   */

  public Vector getAttributes()
  {
    return ( timeSequenceAttributeVector );

  } // END public Vector getAttributes()

//********************************************************//
  
  public String attributeStringTipText()
  {
    return ( "Specify which attributes to find events in." );
  } // END public String attributeStringTipText()

//********************************************************//

  /** 
   * Returns the original option for attributes to find
   * events in
   *
   * @return
   * the original option string
   */

  public String getAttributeString()
  {
    return ( attributeString );

  } // END public String getAttributeString()

//********************************************************//

  /**
   * Sets the event tolerance.
   *
   * @param t
   * the percentage by which values can differ and still
   * be included into an event.
   */

  public void setToleranceString ( String t )
  {
    try
    {
      setTolerance ( Double.parseDouble ( t ) );
    }
    catch ( NumberFormatException ex )
    {
      setTolerance ( TOLERANCE_DEFAULT );
    } // end else use default

  } // END public void setToleranceString ( String t )

//********************************************************//

  /**
   * Sets the event tolerance.
   *
   * @param t
   * the percentage by which values can differ and still
   * be included into an event.
   */

  public void setTolerance ( double t )
  {
    tolerance = t;

  } // END public void setTolerance ( double t )

//********************************************************//

  /**
   * Returns the tolerance value in use.
   *
   * @return
   * the tolerance
   */

  public double getTolerance()
  {
    return ( tolerance );

  } // END public double getTolerance()

//********************************************************//
  
  public String toleranceTipText()
  {
    return ( "Specify the tolerance to use when including values in events." );
  } // END public String toleranceTipText()

//********************************************************//

  /**
   * Sets the minimum required number of values in an event
   *
   * @param n
   * the minimum number of values in an event
   */

  public void setMinNumValues ( String n )
  {
    try
    {
      setMinNumValues ( Integer.parseInt ( n ) );
    }
    catch ( NumberFormatException ex )
    {
      setMinNumValues ( MIN_NUM_VALUES_DEFAULT );
    } // end else use default

  } // END public void setMinNumValues ( String n )

//********************************************************//

  /**
   * Sets the minimum required number of values in an event
   *
   * @param n
   * the minimum number of values in an event
   */

  public void setMinNumValues ( int n )
  {
    minNumValues = n;

  } // END public void setMinNumValues ( int n )

//********************************************************//

  /**
   * Returns the minimum required number of values for events
   *
   * @return
   * the min number
   */

  public int getMinNumValues()
  {
    return ( minNumValues );

  } // END public int getMinNumValues()

//********************************************************//
  
  public String minNumValuesTipText()
  {
    return ( "Specify the minimum required number of values in an event." );
  } // END public String minNumValuesTipText()

//********************************************************//

  /**
   * Sets if the filter should find increase events.
   *
   * @param b
   * true to find increase events, false otherwise
   */

  public void setIncreaseEvents ( boolean b )
  {
    increaseEvents = b;

  } // END public void setIncreaseEvents ( boolean b )

//********************************************************//

  /**
   * Returns if the filter is set to find increase events.
   *
   * @return
   * true if increase events are looked for, false otherwise
   */

  public boolean getIncreaseEvents()
  {
    return ( increaseEvents );

  } // END public boolean getIncreaseEvents()

//********************************************************//
  
  public String increaseEventsTipText()
  {
    return ( "Specify if increase events should be found." );
  } // END public String increaseEventsTipText()

//********************************************************//

  /**
   * Sets if the filter should find decrease events.
   *
   * @param b
   * true to find decrease events, false otherwise
   */

  public void setDecreaseEvents ( boolean b )
  {
    decreaseEvents = b;

  } // END public void setDecreaseEvents ( boolean b )

//********************************************************//

  /**
   * Returns if the filter is set to find decrease events.
   *
   * @return
   * true if decrease events are looked for, false otherwise
   */

  public boolean getDecreaseEvents()
  {
    return ( decreaseEvents );

  } // END public boolean getDecreaseEvents()

//********************************************************//
  
  public String decreaseEventsTipText()
  {
    return ( "Specify if decrease events should be found." );
  } // END public String decreaseEventsTipText()

//********************************************************//

  /**
   * Sets if the filter should find sustain events.
   *
   * @param b
   * true to find sustain events, false otherwise
   */

  public void setSustainEvents ( boolean b )
  {
    sustainEvents = b;

  } // END public void setSustainEvents ( boolean b )

//********************************************************//

  /**
   * Returns if the filter is set to find sustain events.
   *
   * @return
   * true if sustain events are looked for, false otherwise
   */

  public boolean getSustainEvents()
  {
    return ( sustainEvents );

  } // END public boolean getSustainEvents()

//********************************************************//
  
  public String sustainEventsTipText()
  {
    return ( "Specify if sustain events should be found." );
  } // END public String sustainEventsTipText()

//********************************************************//

  /**
   * Finds events of the specified type in the instance
   * for the attribute specified.
   *
   * @return
   * <ul>
   * <li>the events</li>
   * <li>empty string if no events found</li>
   * </ul>
   */

  private String getEvents ( Instance instance, 
			     String attribute,
			     int eventType )
  {
    // get the values
    double [] values = TSUtility.getValues 
      ( instance.stringValue ( getInputFormat().attribute ( attribute ) ) );
    
    // the events
    StringBuffer events = new StringBuffer();
    
    // ok, now we look for the event
    // we need to track beginning and end of events
    int begin = -1;
    int end = -1;

    // BEGIN Just used for sustain events right now

    // we also need to track how much we will tolerate
    // for change and still consider it
    double lower = 0;
    double upper = 0;

    // the same, start with 5% of the first value
    if ( values.length > 1 )
    {
      lower = values [ 0 ] - ( values [ 0 ] * getTolerance() );
      upper = values [ 0 ] + ( values [ 0 ] * getTolerance() );

    } // end if there is at least two values

    // END Just used for sustain events right now

    // we need to track the number of values we have in an event
    int numValues = 0;
    boolean checkEvent = false;

    for ( int i = 0; i < values.length - 1; i++ )
    {
      // compare the current value and the next
      // if it increases
      switch ( eventType )
      {
      case INCREASE_INT:
	// compare the current value and the next
	// if it increases
	checkEvent = values [ i ] < values [ i + 1 ];
	break;

      case DECREASE_INT:
	// compare the current value and the next
	// if it decreases
	checkEvent = values [ i ] > values [ i + 1 ];
	break;

      case SUSTAIN_INT:
	// compare the current value and the next
	// if it stays about the same
	checkEvent = ( lower <= values [ i + 1 ] ) &&
	  ( upper >= values [ i + 1 ] );
	break;

      default:
	checkEvent = false;

      } // end switch event type

      if ( checkEvent )
      {
	switch ( begin )
	{
	case -1:
	  begin = i;
	  end = i + 1;
	  numValues = 2;
	  break;
	default:
	  end = i + 1;
	  numValues++;
	} // end switch on begin
      }  // end if current value less than next value
      else
      {
	// for sustain event
	lower = values [ i + 1 ] - values [ i + 1 ] * getTolerance();
	upper = values [ i + 1 ] + values [ i + 1 ] * getTolerance();
	
	switch ( begin )
	{
	case -1:
	  break;
	default:
	  // make sure we have enough events
	  if ( numValues >= getMinNumValues() )
	  {
	    if ( events.length() > 0 )
	    {
	      events.append ( SETDELIMITER );
	    } // end if this is not the first event
	    else
	    {
	      // then we are starting the set
	      events.append ( SETBEGIN );
	    } // end if beginning of set
	    
	    // add the event
	    events.append ( begin );
	    events.append ( TSUtility.TIME_SEQUENCE_DELIMITER );
	    events.append ( end );

	    // reset our counters
	    begin = -1;
	    end = -1;
	    numValues = 0;
	  } // end if we have enough values
	  break;

	} // end switch on begin
      } // end else not less than
      
    } // end for each pair of values
    
    // if begin and end are set we have one last event to add
    switch ( begin )
    {
    case -1:
      break;
    default:
      // make sure we have enough values
      if ( numValues >= getMinNumValues() )
      {
	if ( events.length() > 0 )
	{
	  events.append ( SETDELIMITER );
	} // end if this is not the first event
	else
	{
	  // then we are starting the set
	  events.append ( SETBEGIN );
	} // end if beginning of set
	
	// add the event
	events.append ( begin );
	events.append ( TSUtility.TIME_SEQUENCE_DELIMITER );
	events.append ( end );
      } // end if we have enough values
      break;
    } // end yet another switch on begin
    
    // if we have any events we must close the set
    if ( events.length() > 0 )
    {
      events.append ( SETEND );
    } // end if we have to close the set

    return ( events.toString() );

  } // END private String getEvents ( Instance instance,
  // String attribute, int eventType )

//********************************************************//

  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input instance
   * structure (any values of instances contained in the object are 
   * ignored) but in addition to the structure, this filter does require 
   * to know the number of instances in the input data set.
   *
   * @return true if the outputFormat may be collected immediately
   * @exception Exception if the format couldn't be set successfully
   */

  public boolean setInputFormat ( Instances instanceInfo ) 
    throws Exception 
  {
    // use filter's set input format
    super.setInputFormat ( instanceInfo );

    // output format may NOT be collected right away
    // we have to set the values for our string attributes
    // first
    return ( false );

  } // END   public boolean setInputFormat ( Instances instanceInfo ) 
  //    throws Exception 

 //********************************************************//

  /**
   * Sets the output format. Takes the buffered input
   * instances and generates valid strings for the new
   * attributes.
   */
  
  private void setOutputFormat() 
  {
    // make a new instances for the output format
    Instances outputFormat = new Instances ( getInputFormat(), 0 );

    // if attributes were specified just do those
    // otherwise we should try to find events in all the time
    // sequence attributes
    if ( timeSequenceAttributeVector == null )
    {
      findTimeSequenceAttributes();

    } // end if attributes not set

    for ( int i = 0; i < timeSequenceAttributeVector.size(); i++ )
    {
      // add increase events?
      if ( increaseEvents )
      {
	// only add an attribute if one does not already exist
	if ( outputFormat.attribute ( timeSequenceAttributeVector.
				      elementAt ( i ) + 
				      INCREASE_EVENTS ) == null )
	{
	  outputFormat.
	    insertAttributeAt ( new Attribute ( timeSequenceAttributeVector.
						elementAt ( i ) + 
						INCREASE_EVENTS, null ),
				outputFormat.numAttributes() );

	} // end if attribute must be added

      } // end if increase events
 
      // add decrease events?
      if ( decreaseEvents )
      {
	// only add if one does not already exist
	if ( outputFormat.attribute ( timeSequenceAttributeVector.
				      elementAt ( i ) + 
				      DECREASE_EVENTS ) == null )
	{
	  outputFormat.
	    insertAttributeAt ( new Attribute ( timeSequenceAttributeVector.
						elementAt ( i ) + 
						DECREASE_EVENTS, null ),
				outputFormat.numAttributes() );
	
	} // end if attribute must be added

      } // end if decrease events
 
      // add sustain events?
      if ( sustainEvents )
      {
	// only add if one does not already exist
	if ( outputFormat.attribute ( timeSequenceAttributeVector.
				      elementAt ( i ) + 
				      SUSTAIN_EVENTS ) == null )
	{
	  outputFormat.
	    insertAttributeAt ( new Attribute ( timeSequenceAttributeVector.
						elementAt ( i ) + 
						SUSTAIN_EVENTS, null ),
				outputFormat.numAttributes() );
	} // end if attribute must be added

      } // end if sustain events
      
    } // end for each event attribute

    setOutputFormat ( outputFormat );
    
  } // END private void setOutputFormat() 

 //********************************************************//

  /**
   * Finds the time sequence attributes in the input 
   * format. Any string attributes with set delimiters in
   * their value will NOT be interpreted as a time sequence
   * attribute.
   */

  protected void findTimeSequenceAttributes()
  {
    // make a new vector for storing the list
    // of new event attributes
    timeSequenceAttributeVector = new Vector();
    
    // go through all the attributes in the input
    // checking the type for string
    for ( int a = 0; a < getInputFormat().numAttributes(); a++ )
    {
      if ( getInputFormat().attribute ( a ).isString() )
      {
	// check the values
	Enumeration enum = getInputFormat().attribute ( a ).
	  enumerateValues();

	// temp string for checking stuff
	String tempString;
	
	while ( enum.hasMoreElements() )
	{
	  // does the delimiter appear? if so add this
	  // attribute to our list and break
	  tempString = (String)enum.nextElement();

	  if ( tempString.indexOf ( TSUtility.TIME_SEQUENCE_DELIMITER ) > -1 )
	  {
	    // make sure a set delimeter does NOT appear
	    if ( ( tempString.indexOf ( SETBEGIN ) < 0 ) &&
		 ( tempString.indexOf ( SETEND ) < 0 ) &&
		 ( tempString.indexOf ( SETDELIMITER ) < 0 ) )
	    {
	      // add it to our list
	      timeSequenceAttributeVector.
		addElement ( getInputFormat().attribute ( a ).name() );
	    // we don't have to look any longer at this attribute's
	    // values
	    break;

	    } // end if a set delimiter does NOT appear

	  } // end if we found the delimiter
	  
	} // end while enum has more
	
      } // end if attribute type string
      
    } // end for each attribute
    
  } // END protected void findTimeSequenceAttributes()
  
 //********************************************************//

  /**
   * Input an instance for filtering. Ordinarily the
   * instance is processed and made available for output
   * immediately. Some filters require all instances be
   * read before producing output.
   *
   * @param instance the input instance
   *
   * @return true if the filtered instance may now be
   * collected with output().
   *
   * @exception IllegalStateException
   * if no input format has been defined.
   * 
   */

  public boolean input ( Instance instance )
  {
    // make sure input format is set
    if ( getInputFormat() == null ) 
      {
	throw new IllegalStateException ( "No input instance format defined" );
      }

    // if this is a new batch clear the queue and 
    // signify that a batch has begun
    if ( m_NewBatch )
      {
	resetQueue();
	m_NewBatch = false;

      } // end if new batch

    // save the instance for processing later
    bufferInput ( instance );

    // the output cannot be collected yet
    return ( false );

  } // END public boolean input ( Instance instance )

//********************************************************//

  /**
   * Signifies that this batch of input to the filter is
   * finished. 
   *
   * @return true
   * if there are instances pending output.
   *
   * @exception IllegalStateException
   * if no input structure has been defined.
   */

  public boolean batchFinished() 
  {
    // make sure we have an input format
    if ( getInputFormat() == null ) 
      {
	throw new IllegalStateException ( "No input instance format defined" );
      }
    
    // if the output format is not defined then set it and
    // generate the output instance
    if ( ! isOutputFormatDefined() ) 
      {
	setOutputFormat();

	// go through all the buffered instances
	for ( int i = 0; i < getInputFormat().numInstances(); i++ )
	{	  	
	  // make a new instance with the output data set
	  Instance newInstance = new Instance ( getOutputFormat().
						numAttributes() );
	  newInstance.setDataset ( getOutputFormat() );
	  
	  // set all the attributes of this new instance
	  // with values from a buffered input instance
	  for ( int a = 0; a < getInputFormat().numAttributes(); a++ )
	  {
	    if ( getOutputFormat().attribute ( a ).isNumeric() )
	    {
	      newInstance.setValue ( a, getInputFormat().instance ( i ).
				     value ( a ) );

	    }  // end if numeric set value using value
	    else
	    {
	      newInstance.setValue ( a, getInputFormat().instance ( i ).
				     stringValue ( a ) );
	    } // end else set using stringValue

	  } // end for all input attributes

	  // get the events found
	  for ( int e = 0; e < timeSequenceAttributeVector.size(); e++ )
	  {
	    // should we look for increase events?
	    if ( increaseEvents )
	    {
	      newInstance.
		setValue ( getOutputFormat().
			   attribute ( timeSequenceAttributeVector.
				       elementAt ( e ) + 
				       INCREASE_EVENTS ),
			   getEvents ( getInputFormat().instance ( i ),
				       (String)timeSequenceAttributeVector.
				       elementAt ( e ), INCREASE_INT ) );
	    } // end if increase events

	    // should we look for decrease events?
	    if ( decreaseEvents )
	    {
	      newInstance.
		setValue ( getOutputFormat().
			   attribute ( timeSequenceAttributeVector.
				       elementAt ( e ) + 
				       DECREASE_EVENTS ),
			   getEvents ( getInputFormat().instance ( i ),
				       (String)timeSequenceAttributeVector.
				       elementAt ( e ), DECREASE_INT ) );
	    } // end if decrease events

	    // should we look for sustain events?
	    if ( sustainEvents )
	    {
	      newInstance.
		setValue ( getOutputFormat().
			   attribute ( timeSequenceAttributeVector.
				       elementAt ( e ) + 
				       SUSTAIN_EVENTS ),
			   getEvents ( getInputFormat().instance ( i ),
				       (String)timeSequenceAttributeVector.
				       elementAt ( e ), SUSTAIN_INT ) );
	    } // end if sustain events
	    
	  } // end for each attribute to find events in

	  // all set, push her on out
	  push ( newInstance );
	    
	} // end for all input instances
	
      } // end if output format not defined
    
    flushInput();
    m_NewBatch = true;
    return ( numPendingOutput() != 0 );

  } // END public boolean batchFinished() 

//********************************************************//

  /**
   * Returns a string describing this filter.
   *
   * @return
   * a description of the filter suitable for displaying
   * in the explorer/experimenter gui
   *
   */

  public String globalInfo()
  {
    return ( "Finds events in time sequence attributes." );

  } // END public String globalInfo()

//********************************************************//

  /**
   * Returns an enumeration describing the available options
   *
   * @return
   * an enumeration of all the available options
   */

  public Enumeration listOptions() 
  {
    Vector newVector = new Vector ( 6 );

    newVector.addElement ( new Option 
			   ( "\tSpecify the attributes to find events in.",
			     "A", 1, "-A <value>" ) );

    newVector.addElement ( new Option 
			   ( "\tSpecifies the tolerance for including values in events.",
			     "T", 1, "-T <value>" ) );

    newVector.addElement ( new Option 
			   ( "\tSpecifies the minimum required number of values in events.",
			     "N", 1, "-N <value>" ) );

    newVector.addElement ( new Option 
			   ( "\tFind increase events.",
			     "I", 0, "-I" ) );

    newVector.addElement ( new Option 
			   ( "\tFind decrease events.",
			     "D", 0, "-D" ) );

    newVector.addElement ( new Option 
			   ( "\tFind sustain events.",
			     "S", 1, "-S" ) );

    return ( newVector.elements() );

  } // END public Enumeration listOptions() 

//********************************************************//

  /**
   * Parses a list of options for this object. Valid
   * options are:
   *
   * @param options
   * the list of options as an array of strings
   *
   * -A attributes <br>
   * Specify the attributes to find events in
   *
   * -I <br>
   * Find increase events
   *
   * -D <br>
   * Find decrease events
   *
   * -S <br>
   * Find sustain events. 
   *
   * -T tolerance <br>
   * The tolerance for event finding
   *
   * -N min number <br>
   * The minimum required number of values for events
   * 
   * @exception Exception
   * if an option is not supported
   */

  public void setOptions ( String[] options ) throws Exception 
  {
    setAttributeString ( Utils.getOption ( 'A', options ) );
    setToleranceString ( Utils.getOption ( 'T', options ) );
    setMinNumValues ( Utils.getOption ( 'N', options ) );

    setIncreaseEvents ( Utils.getFlag ( 'I', options ) );
    setDecreaseEvents ( Utils.getFlag ( 'D', options ) );

    setSustainEvents ( Utils.getFlag ( 'S', options ) );
    
  } // END public void setOptions ( String[] options ) 
  // throws Exception

//********************************************************//

  /**
   * Gets the current settings of the filter.
   *
   * @return
   * an array of strings suitable for passing to setOptions
   */

  public String [] getOptions() 
  {
    String [] options = new String [12];
    int current = 0;

    // add my own spiffy options
    options [ current++ ] = "-A"; options [ current++ ] = getAttributeString();
    options [ current++ ] = "-T"; options [ current++ ] = "" + getTolerance();
    options [ current++ ] = "-N"; options [ current++ ] = "" + getMinNumValues();

    if ( increaseEvents )
    {
      options [ current++ ] = "-I"; options [ current++ ] = "";
    }

    if ( decreaseEvents )
    {
      options [ current++ ] = "-D"; options [ current++ ] = "";
    }

    if ( sustainEvents )
    {
      options [ current++ ] = "-S"; options [ current++ ] = "";
    }

    // fill anything left over with empties
    while ( current < options.length ) 
    {
      options [ current++ ] = "";
    }
    
    // all done cause we're good like that
    return ( options );

  } // END public String [] getOptions()

//********************************************************//
  
  /**
   * Main method for testing this class.
   *
   * @param argv
   * should contain arguments to the filter: 
   * use -h for help
   */

  public static void main ( String [] argv ) 
  {
    try 
      {
	if ( Utils.getFlag ( 'b', argv ) ) 
	  {
	    Filter.batchFilterFile ( new EventFilter(), argv ); 
	  } 
	else 
	  {
	    Filter.filterFile ( new EventFilter(), argv );
	  }
      } 
    catch ( Exception ex ) 
      {
	System.out.println ( ex.getMessage() );
      }

  } // END public static void main ( String [] argv )

//********************************************************//

} // END public class EventFilter extends Filter 
  //  implements OptionHandler
